/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group.                                               *
 * Copyright by the Board of Trustees of the University of Illinois.         *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of HDF5.  The full HDF5 copyright notice, including     *
 * terms governing use, modification, and redistribution, is contained in    *
 * the files COPYING and Copyright.html.  COPYING can be found at the root   *
 * of the source code distribution tree; Copyright.html can be found at the  *
 * root level of an installed copy of the electronic HDF5 document set and   *
 * is linked from the top-level documents page.  It can also be found at     *
 * http://hdfgroup.org/HDF5/doc/Copyright.html.  If you do not have          *
 * access to either file, you may request a copy from help@hdfgroup.org.     *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "nagg_filayopt.h" 
#include "nagg.h"

/*-------------------------------------------------------------------------
 * File: nagg_filayopt.c
 * Purpose: Public API functions
 *
 * This file was copied from the HDF5 1.8.12 version of h5repack.c in October
 * 2013 for its filter and layout options.  h5repack in function names  has
 * been replaced by nagg_filayopt to avoid future collisions. 
 *-------------------------------------------------------------------------
 */

static int check_options(pack_opt_t *options);
static int check_objects(const char* fname, pack_opt_t *options);
static const char* get_sfilter(H5Z_filter_t filtn);
static int have_request(pack_opt_t *options);

/*-------------------------------------------------------------------------
 * Function: nagg_filayopt
 *
 * Purpose: locate all high-level HDF5 objects in the file
 *  and compress/chunk them using options
 *
 * Algorithm: 2 traversals are made to the file; the 1st builds a list of
 *  the objects, the 2nd makes a copy of them, using the options;
 *  the reason for the 1st traversal is to check for invalid
 *  object name requests
 *
 * Return: 0, ok, -1, fail
 *
 * Programmer: pvn@ncsa.uiuc.edu
 *
 * Date: September, 22, 2003
 *
 *-------------------------------------------------------------------------
 */

/*-------------------------------------------------------------------------
 * Function: nagg_filayopt_init
 *
 * Purpose: initialize options
 *
 * Return: 0, ok, -1, fail
 *
 *-------------------------------------------------------------------------
 */

int nagg_filayopt_init(pack_opt_t *options, int verbose) {
	int k, n;

	HDmemset(options, 0, sizeof(pack_opt_t));
	options->min_comp = MIN_FILT_LAYT_APPLY_SIZE; /* default minimum dataset size for applying 
                                                         layout and filters. Use --minimum option 
                                                         to override. */
	options->verbose = verbose;
	options->layout_g = H5D_LAYOUT_ERROR;

	for (n = 0; n < H5_REPACK_MAX_NFILTERS; n++) {
		options->filter_g[n].filtn = -1;
		options->filter_g[n].cd_nelmts = 0;
		for (k = 0; k < CD_VALUES; k++)
			options->filter_g[n].cd_values[k] = 0;
	}

	return (options_table_init(&(options->op_tbl)));
}

/*-------------------------------------------------------------------------
 * Function: nagg_filayopt_end
 *
 * Purpose: free options table
 *
 *-------------------------------------------------------------------------
 */

int nagg_filayopt_end(pack_opt_t *options) {
	return options_table_free(options->op_tbl);
}

/*-------------------------------------------------------------------------
 * Function: nagg_filayopt_addfilter
 *
 * Purpose: add a compression -f option to table
 *   Example: -f dset:GZIP=6
 *
 * Return: 0, ok, -1, fail
 *
 *-------------------------------------------------------------------------
 */
int nagg_filayopt_addfilter(const char* str, pack_opt_t *options) {
	obj_list_t *obj_list = NULL; /* one object list for the -f and -l option entry */
	filter_info_t filter; /* filter info for the current -f option entry */
	int n_objs; /* number of objects in the current -f or -l option entry */
	int is_glb; /* is the filter global */

	/* parse the -f option */
	if (NULL == (obj_list = parse_filter(str, &n_objs, &filter, options, &is_glb)))
		return -1;

	/* if it applies to all objects */
	if (is_glb) {
		int n;

		n = options->n_filter_g++; /* increase # of global filters */

		if (options->n_filter_g > H5_REPACK_MAX_NFILTERS) {
                        NAGG_ERROR("nagg_filayopt_addfilter(): maximum number of filters exceeded for ");
                        HDfprintf(stderr,"<%s>\n", str);
	/*		error_msg("maximum number of filters exceeded for <%s>\n", str); */
			HDfree(obj_list);
			return -1;
		}

		options->filter_g[n] = filter;
	}
	else
		options_add_filter(obj_list, n_objs, filter, options->op_tbl);

	HDfree(obj_list);
	return 0;
}

/*-------------------------------------------------------------------------
 * Function: nagg_filayopt_addlayout
 *
 * Purpose: add a layout option
 *
 * Return: 0, ok, -1, fail
 *
 *-------------------------------------------------------------------------
 */

int nagg_filayopt_addlayout(const char* str, pack_opt_t *options) {

	obj_list_t *obj_list = NULL; /*one object list for the -t and -c option entry */
	int n_objs; /*number of objects in the current -t or -c option entry */
	pack_info_t pack; /*info about layout to extract from parse */
	int j;
	int ret_value = -1;

	init_packobject(&pack);

        /* This catches the case where a layout is set after it's already been set to
         * all, but not the case where it is set to all after a layout has been set 
         * for specific objects.  That has been added to the parse_layout() function
         * where all_layout is set to 1.
         */
	if (options->all_layout == 1) {
                NAGG_ERROR("nagg_filayopt_addlayout(): invalid layout input: 'all' option \
                             is present with other layout options ");
                HDfprintf(stderr,"<%s>\n", str);
		return ret_value;
	}

	/* parse the layout option */
	obj_list = parse_layout(str, &n_objs, &pack, options);
	if (obj_list) {

		/* set layout option */
		options->layout_g = pack.layout;

		/* no individual dataset specified */
		if (options->all_layout == 1) {
			if (pack.layout == H5D_CHUNKED) {
				/* -2 means the NONE option, remove chunking
				 and set the global layout to contiguous */
				if (pack.chunk.rank == -2) {
					options->layout_g = H5D_CONTIGUOUS;
				}
				/* otherwise set the global chunking type */
				else {
					options->chunk_g.rank = pack.chunk.rank;
					for (j = 0; j < pack.chunk.rank; j++)
						options->chunk_g.chunk_lengths[j] =
								pack.chunk.chunk_lengths[j];
				}
			}
		}

		/* individual dataset specified */
		if (options->all_layout == 0)
			ret_value = options_add_layout(obj_list, n_objs, &pack, options->op_tbl);

		HDfree(obj_list);
		ret_value = 0;
	}

	return ret_value;
}

/*-------------------------------------------------------------------------
 * Function: check_options
 *
 * Purpose: print options, checks for invalid options
 *
 * Return: void, return -1 on error
 *
 * Programmer: pvn@ncsa.uiuc.edu
 *
 * Date: September, 22, 2003
 *
 * Modification:
 *   Peter Cao, July 9, 2007
 *   Add "-L, --latest" and other options to pack a file with the latest file format
 *
 *-------------------------------------------------------------------------
 */
static int check_options(pack_opt_t *options) {
	unsigned int i;
	int k, j, has_cp = 0, has_ck = 0;
	char slayout[30];

	/*-------------------------------------------------------------------------
	 * objects to layout
	 *-------------------------------------------------------------------------
	 */
	if (options->verbose && have_request(options) /* only print if requested */) {
		printf("Objects to modify layout are...\n");
		if (options->all_layout == 1) {
			switch (options->layout_g) {
			case H5D_COMPACT:
				strcpy(slayout, "compact");
				break;
			case H5D_CONTIGUOUS:
				strcpy(slayout, "contiguous");
				break;
			case H5D_CHUNKED:
				strcpy(slayout, "chunked");
				break;
			case H5D_LAYOUT_ERROR:
			case H5D_NLAYOUTS:
                        	NAGG_ERROR("check_options(): invalid layout\n");
			/*	error_msg("invalid layout\n"); */
				return -1;
			default:
				strcpy(slayout, "invalid layout\n");
				return -1;
			}
			printf(" Apply %s layout to all\n", slayout);
			if (H5D_CHUNKED == options->layout_g) {
				printf("with dimension [");
				for (j = 0; j < options->chunk_g.rank; j++)
					printf("%d ", (int) options->chunk_g.chunk_lengths[j]);
				printf("]\n");
			}
		}
	}/* verbose */

	for (i = 0; i < options->op_tbl->nelems; i++) {
		char* name = options->op_tbl->objs[i].path;

		if (options->op_tbl->objs[i].chunk.rank > 0) {
			if (options->verbose) {
				printf(" <%s> with chunk size ", name);
				for (k = 0; k < options->op_tbl->objs[i].chunk.rank; k++)
					printf("%d ",
							(int) options->op_tbl->objs[i].chunk.chunk_lengths[k]);
				printf("\n");
			}
			has_ck = 1;
		}
		else if (options->op_tbl->objs[i].chunk.rank == -2) {
			if (options->verbose)
				printf(" <%s> %s\n", name, "NONE (contigous)");
			has_ck = 1;
		}
	}

	if (options->all_layout == 1 && has_ck) {
            NAGG_ERROR("check_options(): invalid chunking input: 'all' option\
                        is present with other layout options\n");
	    return -1;
	}

	/*-------------------------------------------------------------------------
	 * objects to filter
	 *-------------------------------------------------------------------------
	 */

	if (options->verbose && have_request(options) /* only print if requested */) {
		printf("Objects to apply filter are...\n");
		if (options->all_filter == 1) {
			for (k = 0; k < options->n_filter_g; k++) {
				H5Z_filter_t filtn = options->filter_g[k].filtn;
				switch (filtn) {
				case H5Z_FILTER_NONE:
					printf(" Uncompress all\n");
					break;
				case H5Z_FILTER_SHUFFLE:
				case H5Z_FILTER_FLETCHER32:
					printf(" All with %s\n", get_sfilter(filtn));
					break;
				case H5Z_FILTER_SZIP:
				case H5Z_FILTER_DEFLATE:
					printf(" All with %s, parameter %d\n", get_sfilter(filtn),
							options->filter_g[k].cd_values[0]);
					break;
				default:
					printf(" User Defined %d\n", filtn);
					break;
				} /* k */
			};
		}
	} /* verbose */

	for (i = 0; i < options->op_tbl->nelems; i++) {
		pack_info_t pack = options->op_tbl->objs[i];
		char* name = pack.path;

		for (j = 0; j < pack.nfilters; j++) {
			if (options->verbose) {
				printf(" <%s> with %s filter\n", name,
						get_sfilter(pack.filter[j].filtn));
			}

			has_cp = 1;

		} /* j */
	} /* i */

	if (options->all_filter == 1 && has_cp) {
                NAGG_ERROR("check_options(): invalid compression input: 'all' option\
                            is present with other objects\n");
/*		error_msg(
				"invalid compression input: 'all' option\
                            is present with other objects\n"); */
		return -1;
	}

	/*-------------------------------------------------------------------------
	 * check options for the latest format
	 *-------------------------------------------------------------------------
	 */

	if (options->grp_compact < 0) {
                NAGG_ERROR("check_options(): invalid maximum number of links to \
                            store as header messages\n");
	/*	error_msg(
				"invalid maximum number of links to store as header messages\n"); */
		return -1;
	}
	if (options->grp_indexed < 0) {
                NAGG_ERROR("check_options(): invalid minimum number of links to \
                            store in the indexed format\n");
/*		error_msg(
				"invalid minimum number of links to store in the indexed format\n"); */
		return -1;
	}
	if (options->grp_indexed > options->grp_compact) {
                NAGG_ERROR("check_options(): minimum indexed size is greater \
                            than the maximum compact size\n");
/*		error_msg(
				"minimum indexed size is greater than the maximum compact size\n"); */
		return -1;
	}
	for (i = 0; i < 8; i++) {
		if (options->msg_size[i] < 0) {
                	NAGG_ERROR("check_options(): invalid shared message size\n");
		/*	error_msg("invalid shared message size\n"); */
			return -1;
		}
	}

	/*--------------------------------------------------------------------------------
	 * verify new user userblock options; file name must be present
	 *---------------------------------------------------------------------------------
	 */
	if (options->ublock_filename != NULL && options->ublock_size == 0) {
		if (options->verbose) {
			printf(
					"Warning: user block size missing for file %s. Assigning a default size of 1024...\n",
					options->ublock_filename);
			options->ublock_size = 1024;
		}
	}

	if (options->ublock_filename == NULL && options->ublock_size != 0) {
                NAGG_ERROR("check_options(): file name missing for user block\n");
	/*	error_msg("file name missing for user block\n",
				options->ublock_filename); */
		return -1;
	}

	/*--------------------------------------------------------------------------------
	 * verify alignment options; threshold is zero default but alignment not
	 *---------------------------------------------------------------------------------
	 */

	if (options->alignment == 0 && options->threshold != 0) {
                NAGG_ERROR("check_options(): alignment for H5Pset_alignment missing\n");
	/*	error_msg("alignment for H5Pset_alignment missing\n"); */
		return -1;
	}

	return 0;
}

/*-------------------------------------------------------------------------
 * Function: check_objects
 *
 * Purpose: locate all HDF5 objects in the file and compare with user
 *  supplied list
 *
 * Return: 0, ok, -1 no
 *
 * Programmer: Pedro Vicente, pvn@ncsa.uiuc.edu
 *
 * Date: September, 23, 2003
 *
 *-------------------------------------------------------------------------
 */
/* Removed this function - not needed for nagg. */

/*-------------------------------------------------------------------------
 * Function: have_request
 *
 * Purpose: check if a filter or layout was requested
 *
 * Return: 1 yes, 0 no
 *
 * Date: May, 24, 2007
 *
 *-------------------------------------------------------------------------
 */
static int have_request(pack_opt_t *options) {

	if (options->all_filter || options->all_layout || options->op_tbl->nelems)
		return 1;

	return 0;

}

/*-------------------------------------------------------------------------
 * Function: get_sfilter
 *
 * Purpose: return the filter as a string name
 *
 * Return: name of filter, exit on error
 *
 *-------------------------------------------------------------------------
 */

static const char* get_sfilter(H5Z_filter_t filtn) {
	if (filtn == H5Z_FILTER_NONE)
		return "NONE";
	else if (filtn == H5Z_FILTER_DEFLATE)
		return "GZIP";
	else if (filtn == H5Z_FILTER_SZIP)
		return "SZIP";
	else if (filtn == H5Z_FILTER_SHUFFLE)
		return "SHUFFLE";
	else if (filtn == H5Z_FILTER_FLETCHER32)
		return "FLETCHER32";
	else if (filtn == H5Z_FILTER_NBIT)
		return "NBIT";
	else if (filtn == H5Z_FILTER_SCALEOFFSET)
		return "SOFF";
	else
		return "UD";
}

