/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group.                                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of the HDF NPP Aggregation Tool (nagg). The full        *
 * NPP Aggregation Tool copyright notice, including terms governing use,     *
 * modification, and redistribution, is contained in the file COPYING,       *
 * which can be found at the root of the source code distribution tree.      *
 * If you do not have access to this file, you may request a copy from       *
 * help@hdfgroup.org.                                                        *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <regex.h>
#include <nagg.h>

/* The following is the size of a buffer to contain any error messages

   encountered when the regular expression is compiled. */

#define MAX_ERROR_MSG 0x1000
#if 0
#define DEBUG 1
#endif

/* Compile the regular expression described by "regex_text" into
   "r". */

static int compile_regex (regex_t * r, const char * regex_text)
{
    int status = regcomp (r, regex_text, REG_EXTENDED|REG_NEWLINE);
    if (status != 0) {
        char error_message[MAX_ERROR_MSG];
        regerror (status, r, error_message, MAX_ERROR_MSG);
        printf ("Regex error compiling '%s': %s\n",
                 regex_text, error_message);
        return 1;
    }
    return 0;
}

static int match_regex (regex_t * r, char * to_match, inpfile_sort_t *infile)
{
    /* "P" is a pointer into the string which points to the end of the
       previous match. */
    const char * p = to_match;
    /* "N_matches" is the maximum number of matches allowed. */
    const int n_matches = 10;
    /* "M" contains the matches found. */
    regmatch_t m[n_matches];

    int i = 0;
    int nomatch = regexec (r, p, n_matches, m, 0);
    if (nomatch) {
        printf ("No more matches.\n");
        return nomatch;
    }
    i=0;
    int start;
    int finish;
    start = m[i].rm_so + (p - to_match);
    finish = m[i].rm_eo + (p - to_match);
#ifdef DEBUG
    if (i == 0) {
        printf ("$& is ");
    }
    else {
        printf ("$%d is ", i);
    }
    printf ("'%.*s' (bytes %d:%d)\n", (finish - start),
        to_match + start, start, finish);
    printf("Timestamp is %s.\n", to_match + start);
#endif            
    snprintf(infile->timestamp, (finish - start + 1), (to_match + start));
    infile->file = to_match;

    return 0;
}

/* Compare file1 and file2 by timestamp, which is the date and beginning time
 * fields of the filename (which happen to be contiguous). 
 * Return -1 if file1 < file2 (earlier timestamp)
 *         0 if file1 = file2
 *         1 if file1 > file2
 * Comparison is only by timestamp.
 */
static int
inpfile_compar(void *file1, void *file2)
{
    int return_val;
    inpfile_sort_t *pt1, *pt2;

    assert(file1!=NULL);
    assert(file2!=NULL);
    pt1 = (inpfile_sort_t*) file1;
    pt2 = (inpfile_sort_t*) file2;
/*
    pt1 = * (inpfile_sort_t* const *) file1;
    pt2 = * (inpfile_sort_t* const *) file2;
*/
#ifdef DEBUG
printf("enter inpfile_compar:\n");
/*printf("file1=%x, file2=%x\n", file1, file2);*/
printf("file1=%x, %s, %s\n", pt1, pt1->timestamp, pt1->file);
printf("file2=%x, %s, %s\n", pt2, pt2->timestamp, pt2->file);
#endif

    /* compare timestamp which is a string */
    if (0!=(return_val=HDstrcmp((pt1->timestamp), (pt2->timestamp)))){
        /* found difference. return result */
#ifdef DEBUG
        printf("file timestamps are different return_val=%d\n", return_val);
#endif
        goto done;
    };

done:
#ifdef DEBUG
    printf("leaving inpfile_compar with %d\n", return_val);
#endif
    return return_val;    
}

int nagg_sort_inputfiles(int ninputfiles, char **inputfiles, inpfile_sort_t *infiles) 
{
    regex_t r;
    const char * regex_text;
    const char * find_text;
    int i;
    
    regex_text = "d[0-9]{8}_t[0-9]{7}";

    compile_regex(& r, regex_text);
    
    for(i=0; i<ninputfiles; ++i) {
#ifdef DEBUG
        printf ("\nTrying to find '%s' in '%s'\n", regex_text, inputfiles[i]);
        printf("infiles[%d]=%x\n", i, &infiles[i]); 
#endif
        match_regex(& r, inputfiles[i], &infiles[i]);
    }
#ifdef DEBUG
    for(i=0; i<ninputfiles; ++i) {
        printf("Timestamp:  %s  Filename:  %s\n", infiles[i].timestamp, infiles[i].file);
    }
#endif

    HDqsort(&infiles[0], ninputfiles, sizeof(inpfile_sort_t), inpfile_compar);

#ifdef DEBUG
    for(i=0; i<ninputfiles; ++i) {
        printf("Timestamp:  %s  Filename:  %s\n", infiles[i].timestamp, infiles[i].file);
#endif

    for(i=0; i<ninputfiles; ++i) {
        inputfiles[i] = infiles[i].file;
    }
   
    regfree (& r);
    /* For now we're done with infiles, but the timestamps may be useful when
     * processing the input files.  If so, the malloc and free will be moved,
     * probably to nagg_main.c.
     */
}

iet_t nagg_get_beginning_time(char *filename) {
    hid_t file, grp, attr;
    ssize_t size;
    char grp_name[NAGG_Product_Type_size];
    char attr_path[NAGG_Product_Path_size + 1];
    iet_t begintime = 0;

    if((file = H5Fopen(filename, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0) {
        fprintf(stderr, "Failed to open attribute N_Beginning_Time_IET in file %s.\n", filename);
        goto done;
    }
    
    if((size = 1 + H5Lget_name_by_idx(file, "Data_Products", H5_INDEX_NAME, H5_ITER_INC,
            0, NULL, 0, H5P_DEFAULT)) < 0) {
        fprintf(stderr, "Failed to get size of first group in /Data_Products in file %s.\n", filename);
        goto done;
    }

    if((size = H5Lget_name_by_idx(file, "Data_Products", H5_INDEX_NAME, H5_ITER_INC,
            0, grp_name, size, H5P_DEFAULT)) < 0) {
        fprintf(stderr, "Failed to get name of first group in /Data_Products in file %s.\n", filename);
        goto done;
    }

    snprintf(attr_path, 2 * strlen(grp_name) + 24,  "/Data_Products/%s/%s_Gran_0", grp_name, grp_name);

#ifdef DEBUG
    printf("Name:  %s\n", attr_path);
#endif

    if((attr = H5Aopen_by_name(file, attr_path, "N_Beginning_Time_IET", H5P_DEFAULT, H5P_DEFAULT)) < 0) {
        fprintf(stderr, "Failed to open attribute N_Beginning_Time_IET on dataset %s.\n", attr_path);
        goto done;
    } else if(H5Aread(attr, H5T_NATIVE_ULLONG,
            &begintime) < 0) {
        fprintf(stderr, "Failed to read attribute N_Beginning_Time_IET on dataset %s.\n", attr_path);
    }

#ifdef DEBUG
    printf("nagg_get_beginning_time(): Beginning time: %.3f\n", micros2s(begintime));
#endif

    H5Aclose(attr);

done:

    if(file > -1)
        H5Fclose(file);
        
    return begintime;
}
