/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group.                                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of the HDF NPP Aggregation Tool (nagg). The full        *
 * NPP Aggregation Tool copyright notice, including terms governing use,     *
 * modification, and redistribution, is contained in the file COPYING,       *
 * which can be found at the root of the source code distribution tree.      *
 * If you do not have access to this file, you may request a copy from       *
 * help@hdfgroup.org.                                                        *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "H5public.h"
#include "nagg.h"

#if (H5_VERS_MAJOR>1)||((H5_VERS_MAJOR==1)&&(H5_VERS_MINOR>8))
#define H5Rdereference_(x,y,z) H5Rdereference1(x,y,z)
#else
#define H5Rdereference_(x,y,z) H5Rdereference(x,y,z)
#endif

#define WriteGranules_Product_list_max NAGG_Product_list_max+1
#define NAGG_Dataset_Path_max               160
#define NAGG_Group_Path_max                  80
#define NAGG_Granule_index_string_max         4
#define NAGG_String_attribute_valstring_max 128
#define NAGG_Numeric_attr_valstring_max      32

/* #define DEBUG=1 */

/*------------------------------------------------------------------------
 * Structure:  group_info_t
 *
 * Purpose:    pass both id of the group and its complete path to a functor 
 *             for iterating through the group's datasets.  This is needed
 *             for the layout option so the full path of the dataset can be
 *             matched when it is created to find those that should have 
 *             layout or filter changes applied.
 *
 * Programmer: Larry Knox
 *             October 11, 2013
 *
 *------------------------------------------------------------------------
 */
typedef struct {
    hid_t groupid;
    char path[NAGG_Group_Path_max];
} group_info_t;

typedef group_info_t* group_info_p_t;

#define Mission_name_size    10
#define Instrument_name_size  8
#define Dataset_type_tag_size 7

/*------------------------------------------------------------------------
 * Structure:  product_info_t
 *
 * Purpose:    store product group attribute values to be written to the 
 *             output file userblock.
 *
 * Programmer: Larry Knox
 *             October 20, 2013
 *
 *------------------------------------------------------------------------
 */
typedef struct {
    char         product_name [NAGG_Product_Type_size+1];
    char         instrument_name[Instrument_name_size+1];
    char         dataset_type_tag[Dataset_type_tag_size+1];
    char         processing_domain[Domain_size+1];
    char         beginning_date[NAGG_DATE_size+1];
    uint64_t     beginning_orbit;
    char         beginning_time[NAGG_TIME_size+1];
    char         ending_date[NAGG_DATE_size+1];
    uint64_t     ending_orbit;
    char         ending_time[NAGG_TIME_size+1];
    char         beginning_granule_id[NAGG_Granule_ID_size+1];
    char         ending_granule_id[NAGG_Granule_ID_size+1];
} data_product_info_t; 

typedef data_product_info_t* data_product_info_p_t;

/*------------------------------------------------------------------------
 * Structure:  userblock_info_t
 *
 * Purpose:    store file attribute values to be written to the output file
 *             userblock.
 *
 * Programmer: Larry Knox
 *             October 20, 2013
 *
 *------------------------------------------------------------------------
 */
typedef struct {
    hsize_t      size;
    char         mission_name[Mission_name_size];
    char         platform_name[SPACECRAFT_size];
    unsigned int number_data_products;
    data_product_info_p_t data_product;
} userblock_info_t;

typedef userblock_info_t* userblock_info_p_t;

/*------------------------------------------------------------------------
 * Structure:  product_info_t
 *
 * Purpose:    store IDs for input and output files, output file name, index
 *             of last granule written to ouput file, and pointer to the 
 *             previous granule written.
 *
 * Programmer: Larry Knox
 *             December 15, 2011
 *
 *------------------------------------------------------------------------
 */
typedef struct {
    char dpid[DPID_size+1];         /* dpid and product_name initialized         */
    const char *product_name;       /* at beginning - constant throughout        */
    hid_t infile;                   /* infile and infilename determined granule  */
    const char *infilename;         /* by granule - change with granule->file_in */
    int infile_number_of_granules;  /* may not be needed                         */
    hid_t outfile;                  /* outfile, outfilename, and geofilename     */
    const char *outfilename;        /* should be constant from                   */
    const char *geofilename;        /* from start_write through end_write        */
    int last_i_granule;             /* used to get aggregate ending values       */
    int granules_written;           /* mostly used to trigger initialization     */
                                    /* for new output files                      */
    userblock_info_p_t ublockinfo;
} product_info_t;

typedef product_info_t* product_info_p_t;
static product_info_t g_products[WriteGranules_Product_list_max];
static userblock_info_p_t g_userblocks[NAGG_outputfiles_max + 1]; /* +1 for geo file */ 

/* granules per ouput file and granules per input file should be constant for an aggregation 
when bucket boundaries and fill granules are implemented. */
static int g_gransperoutfile;
static int g_gransperinfile;
static int g_numproducts;

static hbool_t g_input_initialized = FALSE;

static int print_links(hid_t group_id, const char *name, 
                       const H5L_info_t *info, void *op_data);
static int apply_specified_layout(char *dsetpath, hid_t outdtype_id, int ndims,
                       hsize_t *dims, hsize_t *maxdims, hid_t outpl_id, 
                       const char *name, hid_t indset_id, hbool_t *layout_set);
static int create_resized_duplicate_dataset(hid_t ingroup, const char *name,
                       const H5L_info_t *info, void *op_data);
static int get_string_attribute_value(hid_t loc_id, const char * name, char *buf);
static int add_string_attribute(hid_t loc_id, 
                                const char *attr_name, const char *value);
static int update_string_attribute(hid_t loc_id, 
                                   const char *attr_name, const char *value);
static int attrit_op_func(hid_t loc_id, const char *attr_name, 
                          const H5A_info_t *ainfo, void *op_data);
static int update_aggregate_ending_attributes(int product_index);
static int copy_attribute_value(hid_t in_loc_id, const char *attr_name, 
                         const char *attr_name_out, hid_t out_loc_id);
static int write_missing_granules(hid_t output_file, hid_t input_file, 
                           const char *outputfile_name, granule_p_t granule, 
                           int last_i_granule, int i_granule);
static int write_fill_granule_attrs(hid_t loc_id, const char *attr_name, 
                             const H5A_info_t *ainfo, void *op_data);
static int write_fillgr_time_attrs(hid_t dset_id, granule_p_t granule);
static int write_dataset_fill_values(hid_t output_file, const char *dsetpath, 
                              const hsize_t *block_coord, int rank);
static int create_aggregate_dataset(hid_t output_file, hid_t input_file, 
                                    granule_p_t granule);
static int copy_granule_hyperslabs(hid_t output_file, hid_t input_file, 
                                   char *src_dest, char *dest_dset, 
                                   const char * outputfilename, int gran_index);
static int make_aggregate_dataset_path(char *buffer, const char *product_name);
static int make_granule_dataset_path(char *buffer, 
                                     const char *product_name, int granidx);
static int make_All_Data_product_group_path(char *buffer, 
                                            const char *product_name);
static int make_Data_Products_product_group_path(char *buffer, 
                                                 const char *product_name);
static char *iet2utc(iet_t iet_time);
static int add_userblock_element(const char *tagname, const char * value, 
                                 char *userblock_string);

static int store_userblock_ending_attributes(int product_index, hid_t agdset_id);
static int store_userblock_file_attributes(int products_in_file);
static int store_userblock_aggregate_beginning_attributes(int product_index);
static int store_userblock_product_attributes(int product_index);

/*-------------------------------------------------------------------------
 * Function:   apply_specified_filters
 *
 * Purpose:    
 *
 * Parameters: dsetpath:     full path to dataset to be created
 *             outdtype_id:  id for the dataset type       (these three 
 *             ndims:        number of dataset dimensions   used to
 *             dims:         array of dimension values      calculate size) 
 *             maxdims:      array of maximum dimensions (must be reset when
 *                           changing CHUNKED to COMPACT or CONTIGUOUS)
 *             outpl_id:     id for output dataset property list
 *             name:         dataset name (for error messages) 
 *             indset_id:    id for input dataset
 *             *layout_set:  bool variable to pass whether layout was set or not -
 *                           if not set here, chunk size may be reset by nagg as 
 *                           deemed appropriate for input and output dataset size
 *
 * Note:       This function will apply any global filter to the dataset or will
 *             check the list and apply any filter specific to the dataset.  The 
 *             function succeeds as long as apply_filters() does not return a 
 *             negative value.  In other words the function succeeds if a filter
 *             is successfully applied or if there is no filter specified for the 
 *             dataset.
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             November 6, 2013
 *
 *-------------------------------------------------------------------------
 */
static int
apply_specified_filters(char *dsetpath, hid_t outdtype_id, int ndims,
                        hsize_t *dims, hsize_t *maxdims, hid_t outpl_id, 
                        const char *name, hid_t indset_id)
{
    pack_info_t* dset_tbl_obj;
    int status = 0;
    int i;
    int has_filter = 0;

    /* if we have global filters, try to apply them. */
    if (filay_options->n_filter_g > 0) { 
        for (i=0; i<filay_options->n_filter_g; ++i) {
            if((status = apply_filters(dsetpath, ndims, dims, 
                                       H5Tget_size(outdtype_id), outpl_id,
                                       filay_options, &has_filter)) < 0) {
                /* Probably the apply_filters function will have printed an error. */
                NAGG_ERROR("apply_specified_filters():  call to apply_filters() failed.\n");
                goto done;
            }
        }
    } else {
    /* otherwise check the table of objects for this dataset, and for filters to be applied. */
        dset_tbl_obj = options_get_object(dsetpath, filay_options->op_tbl);
        if (dset_tbl_obj != NULL) {
            for (i=0; i<dset_tbl_obj->nfilters; ++i) {
                if((status = apply_filters(dsetpath, ndims, dims, 
                                           H5Tget_size(outdtype_id), outpl_id, 
                                           filay_options, &has_filter)) < 0) {
                    /* Probably the set_filter function will have printed an error. */
                    goto done;
                }
            }
        }

    }
done:
    return status;
}

/*-------------------------------------------------------------------------
 * Function:   apply_specified_layout
 *
 * Purpose:    Apply layout changes specified by command line options to 
 *             property lists for output file datasets that will be created
 *             by the create_resized_duplicate_dataset callback function. 
 *
 * Parameters: dsetpath:     full path to dataset to be created
 *             outdtype_id:  id for the dataset type       (these three 
 *             ndims:        number of dataset dimensions   used to
 *             dims:         array of dimension values      calculate size) 
 *             maxdims:      array of maximum dimensions (must be reset when
 *                           changing CHUNKED to COMPACT or CONTIGUOUS)
 *             outpl_id:     id for output dataset property list
 *             name:         dataset name (for error messages) 
 *             indset_id:    id for input dataset
 *             *layout_set:  bool variable to pass whether layout was set or not -
 *                           if not set here, chunk size may be reset by nagg as 
 *                           deemed appropriate for input and output dataset size
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             October 29, 2013
 *
 *-------------------------------------------------------------------------
 */
static int
apply_specified_layout(char *dsetpath, hid_t outdtype_id, int ndims,
                       hsize_t *dims, hsize_t *maxdims, hid_t outpl_id, 
                       const char *name, hid_t indset_id, hbool_t *layout_set)
{
    pack_info_t* dset_tbl_obj;
    int status = 0;
    hsize_t dset_size, new_dset_size;
    int i, j;
    H5D_layout_t  layout_to_apply = H5D_LAYOUT_ERROR;
    *layout_set = TRUE;

    /* if it applies to all datasets, set the layout type. */
    if (filay_options->all_layout == 1)
        layout_to_apply = filay_options->layout_g;
    else if (filay_options->op_tbl->nelems > 0) {
        /* check for dataset in layout table */
        dset_tbl_obj = options_get_object(dsetpath, filay_options->op_tbl);
        if (dset_tbl_obj != NULL)
            layout_to_apply = dset_tbl_obj->layout;
    }

    /* If user specified a layout in the nagg command for all datasets 
     *  or for this specific dataset, set the property list accordingly. 
     */
    switch (layout_to_apply) {
    case H5D_COMPACT:
        /* Total size for a COMPACT dataset must be < 65536 bytes. Calculation 
         * of size is the size of the type times each dimension.  The dataset 
         * dimensions reflect the number of granules in the aggregation.
         */ 
        dset_size = H5Tget_size(outdtype_id);
        for (j=0; j<ndims; j++) {
            dset_size = dset_size * dims[j];
        }
        if (dset_size < 65536) {
            /* Compact datasets cannot have unlimited dimensions or filters.  
               Set all maxdims[i] = dims[i] and remove all compression filters.*/
            for (i=0; i<ndims; ++i)
                maxdims[i] = dims[i];
            if ((status = H5Premove_filter(outpl_id,H5Z_FILTER_ALL)) < 0) {
                NAGG_ERROR("apply_specified_layout(): Failed to remove compression filters for ");
                fprintf(stderr, "applying COMPACT layout to dataset %s\n", name);
                H5Dclose(indset_id);
                goto done;
            }

            if ((status = H5Pset_layout(outpl_id, H5D_COMPACT)) < 0) {
                NAGG_ERROR("apply_specified_layout(): Failed to set layout to COMPACT for dataset ");
                fprintf(stderr, "%s\n", name);
                H5Dclose(indset_id);
                goto done;
            }
        } else {
            /* If we don't return an error here, users can specify compact layout, and it will be applied to all datasets that aren't 64K or larger.  Larger datasets will have layout characteristics either as otherwise specified or else as in the input dataset.  If we returned an error, users would have to predetermine and specify them individually. */
#ifdef DEBUG
printf("apply_specified_layout(): dataset %s is too large for COMPACT layout in the output file.  COMPACT layout cannot be applied to datasets with size 65535 bytes or larger.\n", name);
#endif
            *layout_set = FALSE;
        }
        break;
    case H5D_CONTIGUOUS:
        /* Contiguous datasets cannot have unlimited dimensions or filters.  
           Set all maxdims[i] = dims[i] and remove all compression filters.*/
        for (i=0; i<ndims; ++i)
            maxdims[i] = dims[i];
        if ((status = H5Premove_filter(outpl_id,H5Z_FILTER_ALL)) < 0) {
            NAGG_ERROR("apply_specified_layout(): Failed to remove compression filters for ");
            fprintf(stderr, "applying CONTIGUOUS layout to dataset %s\n", name);
            H5Dclose(indset_id);
            goto done;
        }
        if ((status = H5Pset_layout(outpl_id, H5D_CONTIGUOUS)) < 0) {
            NAGG_ERROR("apply_specified_layout(): Failed to set layout to CONTIGUOUS for dataset ");
            fprintf(stderr, "%s\n", name);
            H5Dclose(indset_id);
            goto done;
        }
        break;
    case H5D_CHUNKED:
        if (dset_tbl_obj != NULL && filay_options->op_tbl->nelems > 0) {
            /* specific layout applies to this dataset */
            if(dset_tbl_obj->chunk.rank == ndims) {
                if((status = H5Pset_chunk(outpl_id, dset_tbl_obj->chunk.rank, dset_tbl_obj->chunk.chunk_lengths)) < 0) {
                     NAGG_ERROR("apply_specified_layout(): Failed to set CHUNKED layout for dataset ");
                     fprintf(stderr, "%s\n", name);
                     H5Dclose(indset_id);
                     goto done;
                } 
            } else {
                NAGG_ERROR("apply_specified_layout(): incorrect rank specified to set CHUNKED layout for dataset ");
                fprintf(stderr, "%s\n", name);
                status = FAIL;
                H5Dclose(indset_id);
                goto done;
            }
        }else {
            /* global chunk layout can be applied to this dataset 
               but only if the layout rank matches the dataset rank */
            if(filay_options->chunk_g.rank == ndims){
                if((status = H5Pset_chunk(outpl_id, filay_options->chunk_g.rank, 
                                          filay_options->chunk_g.chunk_lengths)) < 0) {
                     NAGG_ERROR("apply_specified_layout(): Failed to set CHUNKED layout for dataset ");
                     fprintf(stderr, "%s\n", name);
                     H5Dclose(indset_id);
                     goto done;
                }
            } else {
                *layout_set = FALSE;
            }
        }
        break;
    case H5D_LAYOUT_ERROR:
        /* no layout for this dataset */
        *layout_set = FALSE;
        break;
    case H5D_NLAYOUTS:
    default:
        NAGG_ERROR("check_options(): invalid layout\n");
        status = FAIL;
        *layout_set = FALSE;
        break;
    }
done:
    return status;
}

/*-------------------------------------------------------------------------
 * Function:   create_resized_duplicate_dataset
 *
 * Purpose:    Create a dataset in the output file with similar characteristics 
 *             to a dataset in the input file sized for the new aggregation 
 *             number. 
 *
 * Parameters: ingroup: id of the group containing the input dataset
 *             name:    dataset name
 *             info:    H5L_info_t struct required for callback functions used 
 *                      by H5Literate 
 *             op_data  address of variable containing id of the group containing
 *                      the ouput dataset 
 *
 * Notes:      Creates a dataset without any data, which will be added by calls 
 *             to the function copy_granule_hyperslabs.
 *             Whether ouput datasets are fixed or unlimited will be determined
 i*             by the input dataset characteristics, unless layout or filter 
 *             options indicate changes.
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             April 27, 2012
 *             Modified October 29, 2013
 *
 *-------------------------------------------------------------------------
 */
static int
create_resized_duplicate_dataset(hid_t ingroup, const char *name,
                       const H5L_info_t *info, void *op_data)
{
    int status;
    hid_t indset_id, indspace_id, indtype_id, inpl_id;
    hid_t outdset_id = -1; 
    hid_t outdspace_id, outdtype_id, outpl_id;
    H5O_info_t infobuf;
    int i, ndims;
    hsize_t origdim;
    hsize_t *dims, *maxdims;
    hsize_t newdset_size;
    char dsetpath[NAGG_Dataset_Path_max];
    hbool_t layout_set = FALSE;
    
    HDstrcpy(dsetpath, (*(group_info_p_t)op_data).path);
    HDstrcat(dsetpath, "/");
    HDstrcat(dsetpath, name);

    if ((status = H5Oget_info_by_name (ingroup, name, &infobuf, H5P_DEFAULT)) < 0) {
        fprintf(stderr, "create_resized_duplicate_dataset(): Failed to get info for product group.\n");
        goto done;
    }

    if (infobuf.type == H5O_TYPE_DATASET) {
        if((indset_id = H5Dopen(ingroup, name, H5P_DEFAULT))<0) {
            fprintf(stderr, "create_resized_duplicate_dataset(): Failed to open dataset %s.\n", name);
                status = FAIL;
                H5Gclose(ingroup);
                goto done;
        }
        if ((indspace_id = H5Dget_space(indset_id)) < 0) {
            fprintf(stderr, "create_resized_duplicate_dataset(): Failed to open dataspace for dataset  %s.\n", name);
                status = FAIL;
                H5Dclose(indset_id);
                goto done;
        }
        if ((indtype_id = H5Dget_type(indset_id)) < 0) {
            fprintf(stderr, "create_resized_duplicate_dataset(): Failed to open datatype for dataset %s.\n", name);
                status = FAIL;
                H5Dclose(indset_id);
                goto done;
        }
        if ((inpl_id = H5Dget_create_plist(indset_id)) < 0) {
            fprintf(stderr, "create_resized_duplicate_dataset(): Failed to open dataset property list for dataset %s.\n", name);
                status = FAIL;
                H5Dclose(indset_id);
                goto done;
        } 

        if ((ndims = H5Sget_simple_extent_ndims(indspace_id)) < 0){
            fprintf(stderr, "Failed to get number of dataset dimensions.\n");
                    status = FAIL;
                    goto done;
        }
        if ((outdtype_id = H5Tcopy(indtype_id)) < 0) {
            fprintf(stderr, "create_resized_duplicate_dataset(): Failed to copy datatype for dataset %s.\n", name);
            status = FAIL;
            H5Dclose(indset_id);
            goto done;
        }
        if ((outpl_id = H5Pcopy(inpl_id)) < 0) {
            fprintf(stderr, "create_resized_duplicate_dataset(): Failed to copy property list for dataset %s.\n", name);
            status = FAIL;
            H5Dclose(indset_id);
            goto done;
        }
    
        /* The first dimension is resized for the size of the new aggregation. 
           The others will keep the same size. */

        dims = (hsize_t *) HDmalloc (ndims * sizeof(hsize_t));
        maxdims = (hsize_t *) HDmalloc (ndims * sizeof(hsize_t));
        if ((H5Sget_simple_extent_dims(indspace_id, dims, maxdims)) < 0 ) {
            fprintf(stderr, "Failed to get dataset dimensions.\n");
            status = FAIL;
            goto done;
        }

        origdim = dims[0];

       /* g_gransperinfile and g_gransperoutfile are the old and new aggregation
          numbers (from the AggregateNumberGranules attribute). */
       /* origdim should always be a multiple of g_gransperinfile, and as a result
        * (origdim / g_gransperinfile) will be the correct dim[0] of the hyperslab
        * for the granule.  Multiplying this times the number of granules in the 
        * output file gives the dim[0] for the dataset in the output file.
        */
        dims[0] = ( origdim / g_gransperinfile ) * g_gransperoutfile;

       /* Could have used H5Dget_storage_size(indset_id), but if compression is 
        * changed, the size in the ouput file may be different.  
        * datatype_size * dimensions should be consistent.
        */
        newdset_size = H5Tget_size(outdtype_id);
        for (i=0; i<ndims; i++) {
            newdset_size = newdset_size * dims[i];
        }
      
        if (maxdims[0] != H5S_UNLIMITED)
             maxdims[0] = dims[0];

        /*
         * Always apply layout for COMPACT datasets since HDF5 has a 64K 
         * maximum size for COMPACT datasets. If there is another layout
         * specified, call apply_specified_layout() if the layout is not
         * global regardless of size or if the layout is global and the 
         * dataset size exceeds the minimum.
         */ 
        if (filay_options->layout_g == H5D_COMPACT 
            || (filay_options->layout_g != H5D_LAYOUT_ERROR
                && (filay_options->all_layout == 0
                    || (filay_options->all_layout == 1
                        && newdset_size >= filay_options->min_comp)))
          
                ) {
         
            if((status = apply_specified_layout(dsetpath, outdtype_id, ndims,
                    dims, maxdims, outpl_id, name, indset_id, &layout_set)) < 0) {
                NAGG_ERROR("create_resized_duplicate_dataset():  Applying layout failed for ");
                fprintf(stderr, "%s\n", name);
                goto done;
            }
        }

        /* no layout options were specified on the command line to apply to this dataset,
         * or compact layout was specified but the dataset size is > 65535.
         * Modify the chunk size to reflect the change in the aggregation number.  Don't
         * increase the size if the new size will be above 64K.  If the new aggregation
         * number is smaller than the input aggregation number, reduce the chunk size only
         * when it is larger than the dataset size. 
         */
        if (!layout_set && g_gransperinfile != g_gransperoutfile) {
#define CHUNK_RESIZE_THRESHOLD 65536
            int chnk_size = 1, new_chnk_size, j;
            hsize_t* chnk_dims = (hsize_t *) HDmalloc (ndims * sizeof(hsize_t));

            if ((H5Pget_chunk(inpl_id, ndims, chnk_dims)) < 0) {
                fprintf(stderr, "create_resized_duplicate_dataset(): Failed to get chunk size for dataset %s.\n", name);
                status = FAIL;
                H5Dclose(indset_id);
                goto done;
            }


            for (j=0; j<ndims; j++) {
                chnk_size = chnk_size * chnk_dims[j];
            }
            /* new_chnk_size = chnk_size / g_gransperinfile * g_gransperoutfile; */
            new_chnk_size = ( g_gransperoutfile / g_gransperinfile ) * chnk_size;


            /* Dataset size is increasing */
            if (g_gransperoutfile > g_gransperinfile) {
                if (new_chnk_size < CHUNK_RESIZE_THRESHOLD) {
                    /*chnk_dims[0] = chnk_dims[0] / g_gransperinfile * g_gransperoutfile;*/
                    chnk_dims[0] = ( g_gransperoutfile / g_gransperinfile ) * chnk_dims[0];
                    if ((status = H5Pset_chunk(outpl_id, ndims, chnk_dims)) < 0) {
                        NAGG_ERROR("create_resized_duplicate_dataset():  Failed to set chunk for ");
                        fprintf(stderr, "%s.\n", name);
                        goto done;
                    }
                }
            /* Dataset size is decreasing - chnk_dims[0] should not be larger than the
               dataset's dims[0].   */
            } else if (chnk_dims[0] > dims[0] ){
                chnk_dims[0] =  dims[0];
                H5Pset_chunk(outpl_id, ndims, chnk_dims);
            }
    
            if (chnk_dims)
                HDfree(chnk_dims);
        } /* end chunk resizing. */
       
        if((outdspace_id = H5Screate_simple(ndims, dims, maxdims)) < 0) {
            fprintf(stderr, "create_resized_duplicate_dataset(): Failed to create dataspace for dataset %s.\n", name);
            status = FAIL;
            H5Dclose(indset_id);
            goto done;
        }

        /* --minimum will apply for layout or filter to be applied to all datasets.
         * If specific datasets are specified for a filter or layout, the minimum will
         * not be considered.
         */
        if ((filay_options->n_filter_g > 0
                 && newdset_size >= filay_options->min_comp) 
             || (filay_options->op_tbl->nelems > 0)) {
            if((status = apply_specified_filters(dsetpath, outdtype_id, ndims,
                             dims, maxdims, outpl_id, name, indset_id)) < 0) {
                NAGG_ERROR("create_resized_duplicate_dataset():  Filter was not applied for ");
                fprintf(stderr, "%s\n", name);
                goto done;
            }
        }


        /* op_data should contain the group id where the dataset will be created. */
        if ((outdset_id = H5Dcreate2((*(group_info_p_t)op_data).groupid, name, outdtype_id,
                          outdspace_id, H5P_DEFAULT, outpl_id, H5P_DEFAULT)) < 0) {
            fprintf(stderr, "create_resized_duplicate_dataset(): Failed to create dataset %s.\n", name);
            status = FAIL;
            H5Dclose(indset_id);
            goto done;
        }
        if ((H5Aiterate2(indset_id, H5_INDEX_NAME, H5_ITER_NATIVE, 0, attrit_op_func, &outdset_id)) < 0) {
            fprintf(stderr, "create_resized_duplicate_dataset(): Failed to copy attributes for dataset %s.\n", name);
            status = FAIL;
            H5Dclose(indset_id);
            goto done;
        }

    } else {
        fprintf(stderr, "create_resized_duplicate_dataset(): Unexpected object type for object %s.\n", name);
    }
done:
    if (H5Iis_valid(indset_id))
        H5Dclose(indset_id);
    if (H5Iis_valid(outpl_id))
        H5Pclose(outpl_id);
    if (H5Iis_valid(outdset_id))
        H5Dclose(outdset_id);
    if (dims)
        HDfree(dims);
    if (maxdims)
        HDfree(maxdims);
    return status;

}

/*-------------------------------------------------------------------------
 * Function:   get_product_idx_by_id
 *
 * Purpose:    
 *
 * Note:       
 *
 * Return:     Success:     the index of a product in the array of products.
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             December 15, 2011
 *
 *-------------------------------------------------------------------------
 */
static int get_product_idx_by_id(const char *product_id)
{
    int i;
    for (i=0; i<g_numproducts ; ++i) {
        if (HDstreq(product_id, g_products[i].dpid)) {
            return i;
        }
    }
    /* 0 based index */
    return g_numproducts-1;
}

/*-------------------------------------------------------------------------
 * Function:   get_product_id_by_idx
 *
 * Purpose:
 *
 * Note:
 *
 * Return:     the product DPID matching the index.
 *
 * Programmer: Larry Knox
 *             March 31, 2012
 *
 *-------------------------------------------------------------------------
 */
static const char *get_product_id_by_idx(int index)
{
    if (g_products[index].dpid)
        return g_products[index].dpid;
    else
        return "Internal error:  product name unknown.\n";
}

/*-------------------------------------------------------------------------
 * Function:   get_product_name_by_idx
 *
 * Purpose:    
 *
 * Note:       
 *
 * Return:     the product name matching the index.
 *
 * Programmer: Larry Knox
 *             December 15, 2011
 *
 *-------------------------------------------------------------------------
 */
static const char *get_product_name_by_idx(int index) 
{
    if (g_products[index].product_name)
        return g_products[index].product_name;
    else 
        return "Internal error:  product name unknown.\n";
}

/*-------------------------------------------------------------------------
 * Function:  get_product_granules_written
 *
 * Purpose:    
 *
 * Note:       
 *
 * Return:     the number of granules written for a product.
 *
 * Programmer: Larry Knox
 *             December 15, 2011
 *
 *-------------------------------------------------------------------------
 */
static int get_product_granules_written(int index) 
{
    return g_products[index].granules_written;
}

/*-------------------------------------------------------------------------
 * Function:   increment_granules_written
 *
 * Purpose:    
 *
 * Note:       
 *
 * Return:     the incremented index of the last granule written.
 *
 * Programmer: Larry Knox
 *             December 15, 2011
 *
 *-------------------------------------------------------------------------
 */
static int increment_granules_written(int index)
{
    return g_products[index].granules_written++;
}

/*-------------------------------------------------------------------------
 * Function:   reset_product_granules_written
 *
 * Purpose:    
 *
 * Note:       
 *
 * Return:     the incremented index of the last granule written.
 *
 * Programmer: Larry Knox
 *             December 15, 2011
 *
 *-------------------------------------------------------------------------
 */
static int reset_product_granules_written(int index)
{
    g_products[index].granules_written = 0;
    return g_products[index].granules_written;
}

/*-------------------------------------------------------------------------
 * Function:   get_last_i_granule
 *
 * Purpose:    
 *
 * Note:       
 *
 * Return:     the index of the last granule written.
 *
 * Programmer: Larry Knox
 *             December 15, 2011
 *
 *-------------------------------------------------------------------------
 */
static int get_last_i_granule(int product_index) 
{
    return g_products[product_index].last_i_granule;
}

/*-------------------------------------------------------------------------
 * Function:   set_last_i_granule
 *
 * Purpose:
 *
 * Note:
 *
 * Return:     void
 *
 * Programmer: Larry Knox
 *             February 10, 2012
 *
 *-------------------------------------------------------------------------
 */
static void
set_last_i_granule(int product_index, int i_granule)
{
    g_products[product_index].last_i_granule = i_granule;
}

/*-------------------------------------------------------------------------
 * Function:  reset_input_file
 *
 * Purpose:   Close one input file and open another
 *
 * Note:      Originally intended for aggregating granules where the numbers
 *            of granules in the input files don't match those in the output
 *            files.  It was decided to keep them open.  Function left here 
 *            for future reference if there are too many open files.
 *
 * Return:    Success:     0
 *            Failure:     -1
 *
 * Programmer: Larry Knox
 *             February 10, 2012
 *
 *-------------------------------------------------------------------------
 */
/*
static int
reset_input_file(int product_index, granule_p_t granule)
{
    int     status;
    if (products[product_index]->infile)
        H5Fclose(products[product_index]->infile);


    g_products[product_index]->inwritsofar = 0;
    if ((products[product_index]->infile = H5Fopen(granule->file_in, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0) {
        fprintf(stderr, "Failed to open file %s.\n", granule->file_in);
        status = FAIL;
    }
    return status;
}
*/

/*-------------------------------------------------------------------------
 * Function:   first_granule_file_init
 *
 * Purpose:    Initialize empty output file when first granule is written.
 *
 * Note:       Datasets under the /All_Data group are copied without data from 
 *             the input file to the empty output file, and their storage size 
 *             set for the specified number of granules.  Attributes belonging 
 *             to groups or to the ..._Aggr dataset are copied from the input 
 *             file to the new output file, except for attributes previously 
 *             written in the start_write function when the file was created.
 *             When the aggregation size differs from that of the input files, 
 *             beginning and ending time attributes for the output file will be 
 *             different from those for the input file.
 *
 * Parameters: output_file: handle to the empty output file.
 *             granule_p_t: pointer to an entry in the granule table
 *             i_granule:   index that the granule should have in the new output 
 *             file.
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             December 15, 2011
 *
 *-------------------------------------------------------------------------
 */
static int
first_granule_file_init( hid_t output_file, hid_t input_file, 
                        granule_p_t granule, int i_granule)
{
    int     status;
    hid_t   gcpl, ingroup, adingroup, dpingroup;
    hid_t   outgroup;
    hid_t   dset_id = -1;
    hid_t   attr;
    H5L_info_t *info;
    char    dest_dset[NAGG_Dataset_Path_max];
    char    product_name[NAGG_Product_Type_size];
    int     product_index;
    group_info_t* adoutgrp;
    group_info_t* dpoutgrp;
    char    num_products_str[NAGG_Numeric_attr_valstring_max];
    char    attr_val[NAGG_String_attribute_valstring_max];
    int     num_file_products = g_numproducts;

    adoutgrp = (group_info_t*)HDmalloc(sizeof(group_info_t));
    dpoutgrp = (group_info_t*)HDmalloc(sizeof(group_info_t));

#ifdef DEBUG 
    printf("Initialize file with granule %d for product %s.\n", i_granule, granule->product_id);
#endif    
    product_index = get_product_idx_by_id(granule->product_id);
    reset_product_granules_written(product_index);

    /* Create /Data_Products/<product_name>/<product_name_Aggr dataset.  This 
       should be done after H5LRcopy_reference is called for the first granule.
       The datasets for the references don't exist before that, and attributes
       on the ..._Aggr dataset from the first granule are needed.  Aggregations
       that begin with fill granules will need to wait for the first actual
       granule and extra modification of some attributes will be required.  */

    HDstrcpy(product_name, granule->product_name);
    if (status = make_All_Data_product_group_path(adoutgrp->path, granule->product_name) < 0) {
        NAGG_ERROR("first_granule_file_init(): Failed to make /All_Data...product path.\n");
        goto done;
    }
    if (status = make_Data_Products_product_group_path(dpoutgrp->path, granule->product_name) < 0) {
        NAGG_ERROR("first_granule_file_init(): Failed to make /Data_product...product path.\n");
        goto done;
    }
    if (status = make_aggregate_dataset_path(dest_dset, granule->product_name) < 0) {
        NAGG_ERROR("first_granule_file_init(): Failed to make aggregate dataset path.\n");
        goto done;
    }

    if((dset_id = H5Dopen(input_file, dest_dset, H5P_DEFAULT)) < 0 ) {
        fprintf(stderr, "first_granule_file_init(): Failed to open dataset %s in %s.\n", dest_dset, granule->file_in);
        status = FAIL;
        goto done;
    }
    if ((attr = H5Aopen(dset_id, "AggregateNumberGranules", H5P_DEFAULT)) < 0)  {
        fprintf(stderr, "Failed to open attribute AggregateNumberGranules.\n");
        status = FAIL;
        goto done;
    }
    /* this should be the same for all products.  Currently it's looked up for each. */
    if ((status = H5Aread(attr, H5T_NATIVE_INT, &g_gransperinfile)) < 0) {
        fprintf(stderr, "Failed to read attribute AggregateNumberGranules.\n");
        status = FAIL;
        goto done;
    }
    if (dset_id > -1)
        H5Dclose(dset_id);


    /* Copy the root group attributes. */
    if ((ingroup = H5Gopen(input_file, "/", H5P_DEFAULT)) < 0) {
        fprintf(stderr, "Failed to open input file \"/\"group.\n");
        status = FAIL;
        goto done;
    }
    if ((outgroup = H5Gopen(output_file, "/", H5P_DEFAULT)) < 0) {
        fprintf(stderr, "Failed to open output file \"/\"group.\n");
        status = FAIL;
        goto done;
    }    
    if ((status = H5Aiterate2(ingroup, H5_INDEX_NAME, H5_ITER_NATIVE, 0, 
                              attrit_op_func, &outgroup)) != 0) {
        NAGG_ERROR("first_granule_file_init(): Failed to copy all attributes of \"/\"group to output file.");
        goto done;
    }
    if ((status = get_string_attribute_value(ingroup, "Mission_Name", 
                    g_products[product_index].ublockinfo->mission_name)) < 0) {
        fprintf(stderr, "Failed to get value of /Mission_Name attribute.\n");
        goto done;
    }

    if ((gcpl = H5Pcreate (H5P_LINK_CREATE)) < 0) {
        fprintf(stderr, "Failed to create propery list for output group %s.\n", adoutgrp->path);
        
        status = FAIL;
        goto done;
    }
    if ((status = H5Pset_create_intermediate_group (gcpl, 1)) < 0) {
        fprintf(stderr, "Failed to set intermediate group creation property for output group %s.\n", adoutgrp->path);
        status = FAIL;
        goto done;
    }
    if ((adoutgrp->groupid = H5Gcreate (output_file, adoutgrp->path, gcpl, H5P_DEFAULT, H5P_DEFAULT)) < 0) {  
        fprintf(stderr, "Failed to create output group %s.\n", adoutgrp->path);
        status = FAIL;
        goto done;
    }
    if ((adingroup = H5Gopen(input_file, adoutgrp->path, H5P_DEFAULT)) < 0) {
        fprintf(stderr, "Failed to open input group %s.\n", adoutgrp->path);
        status = FAIL;
        goto done;
    }

    /* Call a function to copy the datasets from /All_Data in
       the file with the first granule to /All_Data in the new file.  The new
       datasets will have no data, which will be added by the write_granules 
       function. */
    if((status = H5Literate(adingroup, H5_INDEX_NAME, H5_ITER_NATIVE, 0, 
                        create_resized_duplicate_dataset, adoutgrp)) < 0) {
        NAGG_ERROR("first_granule_file_init(): Failed creating resized datasets.\n");
        H5Gclose (adingroup);
        goto done;
    }
    if (status < 0) {
        goto done;
    }


    gcpl = H5Pcreate (H5P_LINK_CREATE);
    if ((status = H5Pset_create_intermediate_group (gcpl, 1)) < 0) {
        NAGG_ERROR("first_granule_file_init(): Failed to set create_intermediate_group property.\n");
        goto done;
    }
    if ((dpoutgrp->groupid = H5Gcreate (output_file, dpoutgrp->path, gcpl, H5P_DEFAULT, H5P_DEFAULT)) < 0) {
        NAGG_ERROR("first_granule_file_init(): Failed to create ");
        fprintf(stderr,"%s group in output file.\n", dpoutgrp->path);
        goto done;
    }
 
    
    /* Copy the /Data_Products/<product_name> group attributes */
    dpingroup = H5Gopen(input_file, dpoutgrp->path, H5P_DEFAULT);
    if ((status =  H5Aiterate2(dpingroup, H5_INDEX_NAME, H5_ITER_NATIVE, 0, attrit_op_func, &(dpoutgrp->groupid))) != 0) {
        NAGG_ERROR("first_granule_file_init(): Failed to copy all attributes of ");
        fprintf(stderr,"%s to output file.");
        goto done;
    }

    if ((status = create_aggregate_dataset(output_file, input_file, granule)) < 0) {
        fprintf(stderr, "Failed to create output aggregate dataset.\n");
    }

done:
    if (H5Iis_valid(attr))
    H5Aclose(attr);
    if (H5Iis_valid(gcpl))
    H5Pclose (gcpl);
    if (H5Iis_valid(ingroup))
    H5Gclose (ingroup);
    if (H5Iis_valid(outgroup))
    H5Gclose (outgroup);
    if (H5Iis_valid(dpoutgrp->groupid))
    H5Gclose (dpoutgrp->groupid); 
    if (H5Iis_valid(dpingroup))
    H5Gclose (dpingroup); 
    if (H5Iis_valid(adingroup))
    H5Gclose (adingroup); 
    if (H5Iis_valid(adoutgrp->groupid))
    H5Gclose (adoutgrp->groupid); 
    if(adoutgrp)
        HDfree(adoutgrp);
    if(dpoutgrp)
        HDfree(dpoutgrp);

    return status;
}

/*-------------------------------------------------------------------------
 * Function:   make_granule_dataset_path
 *
 * Purpose:    assemble path to a granule dataset from product name and index
 *
 * Note:
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             February 9, 2012
 *
 *-------------------------------------------------------------------------
 */
static int
make_granule_dataset_path(char *buffer, const char *product_name, int granidx)
{
    int status = SUCCEED;
    char index_str[NAGG_Granule_index_string_max];

    if (granidx > 999) {
        NAGG_ERROR("make granule_dataset_path():  Granule index too large.\n");
        status= FAIL;
        goto done;
    }

    sprintf(index_str,"%d", granidx);

    snprintf(buffer, 16, "/Data_Products/");
    strncat(buffer, product_name,  strlen(product_name));
    strncat(buffer, "/", 1);
    strncat(buffer, product_name,  strlen(product_name));
    strncat(buffer, "_Gran_", 6);
    strncat(buffer, index_str, strlen(index_str));
done:
    return status;
}

/*-------------------------------------------------------------------------
 * Function:   make_aggregate_dataset_path
 *
 * Purpose:    assemble path to an aggregate dataset from product name
 *
 * Note:
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             February 9, 2012
 *
 *-------------------------------------------------------------------------
 */
static int
make_aggregate_dataset_path(char *buffer, const char *product_name)
{
    int status = 0;

    status = snprintf(buffer, 16, "/Data_Products/");
    HDstrncat(buffer, product_name,  HDstrlen(product_name));
    HDstrncat(buffer, "/", 1);
    HDstrncat(buffer, product_name,  HDstrlen(product_name));
    HDstrncat(buffer, "_Aggr", 5);

    return status;
}

/*-------------------------------------------------------------------------
 * Function:   make_All_Data_product_group_path
 *
 * Purpose:    assemble path to an aggregate dataset from product name
 *
 * Note:
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             February 9, 2012
 *
 *-------------------------------------------------------------------------
 */
static int
make_All_Data_product_group_path(char *buffer, const char *product_name)
{
    int status = 0;
    status = snprintf(buffer, 11, "/All_Data/");
    HDstrncat(buffer, product_name,  HDstrlen(product_name));
    HDstrncat(buffer, "_All", 4);

    return status;
}

/*-------------------------------------------------------------------------
 * Function:   make_Data_Products_product_group_path
 *
 * Purpose:    assemble path to an aggregate dataset from product name
 *
 * Note:
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             February 9, 2012
 *
 *-------------------------------------------------------------------------
 */
static int
make_Data_Products_product_group_path(char *buffer, const char *product_name)
{
    int status = 0;
    status = snprintf(buffer, 16, "/Data_Products/");
    HDstrncat(buffer, product_name,  HDstrlen(product_name));

    return status;
}

/*-------------------------------------------------------------------------
 * Function:   write_missing_granules 
 *
 * Purpose:    write a granule with fill values for data when a granule is 
 *             missing
 *
 * Note:       
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             December 15, 2011
 *
 *-------------------------------------------------------------------------
 */
static int
write_missing_granules(hid_t output_file, hid_t input_file, 
                       const char *outputfile_name, granule_p_t granule, 
                       int last_i_granule, int i_granule)
{
    int i, j, k;
    int product_index;
    hdset_reg_ref_t *ref = NULL;
    hdset_reg_ref_t *ref_new = NULL;
    char *reg_name;
    char src_group[NAGG_Group_Path_max];
    char dest_group[NAGG_Group_Path_max];
    char dest_dset[NAGG_Dataset_Path_max];
    char src_dset[NAGG_Dataset_Path_max];
    hsize_t *region_buf;
    size_t rnlength;
    int rank_out;
    size_t numelem;
    size_t numrefs;
    H5S_sel_type sel_type;
    size_t size_ref;
    hid_t dtype, dcpl;
    hid_t group_id;
    hid_t dset_id = -1; 
    hid_t ref_dset_id = -1;
    hid_t dspace, outdspace, attr;
    hid_t fapl_id = -1;
    hsize_t delta;
    htri_t exists;
    int array_size;
    hsize_t *start;
    hsize_t chunk[1] = {32};
    char **dataset_paths;
    int status = SUCCEED;
        ssize_t ocount = 0, numchars;
        hid_t open_objs[3] = {-1, -1, -1};
        char oname[256];
    granule_pattern_p_t gran_pattern;
    /* last_i_granule was previously used to calculate the hyperslab position,
       but now we are writing only the granule sent, so i_granule is more 
       appropriate.  The parameter last_i_granule (not the product structure's
       member) can possibly be eliminated or used for error checking. */
    last_i_granule++;  
    product_index = get_product_idx_by_id(granule->product_id);
    gran_pattern = get_granule_pattern(get_product_id_by_idx(product_index));

    if (NULL == gran_pattern) {
        NAGG_ERROR("write_missing_granules():  Failed to write missing granule.\n");
        fprintf(stderr, "No granule pattern found for product id %s.\n", granule->product_id);  
        status = FAIL;
        goto done;
    }
#ifdef DEBUG
        printf("Write fill granule for ..._Gran_%d for product %s\n", i_granule, granule->product_name);
#endif
        if(g_products[product_index].infilename != gran_pattern->file_in) {
                /* Need to close the open file and open a new one. */
            if (g_products[product_index].infilename != NULL
                && H5Iis_valid(g_products[product_index].infile)) {
#ifdef DEBUG
            ocount = H5Fget_obj_count(g_products[product_index].infile, H5F_OBJ_DATASET);
            printf("There are %d open datasets before the close.\n", (int)ocount);
            ocount = H5Fget_obj_count(g_products[product_index].infile, H5F_OBJ_GROUP);
            printf("There are %d open groups before the close.\n", (int)ocount);
            if(H5Fget_obj_ids(g_products[product_index].infile, H5F_OBJ_GROUP, (size_t)3,
                              open_objs) < 0) printf("Error getting open groups.\n");
            for (j=0;j<ocount; ++j){
                numchars = H5Iget_name(open_objs[j], oname, 255);
                printf("%s is open.\n", oname);
            }
            ocount = H5Fget_obj_count(g_products[product_index].infile, H5F_OBJ_DATATYPE);
            printf("There are %d open datatypes before the close.\n", (int)ocount);
            ocount = H5Fget_obj_count(g_products[product_index].infile, H5F_OBJ_ATTR);
            printf("There are %d open attributes before the close.\n", (int)ocount);
            if(H5Fget_obj_ids(g_products[product_index].infile, H5F_OBJ_ATTR, (size_t)3,
                              open_objs) < 0) printf("Error getting open attributes.\n");
            for (j=0;j<ocount; ++j){
                numchars = H5Iget_name(open_objs[j], oname, 255);
                printf("%s is open.\n", oname);
            }
#endif
                if (status = H5Fclose(g_products[product_index].infile) < 0) {
                    NAGG_ERROR("write_missing_granules(): Failed to close file.\n");
                }
                g_products[product_index].infile = -1;
            }
/*  experimental
 *           fapl_id = H5Pcreate (H5P_FILE_ACCESS);
 *           status = H5Pset_cache(fapl_id, 0, 0, 0, 0);
*/
            if ((g_products[product_index].infile = 
/*                 H5Fopen(gran_pattern->file_in, H5F_ACC_RDONLY, fapl_id)) < 0) { */
                 H5Fopen(gran_pattern->file_in, H5F_ACC_RDONLY,  H5P_DEFAULT)) < 0) {
                NAGG_ERROR("write_missing_granules: Failed to open input file ");
                fprintf(stderr, "%s for product id %s.\n",
                gran_pattern->file_in, granule->product_id);
                status = FAIL;
                goto done;
            }
            g_products[product_index].infilename = granule->file_in;
            /* Get the number of granules in the input file for this product.
             */
            if (status = get_number_granules(&g_products[product_index]) < 0) {
                /* function prints error */
                goto done;
            }
        }


        input_file = g_products[product_index].infile;

        if ( get_product_granules_written(product_index) < 0) {
            if ((status = first_granule_file_init(output_file, input_file, granule, i_granule)) < 0) {
                goto done;
            }
        }

        status = make_granule_dataset_path(dest_dset, granule->product_name, i_granule);
        status = make_granule_dataset_path(src_dset, granule->product_name, granule->granule_input_index);


        if((dset_id = H5Dopen(input_file, src_dset, H5P_DEFAULT)) < 0 ) {
            fprintf(stderr, "Failed to open dataset %s.\n", src_dset);
            status = FAIL;
            goto done;
        }

        if ((dspace= H5Dget_space(dset_id)) < 0) {
            fprintf(stderr, "Failed to get dataspace for dataset %s", src_dset);
            status = FAIL;
            goto done;
        }
        if ((dtype = H5Dget_type(dset_id)) < 0) {
            fprintf(stderr, "Failed to open datatype for dataset.\n");
            status = FAIL;
            goto done;
        }

        if ((outdspace = H5Scopy(dspace)) < 0) { 
            fprintf(stderr, "Failed to copy dataspace.\n");
            status = FAIL;
        }

        dcpl = H5Pcreate (H5P_DATASET_CREATE);
        if((status = H5Pset_chunk (dcpl, 1, chunk)) < 0 ) {
            NAGG_ERROR("write_missing_granules():  Failed to set chunk in property list.\n");
        }

        if((ref_dset_id = H5Dcreate2(output_file, dest_dset, H5T_STD_REF_DSETREG, outdspace, H5P_DEFAULT, dcpl, H5P_DEFAULT)) < 0 ) {
            NAGG_ERROR("write_missing_granules(): Failed to create dataset ");
            fprintf(stderr, "%s in output file.\n", dest_dset);
            status = FAIL;
            goto done;
        }



        if ((numrefs = H5Sget_simple_extent_npoints(dspace)) < 1) {
            fprintf(stderr, "Failed to get number of region references in dataset %s", src_dset);
        }


        /* Allocate memory for arrays of references. */
        ref = (hdset_reg_ref_t *) HDmalloc (numrefs * sizeof(hdset_reg_ref_t));
        ref_new = (hdset_reg_ref_t *) HDmalloc (numrefs * sizeof(hdset_reg_ref_t));

        status = H5Dread(dset_id, H5T_STD_REF_DSETREG, H5S_ALL, H5S_ALL, H5P_DEFAULT, ref);
        if (status < 0) {
            fprintf(stderr, "Failed to read dataset %s.\n", src_dset);
        }

        array_size = 0;

        /* In order to call H5LRcreate_region_references, we need an hsize_t*
           array the size of 2 * rank * numrefs.  We will need to call
           H5LRget_region_info for each ref, summing the hsize_ts, allocate the
           hsize_t array, then call it again for each ref to fill in the arrays. */
        for (j=0; j<(int)numrefs; ++j) {
            status = H5LRget_region_info(input_file, (const hdset_reg_ref_t*)ref[j], &rnlength, NULL, &rank_out, NULL, &sel_type, &numelem, NULL);
            if (status < 0)
                fprintf(stderr, "Failed to get rank and number of region references.\n");
            array_size += (int)numelem * rank_out * 2;
        }

        start = (hsize_t *) HDmalloc(array_size * sizeof(hsize_t));

        dataset_paths = (char **) HDmalloc((int)numrefs * sizeof(char**));

        region_buf = start;
        for (j=0; j<(int)numrefs; ++j) {
            status = H5LRget_region_info(input_file, (const hdset_reg_ref_t*)ref[j], &rnlength, NULL, &rank_out, NULL, &sel_type, &numelem, NULL);
            if (status < 0)
                fprintf(stderr, "Failed to get region references.\n");

            reg_name = (char *) HDmalloc (rnlength);
            dataset_paths[j] = reg_name;

            status = H5LRget_region_info(input_file, (const hdset_reg_ref_t*)ref[j], &rnlength, reg_name, &rank_out, &dtype, &sel_type, &numelem, region_buf);
            if (status < 0)
                fprintf(stderr, "Failed in line 515.\n");

            /* region_buf contains an array of numbers, 2 for each rank of the
               dataset.  The first half of them represent the smallest coordinate
               in each dimension of a hyperslab and the second half the highest
               dimension of the hyperslab (in the same order).  The
               aggregation boundary should be in the first dimension, so only
               region_buf[0] and region_buf[rank] will change for the hyperslab
               of each granule.  Therefore the coordinates of the hyperslab for
               any granule (returned from H5LRget_region_info as region_buf) can
               be used to compute the coordinates of the hyperslab of any other
               granule as follows:  */
            delta = region_buf[rank_out] - region_buf[0];
            region_buf[0] = (delta +1) * i_granule;
            region_buf[rank_out] = region_buf[0] + delta;

            

            /* Here we need to create the reference(s) that H5LRcopy_reference would
               return in the write_granule after copying the data to the new file in
               the output file.  There is no data to copy, so the correct location
               will be calculated and a region reference created.  h5dump will report
               the unwritten data as fill values. */

            /* Dataset fill values are written in an attribute which is available for 
               datasets with no data.  The fill values do not need to be written to the 
               dataset. */

            /* advance to the coordinates for the next dataset's hyperslab */
            region_buf += 2*rank_out;
        }
        if((status = H5LRcreate_region_references(output_file, numrefs, dataset_paths, start, ref_new)) < 0) {
            NAGG_ERROR("write_missing_granules(): Failed to create region references\n");
            goto done;
        }

        status = H5Dwrite(ref_dset_id, H5T_STD_REF_DSETREG, H5S_ALL, H5S_ALL, H5P_DEFAULT, ref_new);

        status = add_string_attribute(ref_dset_id, "N_Granule_Status", "Missing at delivery time");

        status = H5Aiterate2(dset_id, H5_INDEX_NAME, H5_ITER_NATIVE, 0, write_fill_granule_attrs, &ref_dset_id);

        status = write_fillgr_time_attrs(ref_dset_id, granule); 
 
        H5Tclose(dtype);
        H5Sclose(dspace);
        H5Sclose(outdspace);
        H5Dclose(dset_id);
        dset_id = -1;
        H5Dclose(ref_dset_id);
        ref_dset_id = -1;

        for (j=0; j<(int)numrefs; ++j) {
            free(dataset_paths[j]);
        }
        free(dataset_paths);
        dataset_paths = 0;
        free(ref);
        ref = 0;
        free(ref_new);
        ref_new = 0;
        free(start);
        start = 0;

        if (get_product_granules_written(product_index) < 0) {
            /* Create /Data_Products/<product_name>/<product_name_Aggr dataset.  This
               should be done after H5LRcopy_reference is called for the first granule.
               The datasets for the references don't exist before that, and attributes
               on the ..._Aggr dataset from the first granule are needed.  Aggregations
               that begin with fill granules will need to wait for the first actual
               granule and extra modification of some attributes will be required.  */

            /* This is a duplicate of the code in write_granules().  I think we can move
               the code to first_granule_file_init() and eliminate the duplication.
               Better yet we can probably make it a function of its own.  Even if it's
               only called from first_granule_file_init(), it will avoid increasing
               the size of a function that's already too large. */

            input_file = g_products[product_index].infile;

            if ((status = make_All_Data_product_group_path(src_group, granule->product_name)) < 0) {
                NAGG_ERROR("write_missing_granules(): Failed to make /All_Data path for ");
                fprintf(stderr, "%s.\n", granule->product_name);
                goto done;   
            }

            if ((status = make_Data_Products_product_group_path(dest_group, granule->product_name)) < 0) {
                NAGG_ERROR("write_missing_granules(): Failed to make /Data_Products path for ");
                fprintf(stderr, "%s.\n", granule->product_name);
                goto done;   
            }

            if ((status = make_aggregate_dataset_path(dest_dset, granule->product_name)) < 0) { 
                NAGG_ERROR("write_missing_granules(): Failed to make Aggregate dataset path for ");
                fprintf(stderr, "%s.\n", granule->product_name);
                goto done;   
            }

            status = H5LRcreate_ref_to_all(output_file, src_group, dest_dset,
                                           H5_INDEX_NAME, H5_ITER_NATIVE, H5R_OBJECT);
            if (status < 0) {
                NAGG_ERROR("write_missing_granules(): Failed to create dataset\n");
                fprintf(stderr, "Failed to create %s dataset.\n", dest_dset);
            }

            if((dset_id = H5Dopen(input_file, dest_dset, H5P_DEFAULT)) < 0 ) {
                fprintf(stderr, "Failed to open dataset %s.\n", dest_dset);
                status = FAIL;
                goto done;   
            }
            if((ref_dset_id = H5Dopen(output_file, dest_dset, H5P_DEFAULT)) < 0 ) {
                fprintf(stderr, "Failed to open dataset %s in output file.\n", dest_dset);
                status = FAIL;
                goto done;
            }
            if ((H5Aiterate2(dset_id, H5_INDEX_NAME, H5_ITER_NATIVE, 0, attrit_op_func, &ref_dset_id)) < 0)
{
                fprintf(stderr, "Failed to copy attributes for dataset %s.\n", dest_dset);
                status = FAIL;
            }
            H5Dclose(ref_dset_id);
            H5Dclose(dset_id);

            status = create_aggregate_dataset(output_file, input_file, granule); 
          
            H5Fflush(output_file, H5F_SCOPE_GLOBAL);
        }
done:
        if (g_products[product_index].infile_number_of_granules ==
                granule->granule_input_index + 1) {

            if (H5Iis_valid(g_products[product_index].infile)) {
#ifdef DEBUG
            ocount = H5Fget_obj_count(g_products[product_index].infile, H5F_OBJ_DATASET);
            printf("There are %d open datasets before the close.\n", (int)ocount);
            ocount = H5Fget_obj_count(g_products[product_index].infile, H5F_OBJ_GROUP);
            printf("There are %d open groups before the close.\n", (int)ocount);
            if(H5Fget_obj_ids(g_products[product_index].infile, H5F_OBJ_GROUP, (size_t)3,
                              open_objs) < 0) printf("Error getting open groups.\n");
            for (j=0;j<ocount; ++j){
                numchars = H5Iget_name(open_objs[j], oname, 255);
                printf("%s is open.\n", oname);
            }
            ocount = H5Fget_obj_count(g_products[product_index].infile, H5F_OBJ_DATATYPE);
            printf("There are %d open datatypes before the close.\n", (int)ocount);
            ocount = H5Fget_obj_count(g_products[product_index].infile, H5F_OBJ_ATTR);
            printf("There are %d open attributes before the close.\n", (int)ocount);
            if(H5Fget_obj_ids(g_products[product_index].infile, H5F_OBJ_ATTR, (size_t)3,
                              open_objs) < 0) printf("Error getting open attributes.\n");
            for (j=0;j<ocount; ++j){
                numchars = H5Iget_name(open_objs[j], oname, 255);
                printf("%s is open.\n", oname);
            }
#endif
                if (status = H5Fclose(g_products[product_index].infile) < 0) {
                    NAGG_ERROR("write_missing_granules(): Failed to close file.\n");
                }
                g_products[product_index].infile = -1;
            }
        }

        g_products[product_index].infile = -1;
    if (ref)
        free(ref);
    return status;                                                
}


/*-------------------------------------------------------------------------
 * Function:  add_userblock_element 
 *
 * Purpose:   add an XML element with the tagname and value provided to the 
 *            designated string for a hdf5 file userblock
 *
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             November 18, 2013
 *-------------------------------------------------------------------------
 */
int
add_userblock_element(const char *tagname, const char * value, 
                      char *userblock_string)
{
    int status = 0;
    if(tagname == NULL || value == NULL || userblock_string == NULL) {
        NAGG_ERROR("add_userblock_element(): NULL input parameter provided.\n");
        status = FAIL;
        goto done;
    }

    HDstrcat(userblock_string, "<");
    HDstrcat(userblock_string, tagname);
    HDstrcat(userblock_string, ">");
    HDstrcat(userblock_string, value);
    HDstrcat(userblock_string, "</");
    HDstrcat(userblock_string, tagname);
    HDstrcat(userblock_string, ">");
done:
    return status;
}

/*-------------------------------------------------------------------------
 * Function:   userblock_setup 
 *
 * Purpose:    Create ublockinfo structures for gathering the information to
 *             be written to the output file user block.
 *
 * Note:       
 *            
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             November 15, 2013
 *
 *-------------------------------------------------------------------------
 */
int
userblock_setup(int noutfiles, const char *outgeofile,
            char *geoproduct, int nproducts, hsize_t *ublock_size)
{
    int i, status = 0;
    hsize_t ublock_min;
    *ublock_size = 2048;

    if (noutfiles > 1) {
        /* multiple output files for non-geo products - one product in each file. */
        if (outfileformat == PACKAGED) {
            NAGG_ERROR("userblock_setup():  Internal error - file format is PACKAGED \
                but nagg is creating multiple files for each aggregation.\n");
            status = FAIL;
            goto done;
        }
    } else { 
        /*  0 or 1 output file for products.  If 0 products, then only a geo file.
         *  If 1 or more products, geo will also be packaged in the file.*/
        ublock_min = (nproducts + 1) * 1536;
        if (geoproduct && !outgeofile)
            ublock_min += 1536;
        while (*ublock_size < ublock_min) {
            *ublock_size *= 2;
        }
    } 

    for(i=0; i<nproducts; ++i) {
        g_products[i].ublockinfo = (userblock_info_t*) HDmalloc(sizeof(userblock_info_t));
        g_products[i].ublockinfo->data_product
            = (data_product_info_p_t) HDmalloc(sizeof(data_product_info_t));
        g_products[i].ublockinfo->size = *ublock_size;
    }
    /* if nproducts is 0, the only data_product struct will be for the geo 
     * product.  With one or more non-geo products and a geo product, one more
     * is created for the geo product. 
     */
    if (geoproduct) { 
        g_products[nproducts].ublockinfo 
            = (userblock_info_t*) HDmalloc(sizeof(userblock_info_t));
        g_products[nproducts].ublockinfo->data_product
            = (data_product_info_p_t) HDmalloc(sizeof(data_product_info_t));
        g_products[nproducts].ublockinfo->size = 2048;
    }
    
done:
    return status;
}

/*-------------------------------------------------------------------------
 * Function:   start_write
 *
 * Purpose:    Prepare file for writing an aggregation of granules.
 *
 * Parameters: outfiles: list of output files for NPP products.      
 *             noutfiles: number of output files in outfiles.
 *             outgeofile: Geolocation output file; NULL if not needed.
 *             products_list: list of products requested.
 *             nproducts: number of products in products_list.
 *             creationtime: Creation Time stamp of the output files
 *             ngranulesperfile: number of granules per file requested.
 *
 * Notes:      Whether the output is packaged or unpackaged is dependent
 *             on the combination of values for noutfiles, nproducts, 
 *             outgeofile and geoproduct:
 *
 *             noutfiles   nproducts   outgeofile   geoproduct   output
 * 
 *                 0           0          yes          yes       geoproduct in
 *                                                               outgeofile
 *
 *                 1         1 to n       NULL         yes       packaged file
 *                                                               including geoproduct
 *
 *                 1         1 to n       NULL         NULL      packaged file not
 *                                                               including geoproduct
 *
 *                 n           n           yes         yes       unpackaged
 *    
 *                 n           n          NULL         NULL      unpackaged with
 *                                                               no geoproduct
 *             
 *              Other combinations are unsupported.
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             December 15, 2011
 *
 *-------------------------------------------------------------------------
 */
int 
start_write(NPPFileName_t *outfiles, int noutfiles, const char *outgeofile,
    char *geoproduct, char **products_list, int nproducts, 
    const char *creationdate, const char *creationtime, int ngranulesperfile)
{
    int i, product_index;
    int status = SUCCEED;
    hid_t outfile_ids[WriteGranules_Product_list_max];
    hid_t fcpl;
    g_gransperoutfile = ngranulesperfile;
    int prodfileidx;
    static char geo_product_id_backup[DPID_size+1];
    char geo_product_id[DPID_size+1];
    hsize_t ublock_size;

#if defined(DEBUG)
    printf("start_write with ngranulesperfile=%d\n", ngranulesperfile);
    printf("First outfile is %s.\n", outfiles[0]);
#endif
    g_numproducts = nproducts;

    if (nproducts > NAGG_Product_list_max) {
        NAGG_ERROR("start_write():  Illegal number of products in product list.");
        status = FAIL;
        goto done;
    } 

    if ((status = userblock_setup(noutfiles, outgeofile, geoproduct, nproducts, &ublock_size)) < 0) {
        goto done;
    }
    
    fcpl = H5Pcreate (H5P_FILE_CREATE);
    if ((status = H5Pset_userblock(fcpl, ublock_size)) < 0) {
        NAGG_ERROR("start_write(): Failed to set userblock in file creation property list.\n");
        goto done;
    }
    for (i=0;i<noutfiles;i++) {
        if((outfile_ids[i] = H5Fcreate (outfiles[i],  H5F_ACC_TRUNC, fcpl, H5P_DEFAULT)) < 0) {
            fprintf(stderr, "Failed to create output data file %s.\n", outfiles[i]);
            status = FAIL;
            goto done;
        }
    }
    if (outgeofile) {
        if ((status = H5Pset_userblock(fcpl, 2048)) < 0) { /* one product in a separate geo file. */
            NAGG_ERROR("start_write(): Failed to set geofile userblock in file creation property list.\n");
            goto done;
        }
        if((outfile_ids[i] = H5Fcreate (outgeofile,  H5F_ACC_TRUNC, fcpl, H5P_DEFAULT)) < 0) {
            fprintf(stderr, "Failed to create output geolocation file %s.\n", outgeofile);
            status = FAIL;
            goto done;
        }
    }

    /* initialize product_info_t structures for requested products. */
    for (i=0;i<nproducts;i++) {
        if (!g_input_initialized) {
            HDstrcpy(g_products[i].dpid, products_list[i]);
            g_products[i].product_name = get_product_sname_by_id(g_products[i].dpid);
            g_products[i].infilename = NULL;
            g_products[i].infile = -1;
            g_products[i].infile_number_of_granules = 0;
        }
        g_products[i].granules_written = -1;
        /* There is no infile until one is opened in write_granules()
         *  or write_missing_granules(). 
         */
        if (noutfiles == nproducts) {
            /* Each non-GEO product in its own file.  The Number_of_Data_Products 
               value in the user block will be 1 + <geoproduct>. */
            g_products[i].outfilename = outfiles[i];
            g_products[i].outfile = outfile_ids[i];
        } else if (noutfiles == 1) {
            /* 1 output file for all non-GEO products.  The Number_of_Data_Products
               value in the user block will be nproducts + <geoproduct>. */
            g_products[i].outfilename = outfiles[0];
            g_products[i].outfile = outfile_ids[0];
        } else {
            NAGG_ERROR("start_write():  Expected 1 output file or 1 output file per product.\n");   
        }
        g_products[i].geofilename = outgeofile;

        /* the third parameter should be the name of the geolocation file.   If outgeofile 
           also contains the path we may need to remove it for this function call.  */
        if (outgeofile) 
            add_string_attribute(g_products[i].outfile, "N_GEO_Ref", outgeofile);
        add_string_attribute(g_products[i].outfile, "N_HDF_Creation_Date", creationdate);    
        add_string_attribute(g_products[i].outfile, "N_HDF_Creation_Time", creationtime);    

    }

    if (geoproduct) {
        if (outgeofile) {
            /* There will be a separate GEO product output file. */
            g_products[nproducts].outfile = outfile_ids[nproducts];
            g_products[nproducts].granules_written = -1;
            g_products[nproducts].outfilename = outgeofile;
            /* outgeofile has no reference to another geofile. */
            g_products[nproducts].geofilename = NULL;
            HDstrcpy(g_products[i].dpid, geoproduct);
            g_products[nproducts].product_name = get_product_sname_by_id(g_products[nproducts].dpid);
            add_string_attribute(g_products[nproducts].outfile, "N_HDF_Creation_Date", creationdate);    
            add_string_attribute(g_products[nproducts].outfile, "N_HDF_Creation_Time", creationtime);
        } else if (nproducts == 0) {
            /* geolocation only aggregation */
            g_products[nproducts].outfile = outfile_ids[0];
            g_products[nproducts].granules_written = -1;
            g_products[nproducts].outfilename = outfiles[0];
            HDstrcpy(g_products[nproducts].dpid, geoproduct);
            g_products[nproducts].product_name = get_product_sname_by_id(g_products[nproducts].dpid);

        } else {
            /* A GEO product will be packaged in a file with the data products. */
            g_products[nproducts].outfile = outfile_ids[0];
            g_products[nproducts].granules_written = -1;
            g_products[nproducts].outfilename = outfiles[0];
            HDstrcpy(g_products[i].dpid, geoproduct);
            g_products[nproducts].product_name = get_product_sname_by_id(g_products[nproducts].dpid);

        }
        g_numproducts++;
    }
    g_input_initialized = TRUE;

done:
    return status;
} 

/*-------------------------------------------------------------------------
 * Function:  get_number_granules
 *
 * Purpose:   retrieve the number of granules in the product group of a file
 *            vi the Aggregate dataset's AggregateNumberGranules attribute.
 *
 * Parameters: product_info_t structure
 *
 *-------------------------------------------------------------------------
 */
int
get_number_granules(product_info_p_t product)
{
    int status = 0;
    hid_t attr_id = -1;
    char aggr_dset[NAGG_Dataset_Path_max + 1];

    if ((status = make_aggregate_dataset_path(aggr_dset, product->product_name)) < 0) {
        NAGG_ERROR("write_missing_granules(): Failed to make Aggregate dataset path for ");
        fprintf(stderr, "%s.\n", product->product_name);
        goto done;
    }

    if((attr_id = H5Aopen_by_name(product->infile, aggr_dset, "AggregateNumberGranules",  H5P_DEFAULT, H5P_DEFAULT)) < 0) {
        NAGG_ERROR("get_number_granules(): Failed to open attribute ");
        fprintf(stderr, "AggregateNumberGranules in file %s.\n", product->infilename);
        status = FAIL;
        goto done;
    }
    if(status = H5Aread(attr_id, H5T_NATIVE_INT, &product->infile_number_of_granules) < 0) {
        NAGG_ERROR("get_number_granules(): Failed to read attribute ");
        fprintf(stderr, "AggregateNumberGranules in file %s.\n", product->infilename);
        goto done;
     }
 
done:
    if (H5Iis_valid(attr_id))
        H5Aclose(attr_id);
    return status;
}

/*-------------------------------------------------------------------------
 * Function:  write_granules 
 *
 * Purpose:   write a granule to one of the output files. 
 *
 * Parameters:  granule - points at a real granule to be written.
 *              i_granule - the i-th granule within this file.
 *                  0 <= i_granule < number-of-granules-per-file.
 *
 * Notes:      The input file for a product is first opened in this function
 *             for a real granule, or in missing_granules for a fill granule.
 *             We can omit closing the file if the granule written has 
 *             a granule_input_index less than the number of granules in the
 *             input aggregation.  For each granule we will need to compare its
 *             input file name with the name of the current input file.  If 
 *             they don't match, close the current input file, open the granule
 *             input file, get its aggregation number and make it current.
 *
 *             If the string comparisons are inefficient we could create an
 *             index of input files, change the granule's file_in member from 
 *             the file name to the file index and compare the granule's input  
 *             file index against the index of the current product input file.
 *
 *             The end_write() function should close any open input files.
 *
 *             If there is a gap between granules written so far and i_granule,
 *             need to write the missing fill granules. (I think that 
 *             select_granules() is also creating fill granules and sending
 *             each one over.  If so, the gap will have been filled before 
 *             this function is called.)
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             December 15, 2011
 *
 *-------------------------------------------------------------------------
 */
int
write_granules(granule_p_t granule, int i_granule)
{
    int i, j;
    int granule_index;
    int outgran_index;
    char dest_dset[NAGG_Dataset_Path_max];
    char src_dset[NAGG_Dataset_Path_max];
    hid_t output_file;
    hid_t input_file;
/*    hid_t fapl_id; */
    hid_t group_id, group_id2;
    int product_index;
    htri_t exists;
    hbool_t real_granule;
    int status = SUCCEED;
    ssize_t ocount = 0, numchars;
    hid_t open_objs[3] = {-1, -1, -1};
    char oname[256];

#ifdef DEBUG
    printf("1962: Write granule %d for product %s.\n", i_granule, granule->product_name);
#endif

    real_granule = HDstrneq(granule->granule_id, "N/A");

    product_index = get_product_idx_by_id(granule->product_id);
#ifdef DEBUG
    printf("1969; Product_index is %d for product %s.\n", product_index, granule->product_name);
#endif

    output_file = g_products[product_index].outfile;   

    if (real_granule) {
        if(g_products[product_index].infilename != granule->file_in) {
            /* Need to close the open file and open a new one. */
            if (g_products[product_index].infilename != NULL
                    && H5Iis_valid(g_products[product_index].infile)) {
#ifdef DEBUG
                ocount = H5Fget_obj_count(g_products[product_index].infile, H5F_OBJ_DATASET);
                printf("1981: There are %d open datasets before the close.\n", (int)ocount);
                ocount = H5Fget_obj_count(g_products[product_index].infile, H5F_OBJ_GROUP);
                printf("1983: There are %d open groups before the close.\n", (int)ocount);
                if(H5Fget_obj_ids(g_products[product_index].infile, H5F_OBJ_GROUP, (size_t)3,
                                  open_objs) < 0) printf("Error getting open groups.\n");
                for (j=0;j<ocount; ++j){
                    numchars = H5Iget_name(open_objs[j], oname, 255);
                    printf("%s is open.\n", oname);
                }
                ocount = H5Fget_obj_count(g_products[product_index].infile, H5F_OBJ_DATATYPE);
                printf("1991: There are %d open datatypes before the close.\n", (int)ocount);
                ocount = H5Fget_obj_count(g_products[product_index].infile, H5F_OBJ_ATTR);
                printf("1993: There are %d open attributes before the close.\n", (int)ocount);
                if(H5Fget_obj_ids(g_products[product_index].infile, H5F_OBJ_ATTR, (size_t)3,
                                  open_objs) < 0) printf("Error getting open attributes.\n");
                for (j=0;j<ocount; ++j){
                    numchars = H5Iget_name(open_objs[j], oname, 255);
                    printf("%s is open.\n", oname);
                }
#endif
                if (status = H5Fclose(g_products[product_index].infile) < 0) {
                    NAGG_ERROR("write_granules(): Failed to close file.\n");
                }
                g_products[product_index].infile = -1;
            }
/*            fapl_id = H5Pcreate (H5P_FILE_ACCESS);
            status = H5Pset_cache(fapl_id, 0, 0, 0, 0);
            if ((g_products[product_index].infile = H5Fopen(granule->file_in, H5F_ACC_RDONLY, fapl_id)) < 0) {
*/
            if ((g_products[product_index].infile = H5Fopen(granule->file_in, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0) {
                fprintf(stderr, "Failed to open input file %s for product id %s.\n",
                        granule->file_in, granule->product_id);
                status = FAIL;
                goto done;
            }
            g_products[product_index].infilename = granule->file_in;
        }

        input_file = g_products[product_index].infile;
        
        if ( get_product_granules_written(product_index) < 0) {
            if ((status = first_granule_file_init(output_file, input_file, granule, i_granule)) < 0) {
                goto done;
            }
            /* Get the number of granules in the input file for this product. 
             */
            if (status = get_number_granules(&g_products[product_index]) < 0) {
                /* function prints error */
                goto done;
            }
        }

        if ((status = make_granule_dataset_path(dest_dset, granule->product_name, i_granule)) <0) {
            goto done;
        }

        if ((status = make_granule_dataset_path(src_dset, granule->product_name, 
                      granule->granule_input_index)) < 0) {
            goto done;
         }

        /* Provided the granule doesn't exist in the output file, copy it from input to output */
        if((exists = H5Lexists(output_file, dest_dset, H5P_DEFAULT)) == 0) {    
#ifdef DEBUG
            printf("write_granules(): Copy %s from input file to %s in output file.\n", src_dset, dest_dset);
#endif
            if ((H5Ocopy(input_file, src_dset, output_file, dest_dset, 
                         H5P_DEFAULT, H5P_DEFAULT)) < 0) {
                fprintf(stderr, "Failed to copy %s to %s.\n", src_dset,  g_products[product_index].outfilename); 
            }
        } else if (exists > 0) {
            fprintf(stderr, "%s in %s already exists.\n", src_dset,  g_products[product_index].outfilename);
        } else {
            fprintf(stderr, "Call to H5Lexists failed for %s in %s.\n", src_dset, g_products[product_index].outfilename);
        }

        if ((status = copy_granule_hyperslabs(output_file, input_file, src_dset, dest_dset, 
                                      g_products[product_index].outfilename, i_granule)) < 0) {
            goto done;
        }

        if (get_product_granules_written(product_index) < 0) { 
            if((status = create_aggregate_dataset(output_file, input_file, 
                                                  granule)) < 0) {
                goto done;
            }
        }    

    /*  For fill granules, call write_missing_granule */
    } else {
        outgran_index = get_last_i_granule(product_index);
        if ((status = write_missing_granules(output_file, input_file, 
             g_products[product_index].outfilename, granule, outgran_index, i_granule)) < 0)
            goto done;
    } 
    g_products[product_index].last_i_granule = i_granule;
    /* The assignment seems useless - is the function call useful? */
    outgran_index = increment_granules_written(product_index); 
done:
    /* check to see if all granules have been written.  If not,
     * don't close the file.
     */
    if (g_products[product_index].infile_number_of_granules ==
            granule->granule_input_index + 1) {
        if (g_products[product_index].infile > 0) {
#ifdef DEBUG
            ocount = H5Fget_obj_count(g_products[product_index].infile, H5F_OBJ_DATASET);
            printf("There are %d open datasets before the close.\n", (int)ocount);
            ocount = H5Fget_obj_count(g_products[product_index].infile, H5F_OBJ_GROUP);
            printf("There are %d open groups before the close.\n", (int)ocount);
            if(H5Fget_obj_ids(g_products[product_index].infile, H5F_OBJ_GROUP, (size_t)3,
                              open_objs) < 0) printf("Error getting open groups.\n");
            for (j=0;j<ocount; ++j){
                numchars = H5Iget_name(open_objs[j], oname, 255);
                printf("%s is open.\n", oname);
            }
            ocount = H5Fget_obj_count(g_products[product_index].infile, H5F_OBJ_DATATYPE);
            printf("There are %d open datatypes before the close.\n", (int)ocount);
            ocount = H5Fget_obj_count(g_products[product_index].infile, H5F_OBJ_ATTR);
            printf("There are %d open attributes before the close.\n", (int)ocount);
            if(H5Fget_obj_ids(g_products[product_index].infile, H5F_OBJ_ATTR, (size_t)3,
                              open_objs) < 0) printf("Error getting open attributes.\n");
            for (j=0;j<ocount; ++j){
                numchars = H5Iget_name(open_objs[j], oname, 255);
                printf("%s is open.\n", oname);
            }
#endif
            if (H5Iis_valid(g_products[product_index].infile)) {
                if (status = H5Fclose(g_products[product_index].infile) < 0) {
                    NAGG_ERROR("write_granules(): Failed to close file.\n");
                }
            }
            g_products[product_index].infile = -1;
            g_products[product_index].infilename = NULL;
        }
    }
    return status;
}

/*-------------------------------------------------------------------------
 * Function:   create_granule_regrefs
 *
 * Purpose:    In nagg versions 1.5.1 and before, nagg copied region references 
 *             for datasets and hyperslabs that consisted entirely of fill
 *             values.  The resulting hyperslabs in the output files took 
 *             much more space than those in the input files because the 
 *             original NPP files did not have data written for hyperslabs 
 *             containing only a fill value.  The solution that was implemented
 *             checked the storage size of the input hyperslabs and did not
 *             copy those with 0 size.  What was overlooked was that the output
 *             files were missing the region references to the hyperslabs of 
 *             fill values.
 *             
 *             This function reads the object references from the Aggregate 
 *             dataset for a product, and gets the dimensions of each referenced 
 *             dataset.  Creating the region reference for the hyperslabs in
 *             an NPP file requires the beginning corner of the hyperslab and
 *             the size in each dimension.  The beginning corner is the offset 
 *             from 0 in the first dimension and 0 in the remaining dimensions.
 *             The size of each hyperslab in the first dimension is calculated
 *             by dividing the size of the first dimension of the referenced
 *             datset by the  number of granules in the input aggregation.  
 *             The offset is this size times the granule index in the output file.
 *             The size of the hyperslab is the calculated size in the first 
 *             dimension and the dataset size in the other dimensions.
 *
 * Note:
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             March 21, 2014
 *
 *-------------------------------------------------------------------------
 */
static int
create_granule_regrefs(hid_t output_file, hid_t input_file, char *src_dset,
                       char *dest_dset, const char *outfilename, int gran_index,
                       hid_t ref_dset_id, hdset_reg_ref_t *ref_new)
{
    char *pos = NULL;
    unsigned int len = 0;
    int i = 0, j = 0, status = 0;
    hid_t attr_in = -1;
    hid_t obj = -1;
    hid_t aggr_dset_id = -1;
    hid_t dspace = -1;
    int aggrnum = 0;
    char aggr_dset_path[NAGG_Dataset_Path_max];
    int ndims = 0;
    hsize_t *dims;
    hsize_t aggrdims[1];
    hobj_ref_t *rdata;
    char **ref_names;
    hsize_t *region_buf = NULL;
    ssize_t size;
    int bufsize;
       
        /* Get the name of the Aggregate dataset name by replacing "Gran_*" at the end of 
         * src_dset with "Aggr".
         */
        pos = HDstrstr(src_dset, "Gran_");
        len = pos - src_dset;
        HDstrncpy(aggr_dset_path, src_dset, len);
        aggr_dset_path[len] = '\0'; 
        HDstrncat(aggr_dset_path, "Aggr/", 4);
        if((attr_in = H5Aopen_by_name(input_file, aggr_dset_path, "AggregateNumberGranules", H5P_DEFAULT, H5P_DEFAULT)) < 0) {
            fprintf(stderr, "Failed to open attribute %s/AggregateNumberGranules.\n", aggr_dset_path);
            status = FAIL;
            goto done;
        } else if ((status = H5Aread(attr_in, H5T_NATIVE_INT, &aggrnum)) < 0) {
            fprintf(stderr, "Failed to read attribute %s/AggregateNumberGranules.\n", aggr_dset_path);
            status = FAIL;
            goto done;
        }

        aggr_dset_id = H5Dopen(input_file, aggr_dset_path, H5P_DEFAULT);

        dspace = H5Dget_space(aggr_dset_id);
        ndims = H5Sget_simple_extent_dims (dspace, aggrdims, NULL);
        rdata = (hobj_ref_t *) HDmalloc (aggrdims[0] * sizeof (hobj_ref_t));

        if ((status = H5Dread (aggr_dset_id, H5T_STD_REF_OBJ, H5S_ALL, H5S_ALL, H5P_DEFAULT, rdata)) <0 ) {
            NAGG_ERROR("create_granule_regrefs():  Failed to read Aggregate dataset.\n");
            H5Dclose(aggr_dset_id);
            H5Sclose(dspace);
            goto done;
        }

        /* Let's try allocating an array of char*s for all the dataset names. */
        ref_names = (char **) HDmalloc(aggrdims[0] * sizeof (char*));

        for (i=0; i<aggrdims[0]; i++) {
            hid_t refdsetid;
            hid_t refdspace;
            size_t refndims;
            hsize_t *start, *count;

            /*
             * Open the referenced object, get its name and type.
             */
            obj = H5Rdereference_ (aggr_dset_id, H5R_OBJECT, &rdata[i]);

            /*
             * Get the length of the name, allocate space, then retrieve
             * the name.
             */
            size = 1 + H5Iget_name (obj, NULL, 0);
            ref_names[i] = (char *) HDmalloc (size);
            size = H5Iget_name (obj, ref_names[i], size);

            status = H5Oclose (obj);

            /*
             * We now have the name of the dataset, the number of granules 
             * in the input aggregation, and the index of the hyperslab, but still
             * need to get the dataset dimensions and divide the first one by the
             * number of input granules.  Then the first number in the region 
             * reference is the first dimension of a hyperslab times the index of 
             * the hyperslab.  The rest of the first set of region reference numbers
             * will be 0s. The first of the second set of region reference numbers 
             * will be the first number in the first set plus the first dimension 
             * of a hyperslab.  The rest of the second set will be the remaining
             * dataset dimensions.
             */

            refdsetid = H5Dopen(input_file, ref_names[i], H5P_DEFAULT);
            refdspace = H5Dget_space(refdsetid);
            refndims = H5Sget_simple_extent_ndims(refdspace);

            dims = (hsize_t *)HDmalloc(refndims * sizeof(hsize_t));
            start = (hsize_t *)HDmalloc(refndims * sizeof(hsize_t));
            count = (hsize_t *)HDmalloc(refndims * sizeof(hsize_t));

            bufsize = refndims * sizeof(hsize_t) * 2;
            region_buf = (hsize_t *)HDmalloc(bufsize);

            if(refndims != H5Sget_simple_extent_dims(refdspace, dims, NULL)) {
                fprintf(stderr, "create_granule_regrefs(): different numbers of dimensions returned.\n");
                status = FAIL;
                goto done;
            }

            /* enter beginning and ending value for each dimension. */
            /* NPP datasets have granule boundaries in the first dimension.
             * Each hyperslab contains the data for one granule.  The size of
             * each hyperslab in the first dimension is the total size of the
             * first dimension divided by the number of granules.  The other 
             * dimensions are the same for the hyperslabs as for the entire 
             * datset.
             */
            start[0] = gran_index * dims[0] / aggrnum;
            count[0] = (dims[0] / aggrnum);
            for (j=1; j<refndims; ++j) {
                start[j] = 0;
                count[j] = dims[j];
            }

            status = H5Sselect_hyperslab(refdspace, H5S_SELECT_SET, start, NULL, count, NULL); 
            status = H5Rcreate(ref_new[i], output_file, ref_names[i], H5R_DATASET_REGION, refdspace);

            if (status < 0) {
                fprintf(stderr, "Create region references failed on  the %d dataset, %s.\n", i, ref_names[i]);
                goto done;
            }
            if (H5Iis_valid(refdsetid))
            H5Dclose(refdsetid);
            if (H5Iis_valid(refdspace))
            H5Sclose(refdspace);
            if (dims)
                HDfree (dims); 
            if (start)
                HDfree (start); 
            if (count)
                HDfree (count); 
            if (region_buf)
                HDfree(region_buf);
        }
        if ((status = H5Dwrite(ref_dset_id, H5T_STD_REF_DSETREG, H5S_ALL, H5S_ALL, H5P_DEFAULT, ref_new)) < 0) {
            NAGG_ERROR("create_granule_regrefs():  Failed to write region references to Dataset.\n");
        }
        if (H5Iis_valid(aggr_dset_id))
            H5Dclose (aggr_dset_id);
        if (H5Iis_valid(dspace))
            H5Sclose (dspace);
        if (rdata)
            HDfree(rdata); 
        for(i=0;i<aggrdims[0]; ++i) {
            if (ref_names[i])
                HDfree(ref_names[i]); 
        }
        if (ref_names)
            HDfree(ref_names);

done:
    if (H5Iis_valid(attr_in))
        H5Aclose(attr_in);
    return status;
}

/*-------------------------------------------------------------------------
 * Function:   copy_granule_hyperslabs
 *
 * Purpose:    Copy the hyperslab of raw data for a granule to the correct
 *             location in the output file.
 *
 * Note:       
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             February 10, 2012
 *
 *-------------------------------------------------------------------------
 */
static int
copy_granule_hyperslabs(hid_t output_file, hid_t input_file, char *src_dset, 
                        char *dest_dset, const char *outfilename, int gran_index)
{
    hdset_reg_ref_t *ref = NULL;
    hdset_reg_ref_t *ref_new = NULL;;
    char *reg_name = NULL;
    char *aggr_dset_name = NULL;
    hsize_t *region_buf = NULL;
    size_t rnlength = 0;
    int rank_out = 0;
    size_t numelem = 0;
    hssize_t numrefs = 0;
    H5S_sel_type sel_type = -1;
    size_t size_ref = 0;
    hsize_t delta = 0;
    hid_t dtype = -1;
    hid_t dset_id = -1;
    hid_t ref_dset_id = -1;
    hid_t hs_dset_id = -1;
    hsize_t hs_dset_size = 0;
    hid_t dspace = -1;
    int i = 0, j = 0, status = 0;
    int result = 0;

    /* Open input and output granule datasets for copying data to new file and
       writing reference in the new granule dataset for the new data location. */
    if((dset_id = H5Dopen(input_file, src_dset, H5P_DEFAULT)) < 0 ) {
        fprintf(stderr, "Failed to open dataset %s in %s.\n", src_dset, outfilename);
        status =FAIL;
        goto done;
    }

    if((ref_dset_id = H5Dopen(output_file, dest_dset, H5P_DEFAULT)) < 0 ) {
        fprintf(stderr, "Failed to open dataset %s.\n", dest_dset);
        status = FAIL;
        goto done;
    }

    if ((dspace= H5Dget_space(dset_id)) < 0) {
        fprintf(stderr, "Failed to get dataspace for dataset %s", src_dset);
        status = FAIL;
        goto done;
    }
    if ((dtype = H5Dget_type(dset_id)) < 0) {
        fprintf(stderr, "Failed to open datatype for dataset.\n");
        status = FAIL;
        goto done;
    }

    if ((numrefs = H5Sget_simple_extent_npoints(dspace)) < 1) {
        fprintf(stderr, "Failed to get number of region references in dataset %s.", src_dset);
    }
    
    /* Allocate memory for arrays of references. */  
    size_ref = numrefs * sizeof(hdset_reg_ref_t);
    ref = (hdset_reg_ref_t *) HDmalloc (size_ref);
    ref_new = (hdset_reg_ref_t *) HDmalloc (size_ref);
 
    status = H5Dread(dset_id, H5T_STD_REF_DSETREG, H5S_ALL, H5S_ALL, H5P_DEFAULT, ref);
    if (status < 0) {
        NAGG_ERROR("copy_granule_hyperslabs(): Failed to read dataset ");
        fprintf(stderr, "%s.\n", src_dset);
        goto done;
    }
    /* set contents of ref_new to 0s. */
    memset(ref_new, 0, size_ref);

    /* comparison of the read result to all 0s.  If they are the same, the 
     * region references were missing.
     */
    result = memcmp(ref, ref_new, size_ref);

    /* If the region references are missing in the input file the hyperslabs 
     * can't be determined by calling H5LRget_region_info on a NULL reference.
     * We will need to get the dataset references from the input file's aggregate
     * dataset, and for each dataset calculate the reference corners for the 
     * current granule.
     */
    if (result == 0) {
        status = create_granule_regrefs(output_file, input_file, src_dset,
                       dest_dset, outfilename, gran_index, ref_dset_id, ref_new);
        /* after this executes, a few handles are closed and buffers freed, then 
         * status is returned.  No need for "goto done" if the status is FAILED. */
    } else {  

        for (j=0; j<(int)numrefs; ++j) {
       
            /* The first call gets the rank and dimensions for allocating storage space 
               in the buffers for the dataset name and the corners of the region.  The
               second call gets the dataset name (reg_name) and coordinates. */
            status = H5LRget_region_info(dset_id, (const hdset_reg_ref_t*)ref[j], &rnlength, NULL, &rank_out, NULL, &sel_type, &numelem, NULL);
            if (status < 0) {
                fprintf(stderr, "copy_granule_hyperslabs(): H5LRget_region_info() failed to get rank and dimensions.\n");
                goto done;
            }

            reg_name = (char *) HDmalloc (rnlength);
    
            region_buf = (hsize_t *)HDmalloc((int)numelem * rank_out * sizeof(hsize_t) * 2);

     
            status = H5LRget_region_info(dset_id, (const hdset_reg_ref_t*)ref[j], &rnlength, reg_name, &rank_out, &dtype, &sel_type, &numelem, region_buf);
            if (status < 0) {
                fprintf(stderr, "copy_granule_hyperslabs(): failed to get dataset name and coordinates.\n");
                goto done;
            }

            /* region_buf is array of numbers, 2 for each rank of the dataset.  
               The first half of them represent the smallest coordinate in each 
               dimension of a hyperslab and the second half the highest.  The 
               aggregation boundary should be in the first dimension, so only
               the region_buf[0] and region_buf[rank] will change for the hyperslab
               when a granule is moved to a different position.  */ 
            delta = region_buf[rank_out] - region_buf[0]; 
            region_buf[0] = (delta +1)  * gran_index;
            region_buf[rank_out] = region_buf[0] + delta; 

            /* Check first to see if the dataset containing the region has any data.
               if it doesn't, skip to the next reference to avoid writing fill values. */
            hs_dset_id = H5Rdereference_ (input_file, H5R_DATASET_REGION, (const hdset_reg_ref_t*)ref[j]);
            hs_dset_size = H5Dget_storage_size(hs_dset_id);
    
            if ((status = H5Dclose(hs_dset_id)) < 0) {
                fprintf(stderr, "copy_granule_hyperslabs(): failed to close dereferenced dataset.\n");
                goto done;
            }

            if (hs_dset_size == 0) {
            
                status = H5LRcreate_region_references(output_file, numelem, &reg_name,
                             region_buf, ref_new[j]);
    
                if (status < 0) {
                    fprintf(stderr, "Create region references failed on the %d dataset, %s.\n", j, reg_name);
                    goto done;
                }
    
            } else {

                /* Call H5LRcopy_reference to copy the data pointed to by the reference
                   in the input file to the calculated location in the output file.  The 
                   function also returns a region reference to the location of the data
                   in the new file, which is stored in ref_new[j] and written to the 
                   new _Gran_n dataset in the output file after the data for all region
                   references has been copied. */
                status = H5LRcopy_reference(input_file, (const hdset_reg_ref_t*)ref[j], 
                            outfilename, reg_name, region_buf, (hdset_reg_ref_t*)ref_new[j]);   
                if (status < 0) {
                    fprintf(stderr, "Copy reference failed on the %d dataset, %s.\n", j, reg_name);
                    goto done; 
                }
            }
            HDfree(reg_name);
            HDfree(region_buf);
        }
   
        /* Write all the region references for the granule to the new _Gran_n dataset
           in the output file. */
        if((status = H5Dwrite(ref_dset_id, H5T_STD_REF_DSETREG, H5S_ALL, H5S_ALL, H5P_DEFAULT, ref_new)) < 0) {
            NAGG_ERROR("copy_granule_hyperslabs(): failed to write region references to output file.\n");
            if (H5Iis_valid(dtype))
                H5Tclose(dtype);
            if (H5Iis_valid(dspace))
                H5Sclose(dspace);
            goto done;
        }
    }
    if (H5Iis_valid(dtype))
        H5Tclose(dtype);
    if (H5Iis_valid(dspace))
        H5Sclose(dspace);
    if (H5Iis_valid(hs_dset_id))
        H5Dclose(hs_dset_id);

    if (H5Iis_valid(dset_id))
        H5Dclose(dset_id);
    if (H5Iis_valid(ref_dset_id))
         H5Dclose(ref_dset_id); 
    H5Fflush(output_file, H5F_SCOPE_GLOBAL);
done:
    if (ref)
        free(ref);
    ref = NULL;
    if (ref_new)
        free(ref_new);
    ref_new = NULL;
    return status;    
}

/*-------------------------------------------------------------------------
 * Function:   create_aggregate_dataset
 *
 * Purpose:     
 *
 * Note:       
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             February 10, 2012
 *
 *-------------------------------------------------------------------------
 */
static int
create_aggregate_dataset(hid_t output_file, hid_t input_file, granule_p_t granule) 
{
    char src_group[NAGG_Group_Path_max];
    char dest_group[NAGG_Group_Path_max];
    char dest_dset[NAGG_Dataset_Path_max];
    char src_dset[NAGG_Dataset_Path_max];
    hid_t dset_id = -1;
    hid_t ref_dset_id = -1;
    hid_t attr, dpgroup;
    int status = SUCCEED;
    char *attr_val[NAGG_String_attribute_valstring_max];
    int product_index = get_product_idx_by_id(granule->product_id);

        /* Create /Data_Products/<product_name>/<product_name_Aggr dataset.  This 
           should be done after H5LRcopy_reference is called for the first granule.
           The datasets for the references don't exist before that, and attributes
           on the ..._Aggr dataset from the first granule are needed.  Aggregations
           that begin with fill granules will need to wait for the first actual
           granule and extra modification of some attributes will be required.  */

    status = make_All_Data_product_group_path(src_group, granule->product_name);
    status = make_Data_Products_product_group_path(dest_group, 
                                                   granule->product_name);
    status = make_aggregate_dataset_path(dest_dset, granule->product_name);
    status = make_granule_dataset_path(src_dset, granule->product_name, 
                                       granule->granule_input_index);

    status = H5LRcreate_ref_to_all(output_file, src_group, dest_dset,
                                       H5_INDEX_NAME, H5_ITER_NATIVE, H5R_OBJECT);
    if (status < 0) {
        NAGG_ERROR("create_aggregate_dataset(): Failed to create dataset\n");
        fprintf(stderr, "Failed to create %s dataset.\n", dest_dset);
        status = FAIL;
        goto done;
    }

    if((dset_id = H5Dopen(input_file, dest_dset, H5P_DEFAULT)) < 0 ) {
        fprintf(stderr, "Failed to open dataset %s.\n", dest_dset);
        status = FAIL;
        goto done;
    }
    if((ref_dset_id = H5Dopen(output_file, dest_dset, H5P_DEFAULT)) < 0 ) {
        fprintf(stderr, "Failed to open dataset %s in output file.\n", dest_dset);
        status = FAIL;
        goto done;
    }
    if ((H5Aiterate2(dset_id, H5_INDEX_NAME, H5_ITER_NATIVE, 0, attrit_op_func, &ref_dset_id)) < 0) {
        fprintf(stderr, "Failed to copy attributes for dataset %s.\n", dest_dset);
        status = FAIL;
        goto done;
    }

    H5Dclose(ref_dset_id);
    ref_dset_id = -1;
    H5Dclose(dset_id);
    dset_id = -1;

    /* Four product group attribute values should be written to the output file's
       user block as subelements of the Data_Product element for this product. */
    if ((dpgroup = H5Gopen(input_file, dest_group, H5P_DEFAULT)) < 0) {
        fprintf(stderr, "Failed to open input file %s group.\n", 
                           dest_group);
        status = FAIL;
        goto done;
    }

    /* Four beginning attribute values should be written from the first granule to be written 
       to the new output file (or the first fill granule if there is one):  AggregateBeginningDate
       from Beginning_Date, AggregateBeginningGranuleID from N_Granule_ID, AggregateBeginningOrbitNumber
       from N_Beginning_Orbit_Number, and AggregateBeginningTime from Beginning_Time. */

    if((dset_id = H5Dopen(output_file, dest_dset, H5P_DEFAULT)) < 0 ) {
        fprintf(stderr, "Failed to open dataset %s in output file.\n", dest_dset);
        status = FAIL;
        goto done;
    }
    if ((status = update_string_attribute(dset_id, "AggregateBeginningDate", granule->beginning_date)) < 0) {
        NAGG_ERROR("create_aggregate_dataset(): Failed to write value to AggregateBeginningDate\n");
        goto done;
    }
    if ((status = update_string_attribute(dset_id, "AggregateBeginningGranuleID", granule->granule_id)) < 0) {
        NAGG_ERROR("create_aggregate_dataset(): Failed to write value to AggregateBeginningGranuleID\n");
        goto done;
    }
    if ((attr = H5Aopen(dset_id, "AggregateBeginningOrbitNumber",  H5P_DEFAULT)) < 0) {
        NAGG_ERROR("create_aggregate_dataset(): Failed to open attribute AggregateBeginningOrbitNumber for writing\n");
        status = FAIL;
        goto done;
    } 
    if ((status = H5Awrite(attr, H5T_NATIVE_ULLONG, &(granule->orbit_number))) < 0) {
        fprintf(stderr, "Failed to write attribute AggregateBeginningOrbitNumber.\n");
    }
    if(H5Aclose(attr) < 0)
        NAGG_ERROR("create_aggregate_dataset(): Failed to close attribute AggregateBeginningOrbitNumber.\n");
    if ((status = update_string_attribute(dset_id, "AggregateBeginningTime", granule->beginning_time )) < 0) {
        NAGG_ERROR("create_aggregate_dataset(): Failed to write value to AggregateBeginningTime\n");
        goto done;
    }

done:
    if (H5Iis_valid(ref_dset_id))      
        H5Dclose(ref_dset_id);
    if (H5Iis_valid(dset_id))      
        H5Dclose(dset_id);
    if (H5Iis_valid(dpgroup))
        H5Gclose(dpgroup);
    return status;
}

/*-------------------------------------------------------------------------
 * Function:   add_fileinfopart_to_ub_string  
 *
 * Purpose:    Add the specified attribute values from the file attributes to
 *             the userblock string.
 *
 * Note:       
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             November 15, 2013
 *
 *-------------------------------------------------------------------------
 */
static int        
add_fileinfopart_to_ub_string(unsigned int products_in_file, hbool_t last_file,
                              char *user_block_str)
{
    int status = SUCCEED;
    char number_products[NAGG_Numeric_attr_valstring_max];

    if ((status = add_userblock_element("Mission_Name",
                                        g_products[0].ublockinfo->mission_name, 
                                        user_block_str)) < 0) {
        fprintf(stderr, "Failed to set user block value for Mission_Name attribute.\n");
        goto done;
    }

    if ((status = add_userblock_element("Platform_Short_Name",
                                        g_products[0].ublockinfo->platform_name,
                                        user_block_str)) < 0) {
        fprintf(stderr, "Failed to set user block value for Platform_Short_Name attribute.\n");
        goto done;
    }

    if (g_products[0].geofilename && HDstrneq(g_products[0].geofilename, "")
            && !last_file) {
        if ((status = add_userblock_element("N_GEO_Ref",
                                            g_products[0].geofilename,
                                            user_block_str)) < 0) {
            fprintf(stderr, "Failed to set user block value for N_GEO_Ref attribute.\n");
            goto done;
        }
    } else {
        if ((status = add_userblock_element("N_GEO_Ref",
                                                 "", user_block_str)) < 0) {
            fprintf(stderr, "Failed to set user block value for N_GEO_Ref attribute.\n");
            goto done;
        }
    }

    sprintf(number_products, "%u", products_in_file);
    if ((status = add_userblock_element("Number_Of_Data_Products",
                                        number_products, user_block_str)) < 0) {
        fprintf(stderr, "Failed to set user block value for Number_Of_Data_Products attribute.\n");
        goto done;
    }
done:
    return status;
}

/*-------------------------------------------------------------------------
 * Function:   end_write
 *
 * Purpose:    provide notification that an aggregation is finished and the 
 *             files can be closed.
 *
 * Note:       g_numproducts is the total number of products to be aggregated, 
 *             including the geoproduct.
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             December 15, 2011
 *
 *-------------------------------------------------------------------------
 */
int
end_write()
{ 
    int status = SUCCEED;
    int i, files_to_close;


    /* There should be 1 file for each product, or 1 file for all products.
       If there's only 1, closing the file for the first product will close
       the file for all products, otherwise close them all. 
        */
    /* write file attribute subset to userblock strings. If output is packaged,
       there is only one userblock, the file attribute subset can be written once,
       and all the Data_Product attributes should be added to the same userblock
       for the one file.  If output is unpackaged each product will have attributes
       for its own file.*/
    if (outfileformat == PACKAGED || nproducts == 0) {
        files_to_close = 1;
        /* only one file, so the file attributes can be stored in g_products[0]. */
        if ((status = store_userblock_file_attributes(g_numproducts)) < 0) {
            goto done;
        }   
    } else {
        files_to_close = g_numproducts;
        /* Every product is in its own file. The number of data products for each is 1.
         * The values of these attributes are the same for all of the files, so
         * they can still be stored in g_products[0]*/
        if ((status = store_userblock_file_attributes(1)) < 0) {
            goto done;
        }
    }

    /* update the ending attributes and write the Data_Products element of the
       userblock strings for all products.*/
    for (i=0;i<g_numproducts;++i) {

        if ((status = update_aggregate_ending_attributes(i)) < 0) {
            goto done;
        }
    }

    for (i=0;i<g_numproducts;++i) {
        if ((status = store_userblock_product_attributes(i)) < 0) {
            goto done;
        }
    }

    /*  With multiple products written to the same file, all the updates must
        be done before the file is closed, hence the separation of the update
        from closing files. */
    /* status not checked - Error messages printed in function and the next 
       statement is in "done". */
    status = write_userblocks_and_close(files_to_close);

done:

    return status;
}

/*-------------------------------------------------------------------------
 * Function:   add_dataproduct_to_ub_string
 *
 * Purpose:    Add the information for all of the products included in the 
 *             output file to the userblock string.
 *             
 * Note:       
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             November 15, 2013
 *
 *-------------------------------------------------------------------------
 */
int
add_dataproduct_to_ub_string(int product_index, char *user_block_str)
{
    int status = SUCCEED;
    char attr_val[NAGG_String_attribute_valstring_max];

    HDstrcat(user_block_str, "<Data_Product>");

    if ((status = add_userblock_element("N_Collection_Short_Name", 
                    g_products[product_index].ublockinfo->data_product->product_name, user_block_str)) < 0) {
        fprintf(stderr, "Failed to set user block value for N_Collection_Short_Name attribute.\n");
        goto done;
    }

    if ((status = add_userblock_element("Instrument_Short_Name", 
                    g_products[product_index].ublockinfo->data_product->instrument_name, user_block_str)) < 0) {
        fprintf(stderr, "Failed to set user block value for Instrument_Short_Name attribute.\n");
        goto done;
    }

    if ((status = add_userblock_element("N_Dataset_Type_Tag", 
                    g_products[product_index].ublockinfo->data_product->dataset_type_tag, user_block_str)) < 0) {
        fprintf(stderr, "Failed to set user block value for N_Dataset_Type_Tag attribute.\n");
        goto done;
    }

    if ((status = add_userblock_element("N_Processing_Domain", 
                    g_products[product_index].ublockinfo->data_product->processing_domain, user_block_str)) < 0) {
        fprintf(stderr, "Failed to set user block value for N_Processing_Domain attribute.\n");
        goto done;
    }

    /* They should also be written to the user block.  Three can be written now, but the
     * AggregateBeginningGranuleID will need to be retrieved from the 
     * AggregateBeginningGranuleID product to be written after AggregateEndingDate, 
     * AggregateEndingOrbitNumber, and AggregateEndingTime.
     */
    
    if ((status = add_userblock_element("AggregateBeginningDate",
                      g_products[product_index].ublockinfo->data_product->beginning_date,
                      user_block_str)) < 0) {
        fprintf(stderr, "Failed to set user block value for AggregateBeginningDate attribute.\n" );
        goto done;
    }

    sprintf(attr_val, "%llu", g_products[product_index].ublockinfo->data_product->beginning_orbit);
    if ((status = add_userblock_element("AggregateBeginningOrbitNumber", attr_val,
                                         user_block_str)) < 0) {
        fprintf(stderr, "Failed to set user block value for AggregateBeginningOrbitNumber attribute.\n");
        goto done;
    }
     
    if ((status = add_userblock_element("AggregateBeginningTime",
                          g_products[product_index].ublockinfo->data_product->beginning_time,
                          user_block_str)) < 0) {
        fprintf(stderr, "Failed to set user block value for AggregateBeginningTime attribute.\n");
        goto done;
    }
   
    if ((status = add_userblock_element("AggregateEndingDate", 
                      g_products[product_index].ublockinfo->data_product->ending_date,
                      user_block_str)) < 0) {
        fprintf(stderr, "Failed to set user block value for AggregateEndingDate attribute.\n");
        goto done;
    }

    sprintf(attr_val, "%llu", g_products[product_index].ublockinfo->data_product->ending_orbit); 
    if ((status = add_userblock_element("AggregateEndingOrbitNumber",
                                         attr_val, user_block_str)) < 0) {
        fprintf(stderr, "Failed to set user block value for AggregateEndingOrbitNumber attribute.\n");
        goto done;
    }

    if ((status = add_userblock_element("AggregateEndingTime",
                      g_products[product_index].ublockinfo->data_product->ending_time,
                      user_block_str)) < 0) {
        fprintf(stderr, "Failed to set user block value for AggregateEndingTime attribute.\n");
        goto done;
    }
    
    if ((status = add_userblock_element("AggregateBeginningGranuleID",
                          g_products[product_index].ublockinfo->data_product->beginning_granule_id,
                          user_block_str)) < 0) {
        fprintf(stderr, "Failed to set user block value for AggregateBeginningGranuleID attribute.\n"); 
        goto done;
    }

    if ((status = add_userblock_element("AggregateEndingGranuleID", 
                          g_products[product_index].ublockinfo->data_product->ending_granule_id,
                          user_block_str)) < 0) {
        fprintf(stderr, "Failed to set user block value for AggregateEndingGranuleID attribute.\n");
        goto done;
    }
    strcat(user_block_str, "</Data_Product>");

done:
    return status;

}

/*-------------------------------------------------------------------------
 * Function:   write_userblocks_and_close
 *
 * Purpose:    write the userblocks for all of the output files in the 
 *             aggregation, close the files and free the allocated memory. 
 *             
 * Note:       
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             November 15, 2013
 *
 *-------------------------------------------------------------------------
 */
int
write_userblocks_and_close(int files_to_close)
{ 
    int status = SUCCEED;
    int i, j;
    hbool_t last_file = FALSE;
    char *user_block_str = NULL;

    /* If files_to_close is 1, the Data_Product elements for all products will 
       be written to the same userblock and file.  The file information should 
       be written for the first product only. If files_to_close is more than
       1, there should be 1 product in each file, and the file information will
       need writing for each product. */
    if (files_to_close < 1) {
        NAGG_ERROR("write_userblocks_and_close(): There are no files to be closed.\n");
        status = FAIL;
        goto done;
    }

    for (i=0;i<files_to_close;++i) {
        user_block_str = (char *) HDmalloc (g_products[i].ublockinfo->size + 1); 
        FILE *fp;
        ssize_t ocount = 0, numchars;
        hid_t open_objs[3] = {-1, -1, -1};
        char oname[256];
        
        HDstrcpy(user_block_str, "<HDF_UserBlock>");

        /* i is a 0 base index, so the last file is the number of files - 1 */
        if (i == g_numproducts-1)
            last_file = TRUE;
        /* if only one file, add Data_Product elements for all products to the
           userblock string, then write it and close. */
        if (files_to_close == 1) { 

            if ((status = add_fileinfopart_to_ub_string(g_numproducts, last_file, 
                                                        user_block_str)) < 0) {
                goto done;
            }  
            for (j=0; j<g_numproducts; ++j) {
                if ((status = add_dataproduct_to_ub_string(j, user_block_str)) < 0) {
                    goto done;   
                }
            }

        } else {
            if ((status = add_fileinfopart_to_ub_string(1, last_file,
                                                        user_block_str)) < 0) {
                goto done;
            }
            if ((status = add_dataproduct_to_ub_string(i, user_block_str)) < 0) {
            goto done;
            }
        } 

        HDstrcat(user_block_str, "</HDF_UserBlock>");
        if(g_products[i].outfile > -1) {
#ifdef DEBUG
            ocount = H5Fget_obj_count(g_products[i].outfile, H5F_OBJ_DATASET);
            printf("There are %d open datasets before the close.\n", (int)ocount);
            ocount = H5Fget_obj_count(g_products[i].outfile, H5F_OBJ_GROUP);
            printf("There are %d open groups before the close.\n", (int)ocount);
            if(H5Fget_obj_ids(g_products[i].outfile, H5F_OBJ_GROUP, (size_t)3,
                              open_objs) < 0) printf("Error getting open groups.\n");
            for (j=0;j<ocount; ++j){
                numchars = H5Iget_name(open_objs[j], oname, 255);
                printf("%s is open.\n", oname);
            }
            ocount = H5Fget_obj_count(g_products[i].outfile, H5F_OBJ_DATATYPE);
            printf("There are %d open datatypes before the close.\n", (int)ocount);
            ocount = H5Fget_obj_count(g_products[i].outfile, H5F_OBJ_ATTR);
            printf("There are %d open attributes before the close.\n", (int)ocount);
            if(H5Fget_obj_ids(g_products[i].outfile, H5F_OBJ_ATTR, (size_t)3,
                              open_objs) < 0) printf("Error getting open attributes.\n");
            for (j=0;j<ocount; ++j){
                numchars = H5Iget_name(open_objs[j], oname, 255);
                printf("%s is open.\n", oname);
            }
#endif
            if (H5Iis_valid(g_products[i].outfile)) {
                if((H5Fclose(g_products[i].outfile)) < 0) {
                    fprintf(stderr, "Error closing output data file.\n");
                    status = FAIL;
                    goto done;
                }
            }
            g_products[i].outfile = -1;
            if (fp = fopen (g_products[i].outfilename, "r+")) {
                fwrite (user_block_str, 1, strlen(user_block_str), fp);
                fclose (fp);
            } else {
                NAGG_ERROR("write_userblocks_and_close():  Failed to open file ");
                fprintf(stderr, "%s for writing userblock.\n", g_products[i].outfilename);
                goto done;
            }
        }
    }
done:
    if (user_block_str)
        HDfree(user_block_str);
    for (i=0;i<g_numproducts;++i) {
        if (g_products[i].ublockinfo) {
            if (g_products[i].ublockinfo->data_product) {
                HDfree(g_products[i].ublockinfo->data_product);
                g_products[i].ublockinfo->data_product = NULL;
            }
            HDfree(g_products[i].ublockinfo);
            g_products[i].ublockinfo = NULL;
        }
    }
    if (g_products[g_numproducts].ublockinfo) {
        if (g_products[g_numproducts].ublockinfo->data_product) {
            HDfree(g_products[g_numproducts].ublockinfo->data_product);
            g_products[g_numproducts].ublockinfo->data_product = NULL;
        }
        HDfree(g_products[g_numproducts].ublockinfo);
        g_products[g_numproducts].ublockinfo = NULL;
    }
    return status;
    
}

/*-------------------------------------------------------------------------
 * Function:   add_string_attribute
 *
 * Purpose:    Add an attribute of type string to the object identified by the
 *             first parameter.
 *
 * Note:       The attribute will be added as a 1x1 array of strings to match
 *             the files that IDPS delivers.
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             December 15, 2011
 *
 *-------------------------------------------------------------------------
 */
static int
add_string_attribute(hid_t loc_id, const char *attr_name, const char *value) 
{
    herr_t  status;
    hid_t   dataset, datatype, dataspace, dsid;
    hid_t   grpid, spaceid, typeid, attid;
    hsize_t dims[2] = {1,1};
    int     rank = 2;


    if (H5Aexists_by_name(loc_id, ".", attr_name,  H5P_DEFAULT)>0) {
       /* fprintf(stderr, "Note: %s already exists, value will not be written.\n", attr_name); */
        status = 0;
    } else {
        spaceid = H5Screate_simple(rank, dims, dims);
        typeid = H5Tcopy(H5T_C_S1);
        H5Tset_size(typeid, strlen(value) + 1);
        attid = H5Acreate2(loc_id, attr_name, typeid, spaceid,
                           H5P_DEFAULT, H5P_DEFAULT);
        status = H5Awrite(attid, typeid, value);

        if (H5Iis_valid(spaceid))
            H5Sclose(spaceid);
        if (H5Iis_valid(typeid))
            H5Tclose(typeid);
        if (H5Iis_valid(attid))
            H5Aclose(attid);
    }
    return status;
}

/*-------------------------------------------------------------------------
 * Function:   get_string_attribute_value
 *
 * Purpose:    Get the value of a string type attribute.
 *
 * Note:       
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             December 15, 2011
 *
 *-------------------------------------------------------------------------
 */
static int
get_string_attribute_value(hid_t loc_id, const char * name, char *buf)
{
    hid_t   attr;
    hid_t   atype, atype_mem;    /* Attribute type */
    H5T_class_t  type_class;
    int status = SUCCEED;

    if((attr = H5Aopen(loc_id, name, H5P_DEFAULT)) < 0) {
        fprintf(stderr, "Failed to open attribute %s.\n", name);
        status = FAIL;
        goto done;
    } else if(( atype = H5Aget_type(attr))< 0) {
        fprintf(stderr, "Failed to get type for attribute %s.\n", name);
        status = FAIL;
    } else if((type_class = H5Tget_class(atype)) < 0) {
        fprintf(stderr, "Failed to get type class for attribute %s.\n", name);
        status = FAIL;
    } else if ((atype_mem = H5Tget_native_type(atype, H5T_DIR_ASCEND)) < 0) {
        fprintf(stderr, "Failed to get type class for attribute %s.\n", name);
        status = FAIL;
    } else if ((status  = H5Aread(attr, atype_mem, buf)) < 0) {
        fprintf(stderr, "Failed to read attribute %s.\n", name);
    }
done:
    if (H5Iis_valid(atype))
        H5Tclose(atype);
    if (H5Iis_valid(type_class))
        H5Tclose(type_class);
    if (H5Iis_valid(atype_mem))
        H5Tclose(atype_mem);
    if (H5Iis_valid(attr))
        H5Aclose(attr);
    return status;

}

/*-------------------------------------------------------------------------
 * Function:   update_string_attribute 
 *
 * Purpose:    . 
 *
 * Note:       
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             December 15, 2011
 *
 *-------------------------------------------------------------------------
 */
static int
update_string_attribute(hid_t loc_id, const char *attr_name, const char *value)
{
    hid_t attr, atype;
    int status = SUCCEED;

    if((attr = H5Aopen(loc_id, attr_name, H5P_DEFAULT)) < 0) {
        fprintf(stderr, "Failed to open attribute %s.\n", attr_name);
        status = FAIL;
        goto done;
    } else if(( atype = H5Aget_type(attr))< 0) {
        fprintf(stderr, "Failed to get type for attribute %s.\n", attr_name);
        status = FAIL;

    } else if ((status = H5Awrite(attr, atype, value)) < 0) {
        fprintf(stderr, "Failed to write attribute %s.\n", attr_name);
    }

done:
    if (H5Iis_valid(atype))
        H5Tclose(atype);
    if (H5Iis_valid(attr))
        H5Aclose(attr);
    return status;
}

/*-------------------------------------------------------------------------
 * Function:  store_userblock_aggregate_beginning_attributes
 *
 * Purpose:   Store beginning attributes for the aggregation to be written
 *            to the userblock
 *
 * Note:      
 * 
 *            
 *           
 *          
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             November 21, 2013
 *
 *-------------------------------------------------------------------------
 */
static int
store_userblock_beginning_attributes(int product_index, hid_t agdset_id)
{
    int status = SUCCEED;
    hid_t attr;
    char attr_val[NAGG_Numeric_attr_valstring_max];
    uint64_t orbit_number;

    if ((status = get_string_attribute_value(agdset_id, "AggregateBeginningDate", 
                      g_products[product_index].ublockinfo->data_product->beginning_date)) < 0) {
        fprintf(stderr, "Failed to get value of AggregateBeginningDate attribute.\n");
        goto done;
    }


    if((attr = H5Aopen(agdset_id, "AggregateBeginningOrbitNumber", H5P_DEFAULT)) < 0) {
        NAGG_ERROR("store_userblock_beginning_attributes(): Failed to open attribute AggregateBeginningOrbitNumber for reading`\n");
        status = FAIL;
        goto done;
    } else if ((status = H5Aread(attr, H5T_NATIVE_ULLONG, &(g_products[product_index].ublockinfo->data_product->beginning_orbit))) < 0) {
        NAGG_ERROR("store_userblock_beginning_attributes(): Failed to read attribute AggregateBeginningOrbitNumber.\n");
        status = FAIL;
        goto done;
    }

    if ((status = get_string_attribute_value(agdset_id, "AggregateBeginningTime", 
                      g_products[product_index].ublockinfo->data_product->beginning_time)) < 0) {
        fprintf(stderr, "Failed to get value of AggregateBeginningTime attribute.\n");
        goto done;
    }
done:
    
    if (H5Iis_valid(attr)) {
        if(H5Aclose(attr) < 0)
            NAGG_ERROR("store_userblock_beginning_attributes(): Failed to close attribute AggregateBeginningOrbitNumber.\n");
    }
    return status;

}

/*-------------------------------------------------------------------------
 * Function:   store_userblock_ending_attributes
 *
 * Purpose:    Get specified attributes for the end of the aggregation to be
 *             written to the XML userblock.
 *
 * Note:       
 *            
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             November 15, 2013
 *
 *-------------------------------------------------------------------------
 */
static int
store_userblock_ending_attributes(int product_index, hid_t agdset_id)
{
    int status = SUCCEED;
    hid_t attr;
    uint64_t orbit_number;

    
    if ((status = get_string_attribute_value(agdset_id, "AggregateEndingDate", 
                      g_products[product_index].ublockinfo->data_product->ending_date)) < 0) {
        fprintf(stderr, "Failed to get value of AggregateEndingDate attribute.\n");
        goto done;
    }

    if((attr = H5Aopen(agdset_id, "AggregateEndingOrbitNumber", H5P_DEFAULT)) < 0) {
        NAGG_ERROR("store_userblock_ending_attributes(): Failed to open attribute AggregateEndingOrbitNumber for reading`\n");
        status = FAIL;
        goto done;
    } else if ((status = H5Aread(attr, H5T_NATIVE_ULLONG, &(g_products[product_index].ublockinfo->data_product->ending_orbit))) < 0) {
        NAGG_ERROR("store_userblock_ending_attributes(): Failed to read attribute AggregateEndingOrbitNumber.\n");
        status = FAIL;
        goto done;
    }

    if ((status = get_string_attribute_value(agdset_id, "AggregateEndingTime", 
                      g_products[product_index].ublockinfo->data_product->ending_time)) < 0) {
        fprintf(stderr, "Failed to get value of AggregateEndingTime attribute.\n");
        goto done;
    }

    if ((status = get_string_attribute_value(agdset_id, "AggregateBeginningGranuleID", 
                      g_products[product_index].ublockinfo->data_product->beginning_granule_id)) < 0) {
        fprintf(stderr, "Failed to get value of AggregateBeginningGranuleID attribute.\n");
        goto done;
    }

    if ((status = get_string_attribute_value(agdset_id, "AggregateEndingGranuleID", 
                      g_products[product_index].ublockinfo->data_product->ending_granule_id)) < 0) {
        fprintf(stderr, "Failed to get value of AggregateEndingGranuleID attribute.\n");
        goto done;
    }

done:
    if (H5Iis_valid(attr)) {
        if(H5Aclose(attr) < 0)
            NAGG_ERROR("store_userblock_ending_attributes(): Failed to close attribute AggregateEndingOrbitNumber.\n");
    }
    return status;
}

/*-------------------------------------------------------------------------
 * Function:   store_userblock_file_attributes 
 *             
 * Purpose:    Get specified attributes for the output file to be written to
 *             the XML userblock.
 *
 * Note:       
 *            
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             November 15, 2013
 *
 *-------------------------------------------------------------------------
 */
static int
store_userblock_file_attributes(int products_in_file)
{
    int status = SUCCEED;
    hid_t outgroup, attr;
    htri_t result;

    if ((outgroup = H5Gopen(g_products[0].outfile, "/", H5P_DEFAULT)) < 0) {
        fprintf(stderr, "Failed to open output file \"/\"group.\n");
        status = FAIL;
        goto done;
    }

    if ((status = get_string_attribute_value(outgroup, "Mission_Name",
                                        g_products[0].ublockinfo->mission_name)) < 0) {
        fprintf(stderr, "Failed to get value of /Mission_Name attribute.\n");
        goto done;
    }

    if ((status = get_string_attribute_value(outgroup, "Platform_Short_Name", 
                                             g_products[0].ublockinfo->platform_name)) < 0) {
        fprintf(stderr, "Failed to get value of /Platform_Short_Name attribute.\n");
        goto done;
    }

    /* The value for /N_GEO_Ref is in g_products[0].geofilename. */

done:
    g_products[0].ublockinfo->number_data_products = products_in_file;

    if (H5Iis_valid(outgroup)) {
        if (status = H5Gclose(outgroup) < 0) 
            NAGG_ERROR("store_userblock_file_attributes():  Failed to close \"/\" group.\n");
    } 
    return status;
}

/*-------------------------------------------------------------------------
 * Function:  store_userblock_product_attributes
 *
 * Purpose:   Store attributes of the product group to be written
 *            to the userblock
 *
 * Note:
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             November 21, 2013
 *
 *-------------------------------------------------------------------------
 */
static int
store_userblock_product_attributes(int product_index)
{
    int status = SUCCEED;
    hid_t dpgroup;
    char    dpgrpname[NAGG_Group_Path_max];
    char attr_val[NAGG_String_attribute_valstring_max];

    if((status = make_Data_Products_product_group_path(dpgrpname, 
                        g_products[product_index].product_name)) < 0) {
        goto done;
    }

    HDstrcpy(g_products[product_index].ublockinfo->data_product->product_name,
             g_products[product_index].product_name);
    
    if ((dpgroup = H5Gopen(g_products[product_index].outfile, dpgrpname, H5P_DEFAULT)) < 0) {
        fprintf(stderr, "Failed to open group %s.\n", dpgrpname);
        status = FAIL;
        goto done;
    }

    if ((status = get_string_attribute_value(dpgroup, "Instrument_Short_Name", 
                    attr_val)) < 0) {
        fprintf(stderr, "Failed to get value of %s/Instrument_Short_Name attribute.\n", dpgrpname);
        goto done;
    }

    if ((status = get_string_attribute_value(dpgroup, "Instrument_Short_Name", 
                    g_products[product_index].ublockinfo->data_product->instrument_name)) < 0) {
        fprintf(stderr, "Failed to get value of %s/Instrument_Short_Name attribute.\n", dpgrpname);
        goto done;
    }

    if ((status = get_string_attribute_value(dpgroup, "N_Dataset_Type_Tag", 
                    g_products[product_index].ublockinfo->data_product->dataset_type_tag)) < 0) {
        fprintf(stderr, "Failed to get value of %s/N_Dataset_Type_Tag attribute.\n", dpgrpname);
        goto done;
    }

    if ((status = get_string_attribute_value(dpgroup, "N_Processing_Domain", 
                    g_products[product_index].ublockinfo->data_product->processing_domain)) < 0) {
        fprintf(stderr, "Failed to get value of %s/N_Processing_Domain attribute.\n", dpgrpname);
        goto done;
    }

done:
    if (H5Iis_valid(dpgroup)) {
        if (status = H5Gclose(dpgroup) < 0) {
            NAGG_ERROR("store_userblock_product_attributes():  Failed to close ");
            fprintf(stderr, "%s group.\n", dpgrpname);
        }
    }
        return status;
}

/*-------------------------------------------------------------------------
 * Function:  update_aggregate_ending_attributes
 *
 * Purpose:   Write attributes of the Aggregate dataset that are determined by
 *            the last granule in the aggregation.
 *
 * Note:      The attributes "AggregateEndingDate", "AggregateEndingGranuleID",
 *            "AggregateEndingOrbitNumber", and "AggregateEndingTime" can be
 *            copied from the last granule in the aggregation.  The last granule
 *            and the value of the "AggregateNumberGranules" attribute are
 *            determined by the product_gran_idx for the product being aggregated.
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             December 15, 2011
 *
 *-------------------------------------------------------------------------
 */
static int
update_aggregate_ending_attributes(int product_index)
{
    int status;
    char  aggr_dset[NAGG_Dataset_Path_max];
    char  gran_dset[NAGG_Dataset_Path_max];
    int   aggr_gran_num;
    hid_t agdset_id, grdset_id, attr, atype;

    aggr_gran_num = g_gransperoutfile;
    status = make_aggregate_dataset_path(aggr_dset, get_product_name_by_idx(product_index));
    status = make_granule_dataset_path(gran_dset, get_product_name_by_idx(product_index),
                                 get_last_i_granule(product_index));
 
    if((agdset_id = H5Dopen(g_products[product_index].outfile, aggr_dset, H5P_DEFAULT)) < 0 ) {
        NAGG_ERROR("update_aggregate_ending_attributes(): Failed to open output dataset\n"); 
        fprintf(stderr, "Failed to open dataset %s in output file %s for updating ending attributes.\n", aggr_dset, g_products[product_index].outfilename);
        status = FAIL;
        goto done;
    }
    if((grdset_id = H5Dopen(g_products[product_index].outfile, gran_dset, H5P_DEFAULT)) < 0 ) {
        NAGG_ERROR("update_aggregate_ending_attributes(): Failed to open input dataset\n"); 
        fprintf(stderr, "Failed to open dataset %s in output file %s for updating ending attributes.\n", gran_dset, g_products[product_index].outfilename);
        status = FAIL;
        goto done;
    }
 
    if ((status = copy_attribute_value(grdset_id, "Ending_Date", "AggregateEndingDate", agdset_id)) < 0) {
        NAGG_ERROR("update_aggregate_ending_attributes(): Failed to copy value to AggregateEndingDate\n");
        goto done;
    }
    if ((status = copy_attribute_value(grdset_id, "N_Granule_ID", "AggregateEndingGranuleID", agdset_id)) < 0) {
        NAGG_ERROR("update_aggregate_ending_attributes(): Failed to copy value to AggregateEndingGranuleID\n");
        goto done;
    }
    if ((status = copy_attribute_value(grdset_id, "N_Beginning_Orbit_Number", "AggregateEndingOrbitNumber", agdset_id)) < 0) {
        NAGG_ERROR("update_aggregate_ending_attributes(): Failed to copy value to AggregateEndingOrbitNumber\n");
        goto done;
    }
    if ((status = copy_attribute_value(grdset_id, "Ending_Time", "AggregateEndingTime", agdset_id)) < 0) {
        NAGG_ERROR("update_aggregate_ending_attributes(): Failed to copy value to AggregateEndingTime\n");
        goto done;
    }
    if((attr = H5Aopen(agdset_id, "AggregateNumberGranules", H5P_DEFAULT)) < 0) {
        NAGG_ERROR("update_aggregate_ending_attributes(): Failed to open attribute AggregateNumberGranules for writing`\n"); 
        status = FAIL;
        goto done;
    } else if ((status = H5Awrite(attr, H5T_NATIVE_INT, &g_gransperoutfile)) < 0) {
        NAGG_ERROR("update_aggregate_ending_attributes(): Failed to write attribute AggregateNumberGranules.\n");
        status = FAIL;
        goto done;
    }

    /* They should also be written to the user block.  I think the simplest way 
     * to get them now is to read the attributes. 
     */
    if((status=store_userblock_beginning_attributes(product_index, agdset_id)) < 0) {
        status = FAIL;
        goto done;
    }

    if((status=store_userblock_ending_attributes(product_index, agdset_id)) < 0) {
        status = FAIL;
        goto done;
    }

done:
    if (H5Iis_valid(attr)) 
        H5Aclose(attr);
    if (H5Iis_valid(agdset_id)) 
    H5Dclose(agdset_id);
    if (H5Iis_valid(grdset_id)) 
    H5Dclose(grdset_id);
    return status;
}

/*-------------------------------------------------------------------------
 * Function:   write_dataset_fill_values 
 *
 * Purpose:    write fill values to the specified hyperslab of the dataset.
 *
 * Note:       
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             December 15, 2011
 *
 *-------------------------------------------------------------------------
 */
static int
write_dataset_fill_values(hid_t output_file, const char *dsetpath, 
                          const hsize_t *block_coord, int rank)
{
    int         i, status = SUCCEED;
    int         skip = 0;
    hsize_t     numelems;
    hsize_t     dims[1];
    hid_t       dset_id = -1;
    hid_t       dtype, dtype_mem, dspace, dspace_mem;
    size_t      dtype_size;
    hsize_t    *start = NULL,
               *stride = NULL,
               *count = NULL ,
               *block_size = NULL;
    void       *databuf = NULL;
    int8_t           *i8buf = NULL;
    unsigned char    *ui8buf = NULL;
    short int        *i16buf = NULL;
    unsigned short   *ui16buf = NULL;
    int              *i32buf = NULL;
    unsigned int     *ui32buf = NULL; 
    float            *f32buf = NULL;
    long long        *i64buf = NULL;
    double           *f64buf = NULL;

    /* fill values from Common Data Format Control Book - Volume I -
       Overview, Table 3.5.6-1,  p 104 */
    int8_t           i8val = 126;
    unsigned char     ui8val = 254;
    short int        i16val = -998;
    unsigned short   ui16val = 65534;
    int              i32val = -998;
    unsigned long    ui32val = 4294967294UL;
    float            f32val = -999.8;
    long long        i64val = -998;
    double           f64val = -999.8;

    start = (hsize_t *) HDmalloc(rank * sizeof(hsize_t));        
    stride = (hsize_t *) HDmalloc(rank * sizeof(hsize_t));        
    count = (hsize_t *) HDmalloc(rank * sizeof(hsize_t));        
    block_size = (hsize_t *) HDmalloc(rank * sizeof(hsize_t));        
    numelems = 1;

    for (i=0; i<rank; ++i) {
        stride[i] = 1;
        count[i] = 1;
        start[i] = block_coord[i];
        block_size[i] = (block_coord[i + rank] + 1) - block_coord[i];
        numelems *= block_size[i];
    }
    dims[0] = numelems;

    if((dset_id = H5Dopen(output_file, dsetpath, H5P_DEFAULT)) < 0 ) {
        NAGG_ERROR("write_dataset_fill_values(): Failed to open output dataset.\n");
        fprintf(stderr, "Failed to open dataset %s in output file.\n", dsetpath);
        status = FAIL;
        goto done;
    }
    if ((dspace = H5Dget_space(dset_id)) < 0) {
        fprintf(stderr, "Failed to open dataspace for dataset  %s.\n", dsetpath);
        status = FAIL;
        H5Dclose(dset_id);
        goto done;
    }
    if((status = H5Sselect_hyperslab(dspace, H5S_SELECT_SET, start, stride, count, block_size)) < 0) {
        fprintf(stderr, "Failed to select hyperslab for dataset %s.\n", dsetpath);
        H5Dclose(dset_id);
        goto done;
    }
    if ((dtype = H5Dget_type(dset_id)) < 0) {
        fprintf(stderr, "Failed to open datatype for dataset %s.\n", dsetpath);
        status = FAIL;
        H5Dclose(dset_id);
        goto done;
    }
    if ((dtype_mem = H5Tget_native_type(dtype, H5T_DIR_ASCEND)) < 0) {
        fprintf(stderr, "Failed to get type class for dataset %s.\n", dsetpath);
        status = FAIL;
        goto done;
    }
    if ((dtype_size = H5Tget_size(dtype_mem)) == 0) {
        fprintf(stderr, "Failed to get size of datatype for dataset %s.\n", dsetpath);
        status = FAIL;
        H5Dclose(dset_id);
        goto done;
    }
    if ((dspace_mem = H5Screate_simple(1, dims, NULL)) < 0) {
        fprintf(stderr, "Failed to create memory dataspace for dataset %s.\n", dsetpath);
        status = FAIL;
        H5Dclose(dset_id);
        goto done;
    }

    /* allocate memory and fill it with the appropriate fill value for the datatype, then
       write the hyperslab to the dataset. */
    switch(H5Tget_class(dtype)) {    
        case H5T_INTEGER:
            if(H5Tequal(dtype_mem, H5T_NATIVE_UCHAR)) {
                ui8buf = (unsigned char *)HDmalloc(numelems * dtype_size);
                H5Dfill(&ui8val, H5T_NATIVE_UCHAR, ui8buf, H5T_NATIVE_UCHAR, dspace_mem); 
                databuf = ui8buf;
            } else if(H5Tequal(dtype_mem, H5T_NATIVE_CHAR)) {
                i8buf = (int8_t *)HDmalloc(numelems * dtype_size);
                H5Dfill(&i8val, H5T_NATIVE_CHAR, i8buf, H5T_NATIVE_CHAR, dspace_mem); 
                databuf = i8buf;
            } else if(H5Tequal(dtype_mem, H5T_NATIVE_USHORT)) {
                ui16buf = (unsigned short *)HDmalloc(numelems * dtype_size);
                H5Dfill(&ui16val, H5T_NATIVE_USHORT, ui16buf, H5T_NATIVE_USHORT, dspace_mem); 
                databuf = ui16buf;
            } else  if(H5Tequal(dtype_mem, H5T_NATIVE_SHORT)) {
                i16buf = (short int *)HDmalloc(numelems * dtype_size);
                H5Dfill(&i16val, H5T_NATIVE_SHORT, i16buf, H5T_NATIVE_SHORT, dspace_mem); 
                databuf = i16buf;
            } else if(H5Tequal(dtype_mem, H5T_NATIVE_UINT)) {
                ui32buf = (unsigned int *)HDmalloc(numelems * dtype_size);
                H5Dfill(&ui32val, H5T_NATIVE_UINT, ui32buf, H5T_NATIVE_UINT, dspace_mem); 
                databuf = ui32buf;
            } else  if(H5Tequal(dtype_mem, H5T_NATIVE_INT)) {
                i32buf = (int *)HDmalloc(numelems * dtype_size);
                H5Dfill(&i32val, H5T_NATIVE_INT, i32buf, H5T_NATIVE_INT, dspace_mem); 
                databuf = i32buf;
            } else if(H5Tequal(dtype_mem, H5T_NATIVE_LLONG)) {
                i64buf = (long long *)HDmalloc(numelems * dtype_size);
                H5Dfill(&i64val, H5T_NATIVE_LLONG, i64buf, H5T_NATIVE_LLONG, dspace_mem); 
                databuf = i64buf;
            }
            break;
        case H5T_FLOAT:
            if(H5Tequal(dtype_mem, H5T_NATIVE_FLOAT)) {
                f32buf = (float *)HDmalloc(numelems * sizeof(float));
                H5Dfill(&f32val, H5T_NATIVE_FLOAT, f32buf, H5T_NATIVE_FLOAT, dspace_mem); 
                databuf = f32buf;
            } else if(H5Tequal(dtype_mem, H5T_NATIVE_DOUBLE)) {
                f64buf = (double *)HDmalloc(numelems * sizeof(double));
                H5Dfill(&f64val, H5T_NATIVE_DOUBLE, f64buf, H5T_NATIVE_DOUBLE, dspace_mem); 
                databuf = f64buf;
            }
            break;
        default:
            fprintf(stderr, "Unknown type for dataset %s\n", dsetpath);
            status = FAIL;
            goto done;
        break;
    }
    if((status = H5Dwrite(dset_id, dtype_mem, dspace_mem, dspace, H5P_DEFAULT, databuf)) < 0) {
        NAGG_ERROR("write_dataset_fill_values():  Failed to write fill value hyperslab\n");
        goto done;
    }
 
    
done:
    
    if (start) HDfree(start);
    if (stride) HDfree(stride);
    if (count) HDfree(count);
    if (block_size) HDfree(block_size);
    if (databuf) HDfree(databuf);

    if(H5Iis_valid(dtype_mem))
        H5Tclose(dtype_mem);
    if(H5Iis_valid(dtype))
        H5Tclose(dtype);
    if(H5Iis_valid(dspace_mem))
        H5Sclose(dspace_mem);
    if(H5Iis_valid(dspace))
        H5Sclose(dspace);
    if(H5Iis_valid(dset_id))
         H5Dclose(dset_id);

    return status;    
}

/*-------------------------------------------------------------------------
 * Function:   attrit_op_func
 *
 * Purpose:    Copy an attribute from an input file to an output file.  For use
 *             with H5Aiterate2.
 *
 * Note:       
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             December 15, 2011
 *
 *-------------------------------------------------------------------------
 */
static int
attrit_op_func(hid_t loc_id, const char *attr_name, const H5A_info_t *ainfo, void *op_data)
{
    hid_t attr_in, attr_out, atype, aspace, acpl;
    void *buf = 0;
    hsize_t size = 0;
    int status = SUCCEED;

    if (HDstreq(attr_name, "N_GEO_Ref")) {
        /* skip N_GEO_Ref:  will be written later if needed */
        goto done;
    } 
    if (H5Aexists_by_name(*(hid_t *)op_data, ".", attr_name,  H5P_DEFAULT)>0) {
        /* fprintf(stderr, "Note: %s already exists, value will not be written.\n", attr_name); */
        status = 0;
        goto done;
    } else if((attr_in = H5Aopen(loc_id, attr_name, H5P_DEFAULT)) < 0) {
        fprintf(stderr, "Failed to open attribute %s.\n", attr_name);
        status = FAIL;
        goto done;
    } else if(( atype = H5Aget_type(attr_in))< 0) {
        fprintf(stderr, "Failed to get type for attribute %s.\n", attr_name);
        status = FAIL;
    } else if(( aspace = H5Aget_space(attr_in))< 0) {
        fprintf(stderr, "Failed to get dataspace for attribute %s.\n", attr_name);
        status = FAIL;
    }  else if((acpl = H5Aget_create_plist(attr_in)) < 0) {
        fprintf(stderr, "Failed to get creation property list for attribute %s.\n", attr_name);
        status = FAIL;
    }  else if((size = H5Aget_storage_size(attr_in)) == 0) {
        fprintf(stderr, "Failed to get storage size for attribute %s.\n", attr_name);
        status = FAIL;
    } else {
        /* compute and allocate memory for copy buffer */
        buf = HDmalloc (size);

        if ((status = H5Aread(attr_in, atype, buf)) < 0) {
                fprintf(stderr, "Failed to read attribute %s.\n", attr_name);
                status = FAIL;
                goto done;
            } else if ((attr_out = H5Acreate2(*(hid_t *)op_data, attr_name, atype, aspace,
                           acpl, H5P_DEFAULT)) < 0) {
                fprintf(stderr, "Failed to create attribute %s.\n", attr_name);
                status = FAIL;
                goto done;
            } else if ((status = H5Awrite(attr_out, atype, buf)) < 0) {
                fprintf(stderr, "Failed to write attribute %s.\n", attr_name);
                status = FAIL;
            }

            H5Pclose(acpl);
            H5Sclose(aspace);
            H5Tclose(atype);
            H5Aclose(attr_out);
            if (buf)
                HDfree(buf);
            buf = 0;
    }
        
    if(H5Iis_valid(attr_in))
    H5Aclose(attr_in);
done:
    if (buf)
        HDfree(buf);
    return status;
}

/*-------------------------------------------------------------------------
 * Function:   write_fillgr_time_attrs
 *
 * Purpose:    Write time attributes to a fill granule.
 *
 * Note:       
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             February 14, 2012
 *
 *-------------------------------------------------------------------------
 */
static int 
write_fillgr_time_attrs(hid_t dset_id, granule_p_t granule)
{
    hid_t attr, atype, aspace, acpl;
    hsize_t size;
    int status = SUCCEED;
    unsigned long long ietimestamp;
    char begin_hour[3], end_hour[3];
    char    ending_date[NAGG_DATE_size+1];
    char *UTC;

    ietimestamp = granule->granule_start_time_IET;
    if ((attr = H5Aopen(dset_id, "N_Beginning_Time_IET",  H5P_DEFAULT)) < 0) {
        NAGG_ERROR("write_fillgr_time_attrs(): Failed to open attribute N_Beginning_Time_IET for writing\n");
        status = FAIL;
        goto done;
    } else if ((status = H5Awrite(attr, H5T_NATIVE_ULLONG, &ietimestamp)) < 0) {
        fprintf(stderr, "Failed to write attribute N_Beginning_Time_IET.\n");
        status = FAIL;
    }
    H5Aclose(attr);

    ietimestamp = granule->granule_end_time_IET;
    if ((attr = H5Aopen(dset_id, "N_Ending_Time_IET",  H5P_DEFAULT)) < 0) {
        NAGG_ERROR("write_fillgr_time_attrs(): Failed to open attribute N_Ending_Time_IET for writing\n");
        status = FAIL;
        goto done;
    } else if ((status = H5Awrite(attr, H5T_NATIVE_ULLONG, &ietimestamp)) < 0) {
        fprintf(stderr, "Failed to write attribute N_Beginning_Time_IET.\n");
        status = FAIL;
    }
    H5Aclose(attr);

    if ((status = update_string_attribute(dset_id, "Beginning_Date", 
                                       granule->beginning_date)) < 0) {
        NAGG_ERROR("write_fillgr_time_attrs(): Failed to add attribute Beginning_Date.\n");
        goto done;
    }
    
    if ((status = update_string_attribute(dset_id, "Beginning_Time", 
                                       granule->beginning_time)) < 0) {
        NAGG_ERROR("write_fillgr_time_attrs(): Failed to add attribute Beginning_Date.\n");
        goto done;
    }
    
    if ((status = update_string_attribute(dset_id, "Ending_Time", 
                                       granule->ending_time)) < 0) {
        NAGG_ERROR("write_fillgr_time_attrs(): Failed to add attribute Beginning_Date.\n");
        goto done;
    }
    
    /* write "Ending_Date" attribute.  The ending date will be the same as the
       beginning date unless the granule straddles midnight, in which case the 
       Ending_Time (00) hour will be less than the Beginning_Time hour (23).

       Beginning_Time and Ending_Time are strings with the format HHMMSS.mmmmmmZ. */

    HDstrncpy(begin_hour, granule->beginning_time, 2);
    begin_hour[2] = '\0';
    HDstrncpy(end_hour, granule->ending_time, 2);
    end_hour[2] = '\0';

    if(atoi(begin_hour) > atoi(end_hour)) {
        /* Get Ending_Date from end_time_IET */
        UTC =  iet2utc(granule->granule_end_time_IET);
        if (NULL==UTC){
            NAGG_ERROR("write_fillgr_time_attrs(): Failed to convert end_time to ending time");
            status = FAIL;
            goto done;
        }
        sprintf(ending_date, "%8.8s", UTC); 
        if ((status = update_string_attribute(dset_id, "Ending_Date",
                                       ending_date)) < 0) {
            NAGG_ERROR("write_fillgr_time_attrs(): Failed to write attribute Ending_Date.\n");
            goto done;
        }
    } else {
        /* Ending_Date is the same as Beginning_Date */
        if ((status = update_string_attribute(dset_id, "Ending_Date",
                                       granule->beginning_date)) < 0) {
            NAGG_ERROR("write_fillgr_time_attrs(): Failed to write attribute Beginning_Date.\n");
            goto done;
        } 
    }

done:
    return status;
}

/*-------------------------------------------------------------------------
 * Function:   write_fill_granule_attrs
 *
 * Purpose:    Functor to write attributes of an object found by H5Aiterate2.
 *
 * Note:
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             December 15, 2011
 *
 *-------------------------------------------------------------------------
 */
static int
write_fill_granule_attrs(hid_t loc_id, const char *attr_name, const H5A_info_t *ainfo, void *op_data)
{
    hid_t attr_in, attr_out, atype, atype_mem,  aspace, acpl;
    void *buf;
    void *sbuf = NULL;
    hsize_t size;
    hssize_t attrsize;
    int status = SUCCEED;
    int rank = 2;
    hsize_t dims[2] = {1,1};
    /* fill values from Common Data Format Control Book - External Volume V -
       Metadata, Table 4.4.1-1,  p 79 */
    unsigned short ucval = 249;
    unsigned int uival = 65529;
    unsigned long long ullval = 993;
    int ival = -993;
    float fval = -999.3;
    
    if (H5Aexists_by_name(*(hid_t *)op_data, ".", attr_name,  H5P_DEFAULT)>0) {
     /* fprintf(stderr, "Note: %s already exists, value will not be written.\n", attr_name); */
        status = 0;
        goto done;
    } else if((attr_in = H5Aopen(loc_id, attr_name, H5P_DEFAULT)) < 0) {
        fprintf(stderr, "Failed to open attribute %s.\n", attr_name);
        status = FAIL;
        goto done;
    } else if(( atype = H5Aget_type(attr_in))< 0) {
        fprintf(stderr, "Failed to get type for attribute %s.\n", attr_name);
        status = FAIL;
    } else if ((atype_mem = H5Tget_native_type(atype, H5T_DIR_ASCEND)) < 0) {
        fprintf(stderr, "Failed to get type class for attribute %s.\n", attr_name);
        
    } else if(( aspace = H5Screate_simple(rank, dims, NULL))< 0) {
        fprintf(stderr, "Failed to get dataspace for attribute %s.\n", attr_name);
        status = FAIL;
    }  else if((acpl = H5Aget_create_plist(attr_in)) < 0) {
        fprintf(stderr, "Failed to get creation property list for attribute %s.\n", attr_name);
        status = FAIL;
    }  else if((size = H5Aget_storage_size(attr_in)) == 0) {
        fprintf(stderr, "Failed to get storage size for attribute %s.\n", attr_name);
        status = FAIL;
    } else {
        attrsize = H5Sget_simple_extent_npoints(aspace);
 
        if ((attr_out = H5Acreate2(*(hid_t *)op_data, attr_name, atype, aspace,
                       acpl, H5P_DEFAULT)) < 0) {
            fprintf(stderr, "Failed to create attribute %s.\n", attr_name);
            status = FAIL;
            goto done;
        }

        switch(H5Tget_class(atype)) {    
            case H5T_INTEGER:
                if(H5Tequal(atype, H5T_STD_U8BE) == TRUE || 
                   H5Tequal(atype, H5T_STD_U8LE) == TRUE) {
                    buf = &ucval;
                } else if(H5Tequal(atype, H5T_STD_U64BE) == TRUE ||
                          H5Tequal(atype, H5T_STD_U64LE) == TRUE) {
                    buf = &ullval;
                } else if(H5Tequal(atype, H5T_STD_U32BE) == TRUE ||
                        H5Tequal(atype, H5T_STD_U32LE) == TRUE) {
                    buf = &uival;
                } else {
                    buf = &ival;
                }
                break;
            case H5T_FLOAT:
                buf = &fval;
                break;
            case H5T_STRING:
                /* compute and allocate memory for copy buffer */
                sbuf = HDmalloc (size);
                sprintf(sbuf, "%s", "N/A");
                buf = sbuf;
                break;
            default:
                fprintf(stderr, "Unknown type for attribute %s\n", attr_name);
                status = FAIL;
                goto done;
            break;
        }
        
        if ((status = H5Awrite(attr_out, atype_mem, buf)) < 0) {
            fprintf(stderr, "Failed to write attribute %s.\n", attr_name);
            status = FAIL;
        }

        H5Pclose(acpl);
        H5Sclose(aspace);
        H5Tclose(atype);
        H5Aclose(attr_out);
    }
        
    H5Aclose(attr_in);
done:
    if (sbuf)
        HDfree(sbuf);
    return status;
}
 
/*-------------------------------------------------------------------------
 * Function:  copy_attribute_value 
 *
 * Purpose:   Functor to copy attributes of an object found be H5Aiterate2. 
 *
 * Note:       
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             December 15, 2011
 *
 *-------------------------------------------------------------------------
 */
static int
copy_attribute_value(hid_t in_loc_id, const char *attr_name_in, const char *attr_name_out, hid_t out_loc_id)
{
    hid_t attr_in, attr_out, atype;
    void *buf = NULL;
    hsize_t size;
    int status;

    if((attr_in = H5Aopen(in_loc_id, attr_name_in, H5P_DEFAULT)) < 0) {
        fprintf(stderr, "Failed to open attribute %s.\n", attr_name_in);
        status = FAIL;
        goto done;
    } else if(( atype = H5Aget_type(attr_in))< 0) {
        fprintf(stderr, "Failed to get type for attribute %s.\n", attr_name_in);
        status = FAIL;
    }  else if((size = H5Aget_storage_size(attr_in)) == 0) {
        fprintf(stderr, "Failed to get storage size for attribute %s.\n", attr_name_in);
        status = FAIL;
    } else {
        /* compute and allocate memory for copy buffer */
        buf = HDmalloc (size);

        if ((status = H5Aread(attr_in, atype, buf)) < 0) {
            fprintf(stderr, "Failed to read attribute %s.\n", attr_name_in);
            status = FAIL;
            goto done;
        } else if ((attr_out = H5Aopen(out_loc_id, attr_name_out, H5P_DEFAULT)) < 0) {
            fprintf(stderr, "Failed to open attribute %s.\n", attr_name_out);
            status = FAIL;
            goto done;
        } else if ((status = H5Awrite(attr_out, atype, buf)) < 0) {
            fprintf(stderr, "Failed to write attribute %s.\n", attr_name_out);
        }
    }

done:
    if (H5Iis_valid(atype))
        H5Tclose(atype);
    if (H5Iis_valid(attr_in))
        H5Aclose(attr_in);
    if (H5Iis_valid(attr_out))
        H5Aclose(attr_out);
    if (buf)
        HDfree(buf);
    return status;
}

/*-------------------------------------------------------------------------
 * Function:   print_links
 *
 * Purpose:    Debug function to print names of objects in a group.
 *
 * Note:       
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             December 15, 2011
 *
 *-------------------------------------------------------------------------
 */
static int 
print_links(hid_t group_id, const char *name, const H5L_info_t *info, void *op_data) 
{
    int     status;
    H5O_info_t infobuf;

    if((status = H5Oget_info_by_name (group_id, name, &infobuf, H5P_DEFAULT)) < 0) {
        fprintf(stderr, "Failed to get info for %s.\n", name);
        goto done;
    }
    switch (infobuf.type) {
        case H5O_TYPE_GROUP:
            printf ("  Group: %s\n", name);
            break;
        case H5O_TYPE_DATASET:
            printf ("  Dataset: %s\n", name);
            break;
        case H5O_TYPE_NAMED_DATATYPE:
            printf ("  Datatype: %s\n", name);
            break;
        default:
            printf ( "  Unknown: %s\n", name);
    }
done:
    return status;
}

/* This function is copied from nagg_select_granules.c.  If we don't add 
   ending_date to the granule_t structure, the function should probably be
   made public there or moved to a common utility functions file. */

/* Convert IET time (in microsecons) to UTC format.
 * Return value:
 *    If succeeded, return a char string that points to the UTC format;
 *    otherwise, return NULL.
 * Note that the return value points to a statically allocated string which
 * might be overwritten by subsequent calls to iet2utc.
 */
#define IET_UNIX_diff   378691234L;     /* Difference in seconds between IET */
                                        /* and Unix time */
#define UTC_SIZE        22      /* YYYYMMDD.HHMMSS.ssssss */
#define million_microsec        1000000     /* one million microsec = 1 sec */
static char utc_string[UTC_SIZE+1];
static char *iet2utc(iet_t iet_time)
{
    time_t unixtime;
    struct tm *t;

    unixtime = iet_time/million_microsec - IET_UNIX_diff;
    t = gmtime(&unixtime);
    sprintf(utc_string, "%4d%02d%02d.%02d%02d%02d.%06llu",
        t->tm_year+1900, t->tm_mon+1, t->tm_mday,
        t->tm_hour, t->tm_min, t->tm_sec,
        iet_time-(iet_time/million_microsec)*million_microsec);
#ifdef DEBUG
    printf("utc_string=%s\n", utc_string);
#endif
    return(utc_string);
}
