#! /bin/sh
#
# Copyright by The HDF Group.
# All rights reserved.
#
# This file is part of the HDF NPP Aggregation Tool (nagg). The full
# NPP Aggregation Tool copyright notice, including terms governing use,
# modification, and redistribution, is contained in the file COPYING,
# which can be found at the root of the source code distribution tree.
# If you do not have access to this file, you may request a copy from
# help@hdfgroup.org.
#
#
# Functions in nagg_functions.sh run tests in this file with the following
# expectations:
#    CHECK_OUTPUT and CHECK_FAIL_OUTPUT support these extra check options if
#        entered after the expected output file, before the nagg command options:
#        dump: (no parameters) this option dumps all granules from the output
#              files, appended to stdout and stderr output from the nagg command.
#        h5dump:  syntax: h5dump "<h5dump options>"
#                         "" around the h5dump options separates them from the 
#                         nagg options that follow
#        h5dump_grep:  
#            syntax: h5dump_grep "<h5dump options>" "[<grep options>]" '<grep pattern>'
#                    <grep options> can be omitted, but the "" are required.  
#                    There are no current tests that use grep options, but if
#                    any grep options are desired, see the current tests that use 
#                    h5dump_grep and add the options inside the empty "".
#
# CHECK <nagg commmand parameters>          
#    runs nagg with the parameters listed. Passes only if it returns 0.
# CHECK_OUTPUT <expected output file> <[extra check options] nagg command parameters>    
#    runs nagg with the parameters listed. Passes only if it returns 0 AND
#    produces output that matches that in the <expected output file>
#    which is found in the <nagg source>/test/testfiles directory.
# CHECK_FAIL <nagg commmand parameters>
#    runs nagg with the parameters listed. Passes only if it returns 1.
# CHECK_FAIL_OUTPUT <expected output file> <[extra check options] nagg command parameters>    
#    runs nagg with the parameters listed. Passes only if it returns 1 AND
#    produces output that matches that in the <expected output file>
#    which is found in the <nagg source>/test/testfiles directory.
# CHECK_VERSION tests for expected patterns in the output of "nagg --version".  
#    The function anticipates version number changes, but if the format or text
#    of "nagg --version" is changed the function should be updated to match in
#    nagg_functions.sh
#
# The date and time fields of output files are determined from the data in the 
# files processed, but the creation field of the filenames is determined by the 
# time when the command is run, which cannot be foreknown.  Therefore the 
# creation field in filenames in the expected output files must be replaced by
# cYYYYMMDDHHMMSSssssss.


# Tests for the HDF NPP aggregation tool


# The build (current) directory might be different than the source directory.
if test -z "$srcdir"; then
    srcdir=.
fi

# Include Nagg test functions.
. $srcdir/nagg_functions.sh

# Main body goes here.
dir=$SRC_TOOL_TESTFILES
#======================================#
# Positive tests: Expected to succeed. #
# Exit code is zero.                   #
#======================================#
echo ============================================
echo The following tests are expected to succeed.
echo ============================================
# tests expected to succeed (exit code is 0)

#INT-1.1.1
CHECK -h
#INT-1.1.2
CHECK_OUTPUT usage.txt --help
# Test --debug option before using it for the remaining tests.
# Check it can show correct output.
#INT-1.1.3
CHECK --debug --help

# Check --version
#INT-1.2.1
CHECK_VERSION
#INT-1.2.2
CHECK --version

# Test real data files
# REDRO files each has 16 granules. Test if nagg can deaggregate them.
dfile=$dir/REDRO_npp_d20030125_t0702533_e0711257_b00014_c20111025170507190780_unkn_ada.h5
# Nagg should be able to find the geolocation file. Leave it here for reference.
#geofile=GCRIO_npp_d20030125_t0702533_e0711257_b00014_c20111025170507184515_unkn_ada.h5
#AGG-1.1
CHECK_OUTPUT redro1.txt -t REDRO $dfile
#AGG-1.1.1
CHECK_OUTPUT redro3.txt -t REDRO -n 3 $dfile
# The granule duration of a REDRO product is 31.997 seconds.
# Therefore -A 95 == -n 3 while -A 96 == -n 4.
#AGG-1.4.1
CHECK_OUTPUT redro3.txt -t REDRO -A 95 $dfile
#AGG-1.4.2
CHECK_OUTPUT redro4.txt -t REDRO -A 96 $dfile
# Verify the above two cases by dumping the produced output files.
CHECK_OUTPUT redro1-dump.txt dump -t REDRO $dfile
CHECK_OUTPUT redro3-dump.txt dump -t REDRO -n 3 $dfile

# Test -d option.
# Redundent. "-d ." is same as not specifying it as where files are
# created. Output file names still contain "./" as prefix.
CHECK_OUTPUT redro1-d-redundant.txt -d . -t REDRO $dfile
# Create a temporary directory and send the output files to it.
tmpdir=tmpdir
test -e $tmpdir || mkdir $tmpdir
CHECK_OUTPUT redro1-d.txt -d $tmpdir -t REDRO $dfile
rm -rf $tmpdir

# Redo above but with -S.  Expect to create unpackaged (Simple) files.
CHECK_OUTPUT redroS.txt -S -t REDRO $dfile
CHECK_OUTPUT redro3S.txt -S -t REDRO -n 3 $dfile
# Verify that --debug shows valid external geo filenames and null geo filenames
# that are expected on output files in unpackaged format.
CHECK_OUTPUT redro3S-dump.txt dump -S -t REDRO -n 3 $dfile

# test aggregating REDRO/GCRIO files with duplicate granules of different versions.
dfile=$dir/REDRO_npp_d20120513_t000*_noaa_ops.h5
#AGG-1.2.1
CHECK_OUTPUT redro32.txt -n32 -tREDRO $dfile

# SATMS files each has 1 granule. Test if nagg can aggregate them.
dfile="$dir/SATMS*.h5"
CHECK_OUTPUT satms8.txt -t SATMS -n 8 $dfile

# SATMS, TATMS compatible input granules with GEO GATMO unpackaged input, packaged 
# output. 
dfile="$dir/SATMS_npp_d201204*.h5 $dir/TATMS_npp_d201204*.h5"
CHECK_OUTPUT satms_tatms2.txt -n2 -t SATMS,TATMS $dfile
# Same again but switch the order of the -t list. Should have the same output.
CHECK_OUTPUT satms_tatms2.txt -n2 -t TATMS,SATMS $dfile

# SATMS, TATMS compatible input granules with GEO GATMO unpackaged input, packaged 
# output.  Missing TATMS granules are to be replaced by fill granules.
dfile="$dir/SATMS_npp_d201204*.h5 $dir/TATMS_npp_d20120404*.h5"
##CHECK_OUTPUT satms_tatms2m.txt -n2 -t SATMS,TATMS $dfile
# Verify that fill granules are added accordingly.
#AGG-3.3.1
CHECK_OUTPUT satms_tatms2m-dump.txt dump -n2 -t SATMS,TATMS $dfile

# ICSTT with non-matching geofilename using -g yes - unpackaged output
echo Test -g yes unpackaged
dfile=$dir/ICSTT_npp_d20030125_t0705359_e0713337_b00014_c20111025184111991145_unkn_ada.h5
CHECK -n 5 -t ICSTT -S -g yes $dfile

# Single product SATMS packaged with GEO GATMO - packaged output
echo Test packaged input packaged output
dfile=$dir/GATMO-SATMS_npp_d20030126_t0402304_e0411338_b00029_c20120126220649940102_devl_ops.h5
CHECK -n 7 -t SATMS $dfile

# Same again, but with unpackaged output
echo Test packaged input unpackaged output
dfile=$dir/GATMO-SATMS_npp_d20030126_t0402304_e0411338_b00029_c20120126220649940102_devl_ops.h5
CHECK -n 7 -S -t SATMS $dfile

# 4 compatible unpackaged products without GEO - packaged output 
echo Test 4 compatible products unpackaged input packaged output 
dfile="$dir/SVM04_npp_d20120229_t0530029_e0535433_b01757_c20120229113543649418_noaa_ops.h5 $dir/SVM05_npp_d20120229_t0530029_e0535433_b01757_c20120229113543649746_noaa_ops.h5 $dir/SVM06_npp_d20120229_t0530029_e0535433_b01757_c20120229113543650046_noaa_ops.h5 $dir/SVM09_npp_d20120229_t0530029_e0535433_b01757_c20120229113543650958_noaa_ops.h5"
CHECK -n2 -g no -t SVM04,SVM05,SVM06,SVM09 $dfile

outdir="tmptest-out"
test -e $outdir || mkdir $outdir
# Same again, but with unpackaged output in output directory
#INT-1.3.6
echo Test 4 compatible products unpackaged input unpackaged output
CHECK -n2 -g no -d $outdir -S -t SVM04,SVM05,SVM06,SVM09 $dfile

dfile="$outdir/SVM0*.h5"
echo Test aggregating the previously deaggregated all-fill-value files 
#INT-1.3.6.1
CHECK -n4 --geolocation=0 --directory=$outdir -t SVM04,SVM05,SVM06,SVM09 $dfile
rm -rf $outdir

# ICSTT with unpackaged input to produce unpackaged output without GEO
echo Test -g no unpackaged
dfile=$dir/ICSTT_npp_d20030125_t0705359_e0713337_b00014_c20111025184111991145_unkn_ada.h5
CHECK -n 5 -t ICSTT -g no $dfile

# SVI02,SVI03 no GEO, fixed dimension files, one missing granule
echo Test -g no fixed dimensions
dfile="$dir/SVI02_aqu_d20110624_t1739177_e1740275_b48626_c20110624210756057075_grav_dev.h5 $dir/SVI03_aqu_d20110624_t1739177_e1740275_b48626_c20110624210732114546_grav_dev.h5 $dir/SVI03_aqu_d20110624_t1738068_e1739166_b48626_c20110624210810775985_grav_dev.h5"
CHECK -n2 -tSVI02,SVI03 -g no $dfile 

# GATMO-SATMS packaged input and output without GEO
echo Test -g no packaged
dfile=$dir/GATMO-SATMS_npp_d20030126_t0402304_e0411338_b00029_c20120126220649940102_devl_ops.h5
CHECK -n 5 -t SATMS -g no $dfile

# GEO product alone - -g GCRIO option
echo Test -g "<geoproduct>"
dfile=$dir/GCRIO_npp_d20030125_t0702533_e0711257_b00014_c20111025170507184515_unkn_ada.h5
CHECK -n 8 -g GCRIO $dfile

# Tests to validate --like -l option.
echo "Test just -l (no -n nor -t)"
# REDRO files each has 16 granules. The -l should produce 16 granules output,
# instead of the default 1 granule.
# Nagg should be able to find the geolocation file. Leave it here for reference.
#geofile=GCRIO_npp_d20030125_t0702533_e0711257_b00014_c20111025170507184515_unkn_ada.h5
exfile=$dir/REDRO_npp_d20030125_t0702533_e0711257_b00014_c20111025170507190780_unkn_ada.h5
dfile=$dir/REDRO_npp_d20030125_t0702533_e0711257_b00014_c20111025170507190780_unkn_ada.h5
CHECK_OUTPUT redro16.txt -l $exfile $dfile
# check --like works the same
CHECK_OUTPUT redro16.txt --like $exfile $dfile
# Override -l with -n 3 (-p should not be in)
CHECK_OUTPUT redro3l.txt -l $exfile -n 3 $dfile
# Test that -l with a 16 granule REDRO example file with an N_GEO_Ref attribute aggregates the 
# single granule input REDRO files with unpackaged output.  The ouput is produced in a pair of
# 7 granule files and a pair of 1 granule files because there are only 8 pairs of input granules
# and they straddle a bucket boundary.
dfile=$dir/REDRO_npp_d20120513_t000*.h5
CHECK_OUTPUT redrol16.txt -l $exfile $dfile  
echo "Test --like with multiple products example file"
# test --like form of option with 2 compatible products + GEO, 5 granules of each product in
# de-aggregated input files.  Produces 3 packaged output files.
exfile=$dir/GATMO-SATMS-TATMS_npp_d20120403_t2359123_e0000159_b02250_c20120830155722719166_unkn_xxx.h5
dfile="$dir/SATMS_npp_d2012040*.h5 $dir/TATMS_npp_d2012040*.h5"
CHECK_OUTPUT satms_tatms2.txt --like $exfile $dfile
# Override -l of two products with -t one product
CHECK_OUTPUT satms2.txt -t SATMS -l $exfile $dfile
# Check if -g no(0) works right with -l
CHECK_OUTPUT satms_tatms2-g0.txt -g no -l $exfile $dfile
# Check if -l <example file with no geo> works like -g no (same input and output)  
exfile=$dir/SATMS-TATMS_npp_d20120403_t2359123_e0000159_b02250_c20130226213950691232_unkn_xxx.h5
CHECK_OUTPUT satms_tatms2-g0.txt -l $exfile $dfile
# test that GEO only example file performs GEO only aggregation
exfile=$dir/GATRO_npp_d20130126_t0153459_e0154157_b06465_c20130126080424077327_noaa_ops.h5
dfile=$dir/GATRO_npp_d20130126_t015*.h5
CHECK_OUTPUT gatro1.txt -n4 -l $exfile $dfile

#Test -t with -g
# This is redundant but allowed.  Useful tests will be added when extended terrain corrected
# compatibility has been added
dfile=$dir/REDRO_npp_d20030125_t0702533_e0711257_b00014_c20111025170507190780_unkn_ada.h5
# Only one geoproduct is allowed
CHECK -n 4 -t REDRO -g GCRIO $dfile

#Test extended compatibility with no geolocation
dfile="$dir/SATMR_npp_d20130126_t015*.h5 $dir/ICSTT_npp_d20130126_t015*.h5"
CHECK_OUTPUT icstt-satmr4-g0.txt -n4 -tSATMR,ICSTT -g no $dfile
#Test for default TC GEO  
CHECK_OUTPUT icstt-satmr4.txt -n4 -tICSTT,SATMR $dfile
#Test combination of products with Not TC GEO
dfile="$dir/SATMR_npp_d20130126_t015*.h5 $dir/ICSTT_npp_d20130126_t015*.h5 $dir/GATRO_npp_d20130126_t015*.h5"
CHECK_OUTPUT gatro-icstt-satmr4.txt -n4 -tICSTT,SATMR -g GATRO $dfile

# Test extended geolocation functionality
echo ============================================
echo Test extended geolocation functionality
echo ============================================

# normal TC case (sanity check)
CHECK_OUTPUT gcrio-sanity-check.txt -t ICSTT -g GCRIO -n 8 \
    $dir/ICSTT_npp_d20130126_t0153459_e0154157_b06465_c20130126034916192115_noaa_ops.h5 \
    $dir/ICSTT_npp_d20130126_t0154179_e0154477_b06465_c20130126034948250511_noaa_ops.h5 \
    $dir/ICSTT_npp_d20130126_t0154499_e0155197_b06465_c20130126035020215772_noaa_ops.h5 \
    $dir/ICSTT_npp_d20130126_t0155219_e0155517_b06465_c20130126035052218071_noaa_ops.h5

# normal non-TC case (sanity check)
CHECK_OUTPUT gatro-sanity-check.txt -t SATMR -g GATRO -n 8 \
    $dir/SATMR_npp_d20130126_t0153459_e0154157_b06465_c20130126080936834465_noaa_ops.h5 \
    $dir/SATMR_npp_d20130126_t0154179_e0154477_b06465_c20130126080936834465_noaa_ops.h5 \
    $dir/SATMR_npp_d20130126_t0154499_e0155197_b06465_c20130126080936834465_noaa_ops.h5 \
    $dir/SATMR_npp_d20130126_t0155219_e0155517_b06465_c20130126080936834465_noaa_ops.h5

# 5.2.2: TC products with non-TC geolocation
CHECK_OUTPUT icstt-gatro-ext.txt -t ICSTT -g GATRO -n 5 \
    $dir/ICSTT_npp_d20130126_t0154179_e0154477_b06465_c20130126034948250511_noaa_ops.h5 \
    $dir/ICSTT_npp_d20130126_t0154499_e0155197_b06465_c20130126035020215772_noaa_ops.h5 \
    $dir/GATRO_npp_d20130126_t0154179_e0154477_b06465_c20130126080424077327_noaa_ops.h5

# 5.2.2: non-TC products with TC geolocation
CHECK_OUTPUT satmr-gcrio-ext.txt -t SATMR -g GCRIO -n 5 \
    $dir/SATMR_npp_d20130126_t0153459_e0154157_b06465_c20130126080936834465_noaa_ops.h5 \
    $dir/SATMR_npp_d20130126_t0154179_e0154477_b06465_c20130126080936834465_noaa_ops.h5 \
    $dir/GCRIO_npp_d20130126_t0154179_e0154477_b06465_c20130126034948250830_noaa_ops.h5

# 5.2.5: TC and non-TC products with non-TC geolocation
CHECK_OUTPUT gatro-mixed-ext.txt -t SATMR,ICSTT -g GATRO -n 5 \
    $dir/SATMR_npp_d20130126_t0153459_e0154157_b06465_c20130126080936834465_noaa_ops.h5 \
    $dir/SATMR_npp_d20130126_t0154179_e0154477_b06465_c20130126080936834465_noaa_ops.h5 \
    $dir/ICSTT_npp_d20130126_t0154179_e0154477_b06465_c20130126034948250511_noaa_ops.h5 \
    $dir/ICSTT_npp_d20130126_t0154499_e0155197_b06465_c20130126035020215772_noaa_ops.h5 \
    $dir/GATRO_npp_d20130126_t0154179_e0154477_b06465_c20130126080424077327_noaa_ops.h5

# 5.2.5: TC and non-TC products with TC geolocation
CHECK_OUTPUT gcrio-mixed-ext.txt -t SATMR,ICSTT -g GCRIO -n 5 \
    $dir/SATMR_npp_d20130126_t0153459_e0154157_b06465_c20130126080936834465_noaa_ops.h5 \
    $dir/SATMR_npp_d20130126_t0154179_e0154477_b06465_c20130126080936834465_noaa_ops.h5 \
    $dir/ICSTT_npp_d20130126_t0154179_e0154477_b06465_c20130126034948250511_noaa_ops.h5 \
    $dir/ICSTT_npp_d20130126_t0154499_e0155197_b06465_c20130126035020215772_noaa_ops.h5 \
    $dir/GCRIO_npp_d20130126_t0154179_e0154477_b06465_c20130126034948250830_noaa_ops.h5

# 5.2.4,5.2.6: TC and non-TC products with TC geolocation, S option
CHECK_OUTPUT gcrio-mixed-s-ext.txt -S -t SATMR,ICSTT -g GCRIO -n 5 \
    $dir/SATMR_npp_d20130126_t0153459_e0154157_b06465_c20130126080936834465_noaa_ops.h5 \
    $dir/SATMR_npp_d20130126_t0154179_e0154477_b06465_c20130126080936834465_noaa_ops.h5 \
    $dir/ICSTT_npp_d20130126_t0154179_e0154477_b06465_c20130126034948250511_noaa_ops.h5 \
    $dir/ICSTT_npp_d20130126_t0154499_e0155197_b06465_c20130126035020215772_noaa_ops.h5 \
    $dir/GATRO_npp_d20130126_t0154179_e0154477_b06465_c20130126080424077327_noaa_ops.h5

# 5.2.7: TC and non-TC products with no geolocation specified
CHECK_OUTPUT mixed-nogeo-ext.txt -t ICSTT,SATMR -n 5 -g no \
    $dir/SATMR_npp_d20130126_t0153459_e0154157_b06465_c20130126080936834465_noaa_ops.h5 \
    $dir/SATMR_npp_d20130126_t0154179_e0154477_b06465_c20130126080936834465_noaa_ops.h5 \
    $dir/ICSTT_npp_d20130126_t0154179_e0154477_b06465_c20130126034948250511_noaa_ops.h5 \
    $dir/ICSTT_npp_d20130126_t0154499_e0155197_b06465_c20130126035020215772_noaa_ops.h5 \

# 5.2.4,5.2.7,5.2.8: TC and non-TC products with no geolocation specified
CHECK_OUTPUT mixed-nogeo2-ext.txt -t SATMR,ICSTT -n 5 -g no \
    $dir/SATMR_npp_d20130126_t0153459_e0154157_b06465_c20130126080936834465_noaa_ops.h5 \
    $dir/SATMR_npp_d20130126_t0154179_e0154477_b06465_c20130126080936834465_noaa_ops.h5 \
    $dir/ICSTT_npp_d20130126_t0154179_e0154477_b06465_c20130126034948250511_noaa_ops.h5 \
    $dir/ICSTT_npp_d20130126_t0154499_e0155197_b06465_c20130126035020215772_noaa_ops.h5 \

# 5.2.4,5.2.7,5.2.8: TC and non-TC products with default geolocation (no -g option)
CHECK_OUTPUT mixed-defaultgeo-ext.txt -t SATMR,ICSTT -n 5 \
    $dir/SATMR_npp_d20130126*_noaa_ops.h5 \
    $dir/ICSTT_npp_d20130126*_noaa_ops.h5\

# 5.2.4,5.2.6,5.2.7,5.2.8: TC and non-TC products with default geolocation (no -g option)
# non-TC geolocation is included in input but is correctly ignored for output. 
CHECK_OUTPUT mixed-defaultgeo-ignoreNTCgeo-ext.txt -t SATMR,ICSTT -n 5 -S \
    $dir/SATMR_npp_d20130126*_noaa_ops.h5 \
    $dir/ICSTT_npp_d20130126*_noaa_ops.h5\
    $dir/GATRO_npp_d20130126*_noaa_ops.h5\


# Test --layout, --filter --minimum options. 
echo ============================================
echo Test --layout, --filter --minimum options.
echo ============================================
#test with new option --layout
dfile=$dir/ICSTT*
CHECK -t ICSTT -n 4 --layout CONTI $dfile

# SATMS, TATMS compatible input granules with GEO GATMO unpackaged input, packaged
# output, chunking layout changed on the BeamLatitude and BeamLongitude datasets in 
# the GEO GATMO group.
##dfile="$dir/SATMS_npp_d201204*.h5 $dir/TATMS_npp_d20120404*.h5"
dfile="$dir/SATMS_npp_d201204*.h5 $dir/TATMS_npp_d201204*.h5"
CHECK_OUTPUT satms_tatms2chunktest.txt h5dump "-d /All_Data/ATMS-SDR-GEO_All/BeamLatitude -d /All_Data/ATMS-SDR-GEO_All/BeamLongitude -p -H" --layout=/All_Data/ATMS-SDR-GEO_All/BeamLongitude:CHUNK=24x12x5 --layout=/All_Data/ATMS-SDR-GEO_All/BeamLatitude:CHUNK=12x48x5 -n2 -t SATMS,TATMS $dfile
# Same with -S options (unpackaged)
CHECK_OUTPUT satms_tatms2chunktest_S.txt h5dump "-d /All_Data/ATMS-SDR-GEO_All/BeamLatitude -d /All_Data/ATMS-SDR-GEO_All/BeamLongitude -p -H" --layout /All_Data/ATMS-SDR-GEO_All/BeamLongitude:CHUNK=24x12x5 --layout=/All_Data/ATMS-SDR-GEO_All/BeamLatitude:CHUNK=12x48x5 -n2 -t SATMS,TATMS -S $dfile
# change various small datasets in all 3 products to CONTIGUOUS storage
#The ouput for this test contains an offset for each CONTIGUOUS dataset which can change each time the test is run.  Commented out pending the addition of a grep option to further filter the output.
#CHECK_OUTPUT satms_tatms2conti_test.txt h5dump "-d/All_Data/ATMS-SDR_All/BrightnessTemperatureFactors -d/All_Data/ATMS-SDR_All/InstrumentMode -d/All_Data/ATMS-SDR_All/PadByte1 -d/All_Data/ATMS-SDR_All/QF11_GRAN_QUADRATICCORRECTION -d/All_Data/ATMS-TDR_All/AntennaTemperatureFactors -d/All_Data/ATMS-TDR_All/QF11_GRAN_QUADRATICCORRECTION -d/All_Data/ATMS-SDR-GEO_All/MidTime -d/All_Data/ATMS-SDR-GEO_All/PadByte1 -p -H" --layout=/All_Data/ATMS-SDR_All/BrightnessTemperatureFactors,/All_Data/ATMS-SDR_All/InstrumentMode,/All_Data/ATMS-SDR_All/PadByte1,/All_Data/ATMS-SDR_All/QF11_GRAN_QUADRATICCORRECTION,/All_Data/ATMS-TDR_All/AntennaTemperatureFactors,/All_Data/ATMS-TDR_All/QF11_GRAN_QUADRATICCORRECTION,/All_Data/ATMS-SDR-GEO_All/MidTime,/All_Data/ATMS-SDR-GEO_All/PadByte1:CONTI -n2 -t SATMS,TATMS $dfile
# apply COMPACT storage to all sufficiently small (<65536 bytes) datasets  
# (nagg only applies layout and filter changes to the /All_Data group).
CHECK_OUTPUT satms_tatms2compact_test.txt h5dump "-g/All_Data -p -H" --layout=/All_Data/ATMS-SDR_All/BeamTime,/All_Data/ATMS-SDR_All/BrightnessTemperatureFactors,/All_Data/ATMS-SDR_All/GainCalibration,/All_Data/ATMS-SDR_All/InstrumentMode,/All_Data/ATMS-SDR_All/NEdTCold,/All_Data/ATMS-SDR_All/NEdTWarm,/All_Data/ATMS-SDR_All/PadByte1,/All_Data/ATMS-SDR_All/QF10_GRAN_HEALTHSTATUS,/All_Data/ATMS-SDR_All/QF11_GRAN_QUADRATICCORRECTION,/All_Data/ATMS-SDR_All/QF11_GRAN_QUADRATICCORRECTION,/All_Data/ATMS-SDR_All/QF12_SCAN_KAVPRTCONVERR,/All_Data/ATMS-SDR_All/QF13_SCAN_WGPRTCONVERR,/All_Data/ATMS-SDR_All/QF14_SCAN_SHELFPRTCONVERR,/All_Data/ATMS-SDR_All/QF15_SCAN_KAVPRTTEMPLIMIT,/All_Data/ATMS-SDR_All/QF16_SCAN_WGPRTTEMPLIMIT,/All_Data/ATMS-SDR_All/QF17_SCAN_KAVPRTTEMPCONSISTENCY,/All_Data/ATMS-SDR_All/QF18_SCAN_WGPRTTEMPCONSISTENCY,/All_Data/ATMS-SDR_All/QF19_SCAN_ATMSSDR,/All_Data/ATMS-SDR_All/QF1_GRAN_HEALTHSTATUS,/All_Data/ATMS-SDR_All/QF20_ATMSSDR,/All_Data/ATMS-SDR_All/QF21_ATMSSDR,/All_Data/ATMS-SDR_All/QF22_ATMSSDR,/All_Data/ATMS-SDR_All/QF2_GRAN_HEALTHSTATUS,/All_Data/ATMS-SDR_All/QF3_GRAN_HEALTHSTATUS,/All_Data/ATMS-SDR_All/QF4_GRAN_HEALTHSTATUS,/All_Data/ATMS-SDR_All/QF5_GRAN_HEALTHSTATUS,/All_Data/ATMS-SDR_All/QF6_GRAN_HEALTHSTATUS,/All_Data/ATMS-SDR_All/QF7_GRAN_HEALTHSTATUS,/All_Data/ATMS-SDR_All/QF8_GRAN_HEALTHSTATUS,/All_Data/ATMS-SDR_All/QF9_GRAN_HEALTHSTATUS,/All_Data/ATMS-TDR_All/AntennaTemperatureFactors,/All_Data/ATMS-TDR_All/BeamTime,/All_Data/ATMS-TDR_All/InstrumentMode,/All_Data/ATMS-TDR_All/PadByte11,/All_Data/ATMS-TDR_All/QF10_GRAN_HEALTHSTATUS,/All_Data/ATMS-TDR_All/QF11_GRAN_QUADRATICCORRECTION,/All_Data/ATMS-TDR_All/QF12_SCAN_KAVPRTCONVERR,/All_Data/ATMS-TDR_All/QF13_SCAN_WGPRTCONVERRR,/All_Data/ATMS-TDR_All/QF14_SCAN_SHELFPRTCONVERR,/All_Data/ATMS-TDR_All/QF15_SCAN_KAVPRTTEMPLIMIT,/All_Data/ATMS-TDR_All/QF16_SCAN_WGPRTTEMPLIMIT,/All_Data/ATMS-TDR_All/QF17_SCAN_KAVPRTTEMPCONSISTENCY,/All_Data/ATMS-TDR_All/QF18_SCAN_WGPRTTEMPCONSISTENCY,/All_Data/ATMS-TDR_All/QF19_SCAN_ATMSSDR,/All_Data/ATMS-TDR_All/QF1_GRAN_HEALTHSTATUS,/All_Data/ATMS-TDR_All/QF20_ATMSSDR,/All_Data/ATMS-TDR_All/QF21_ATMSSDR,/All_Data/ATMS-TDR_All/QF22_ATMSSDR,/All_Data/ATMS-TDR_All/QF2_GRAN_HEALTHSTATUS,/All_Data/ATMS-TDR_All/QF3_GRAN_HEALTHSTATUS,/All_Data/ATMS-TDR_All/QF4_GRAN_HEALTHSTATUS,/All_Data/ATMS-TDR_All/QF5_GRAN_HEALTHSTATUS,/All_Data/ATMS-TDR_All/QF6_GRAN_HEALTHSTATUS,/All_Data/ATMS-TDR_All/QF7_GRAN_HEALTHSTATUS,/All_Data/ATMS-TDR_All/QF8_GRAN_HEALTHSTATUS,/All_Data/ATMS-TDR_All/QF9_GRAN_HEALTHSTATUS,/All_Data/ATMS-SDR-GEO_All/Height,/All_Data/ATMS-SDR-GEO_All/Latitude,/All_Data/ATMS-SDR-GEO_All/Longitude,/All_Data/ATMS-SDR-GEO_All/MidTime,/All_Data/ATMS-SDR-GEO_All/PadByte1,/All_Data/ATMS-SDR-GEO_All/QF1_ATMSSDRGEO,/All_Data/ATMS-SDR-GEO_All/SCAttitude,/All_Data/ATMS-SDR-GEO_All/SCPosition,/All_Data/ATMS-SDR-GEO_All/SCVelocity,/All_Data/ATMS-SDR-GEO_All/SatelliteAzimuthAngle,/All_Data/ATMS-SDR-GEO_All/SatelliteRange,/All_Data/ATMS-SDR-GEO_All/SatelliteZenithAngle,All_Data/ATMS-SDR-GEO_All/SolarAzimuthAngle,/All_Data/ATMS-SDR-GEO_All/SolarZenithAngle,/All_Data/ATMS-SDR-GEO_All/StartTime:COMPA --minimum=1 -n2 -t SATMS,TATMS $dfile 
# De-aggregate, unpackage apply new chunk size 12x24x5 where applicable, and SHUFFLE + GZIP=7 filters to GATMO-SATMS file.
dfile="$dir/GATMO-SATMS_npp_d20030126_t0402304_e0411338_b00029_c20120126220649940102_devl_ops.h5"
CHECK_OUTPUT gatmo_satms_chunk_filters.txt  h5dump "-d /All_Data/ATMS-SDR-GEO_All/BeamLatitude -d /All_Data/ATMS-SDR-GEO_All/BeamLongitude -p -H" --layout CHUNK=12x24x5 --filter SHUF --filter GZIP=7 -S -tSATMS $dfile
# Check all filters
dfile="$dir/REDRO_npp_d20030125*.h5"
CHECK_OUTPUT gcrio_redro_shuf_gz_min.txt h5dump "-d/All_Data/CrIMSS-EDR_All/H2O -d/All_Data/CrIMSS-EDR_All/Pressure -d/All_Data/CrIMSS-EDR_All/Temperature -p -H" -n32 -t REDRO --filter SHUF --filter GZIP=7 --minimum 100000 $dfile
dfile="$dir/GCRIO-ICALI-ICALM-ICCCR-SATMR_npp_d20130126_t0153459_e0155517_b06465_c20130301214009160670_unkn_xxx.h5"
CHECK_OUTPUT gcrio-mixed-applynone.txt h5dump "-d/All_Data/ATMS-REMAP-SDR_All/BrightnessTemperature -d/All_Data/CrIMSS-CrIS-AVMP-LOS-IR-IP_All/moistProfile -d/All_Data/CrIMSS-CrIS-AVMP-LOS-MW-IP_All/moistProfile -d/All_Data/CrIMSS-CrIS-CLOUD-CLEARED-RAD-IP_All/radiance -d/All_Data/CrIMSS-EDR-GEO-TC_All/Height -d/All_Data/CrIMSS-EDR-GEO-TC_All/Latitude -d/All_Data/CrIMSS-EDR-GEO-TC_All/Longitude -d/All_Data/CrIMSS-EDR-GEO-TC_All/SatelliteAzimuthAngle -d/All_Data/CrIMSS-EDR-GEO-TC_All/SatelliteRange -d/All_Data/CrIMSS-EDR-GEO-TC_All/SatelliteZenithAngle -d/All_Data/CrIMSS-EDR-GEO-TC_All/SolarAzimuthAngle -d/All_Data/CrIMSS-EDR-GEO-TC_All/SolarZenithAngle -p -H " -n4 -t ICALI,ICALM,ICCCR,SATMR --filter=NONE --minimum=1 $dfile
CHECK_OUTPUT gcrio-mixed-applycompact-3none.txt  h5dump "-d/All_Data/ATMS-REMAP-SDR_All/BrightnessTemperature -d/All_Data/CrIMSS-CrIS-AVMP-LOS-IR-IP_All/moistProfile -d/All_Data/CrIMSS-CrIS-AVMP-LOS-MW-IP_All/moistProfile -d/All_Data/CrIMSS-CrIS-CLOUD-CLEARED-RAD-IP_All/radiance -d/All_Data/CrIMSS-EDR-GEO-TC_All/Height -p -H " -n8 -t ICALI,ICALM,ICCCR,SATMR --layout=COMPA --filter=/All_Data/CrIMSS-CrIS-AVMP-LOS-IR-IP_All/moistProfile,/All_Data/CrIMSS-CrIS-AVMP-LOS-MW-IP_All/moistProfile,/All_Data/CrIMSS-CrIS-CLOUD-CLEARED-RAD-IP_All/radiance:NONE $dfile
#FLET
dfile="$dir/SVI0*.h5"
CHECK_OUTPUT svi023-flet.txt h5dump_grep "-d/All_Data/VIIRS-I2-SDR_All/QF1_VIIRSIBANDSDR -d/All_Data/VIIRS-I2-SDR_All/Radiance -d/All_Data/VIIRS-I3-SDR_All/QF1_VIIRSIBANDSDR -d/All_Data/VIIRS-I3-SDR_All/Radiance -d/All_Data/VIIRS-I3-SDR_All/Reflectance -p -H " "" 'COMPRESSION\|CHUNKED\|FLETCHER32' --filter=FLET --minimum=102400 -n4 -tSVI02,SVI03 -g no $dfile
#NBIT
CHECK_OUTPUT svi023-nbit.txt h5dump_grep "-d/All_Data/VIIRS-I2-SDR_All/QF1_VIIRSIBANDSDR -d/All_Data/VIIRS-I2-SDR_All/Radiance -d/All_Data/VIIRS-I3-SDR_All/QF1_VIIRSIBANDSDR -d/All_Data/VIIRS-I3-SDR_All/Radiance -d/All_Data/VIIRS-I3-SDR_All/Reflectance -p -H " "" 'COMPRESSION\|CHUNKED' --minimum=102400 --filter=NBIT -n4 -tSVI02,SVI03 -g no $dfile
#SOFF
CHECK_OUTPUT svi023-soff.txt h5dump_grep "-d/All_Data/VIIRS-I2-SDR_All/QF1_VIIRSIBANDSDR -d/All_Data/VIIRS-I2-SDR_All/Radiance -d/All_Data/VIIRS-I3-SDR_All/QF1_VIIRSIBANDSDR -d/All_Data/VIIRS-I3-SDR_All/Radiance -d/All_Data/VIIRS-I3-SDR_All/Reflectance -p -H " "" 'COMPRESSION\|CHUNKED' --minimum=102400 --filter=SOFF=2,DS -n4 -tSVI02,SVI03 -g no $dfile
#SZIP - test only if HDF5 was built with szlib
if [ -z "$CC" ]; then
  CC=h5cc
fi
if ($CC -show | grep lsz); then
  CHECK_OUTPUT svi023-szip.txt h5dump_grep "-d/All_Data/VIIRS-I2-SDR_All/QF1_VIIRSIBANDSDR -d/All_Data/VIIRS-I2-SDR_All/Radiance -d/All_Data/VIIRS-I3-SDR_All/QF1_VIIRSIBANDSDR -d/All_Data/VIIRS-I3-SDR_All/Radiance -d/All_Data/VIIRS-I3-SDR_All/Reflectance -p -H " "" 'COMPRESSION\|CHUNKED\|PIXELS\|MODE\|CODING\|BYTE_ORDER\|HEADER' --minimum=102400 --filter=SZIP=32,NN -n4 -tSVI02,SVI03 -g no $dfile
else
  echo "HDF5 was not built with szip; nagg not tested with szip"
fi
#UD - user defined, dynamically loaded compression filters require an external filter 
# and an external filter project.  Because of the setup needed we won't run a test as 
# part of make check.  We could include a commented out test with setup instructions?

dfile="$dir/GCRIO-ICALI-ICALM-ICCCR-SATMR_npp_d20130126_t0153459_e0155517_b06465_c20130301214009160670_unkn_xxx.h5"
# Global COMPACT layout will be applied to all datasets smaller than 64K;  
# minimum is disregarded for COMPACT layout
CHECK_OUTPUT gcrio-mixed-applycompact-all.txt  h5dump_grep "-p -H" "" 'COMPRESSION\|COMPACT' --layout=COMPA -n8 -tICALI,ICALM,ICCCR,SATMR $dfile

# 1. applying COMPACT or CONTIGUOUS layout to a compressed dataset
#    removes compression
# 2. COMPACT and CONTIGUOUS layouts and compression filters can be applied
#    together if no common datasets are specified.
CHECK_OUTPUT gcrio-mixed-applycompactcontiguous.txt h5dump_grep "-p -H" "" 'COMPRESSION\|COMPACT\|CONTIGUOUS' --layout=/All_Data/CrIMSS-EDR-GEO-TC_All/Height,/All_Data/CrIMSS-EDR-GEO-TC_All/Latitude,/All_Data/CrIMSS-EDR-GEO-TC_All/Longitude,/All_Data/CrIMSS-EDR-GEO-TC_All/SatelliteAzimuthAngle,/All_Data/CrIMSS-EDR-GEO-TC_All/SatelliteRange,/All_Data/CrIMSS-EDR-GEO-TC_All/SatelliteZenithAngle,/All_Data/CrIMSS-EDR-GEO-TC_All/SolarAzimuthAngle,/All_Data/CrIMSS-EDR-GEO-TC_All/SolarZenithAngle:COMPA --layout=/All_Data/ATMS-REMAP-SDR_All/BrightnessTemperature,/All_Data/CrIMSS-CrIS-AVMP-LOS-IR-IP_All/moistProfile,/All_Data/CrIMSS-CrIS-AVMP-LOS-MW-IP_All/moistProfile,/All_Data/CrIMSS-CrIS-CLOUD-CLEARED-RAD-IP_All/radiance:CONTIGUOUS -n8 -tICALI,ICALM,ICCCR,SATMR $dfile

# 1. COMPACT and CONTIGUOUS layouts can be applied with --filter NONE
# 2. --layout and --filter that specify specific datasets will take 
#    precedence over --minimum
CHECK_OUTPUT gcrio-mixed-applyccNONEmin.txt h5dump_grep "-p -H" "" 'COMPRESSION\|COMPACT\|CONTIGUOUS' --layout=/All_Data/CrIMSS-EDR-GEO-TC_All/Height,/All_Data/CrIMSS-EDR-GEO-TC_All/Latitude,/All_Data/CrIMSS-EDR-GEO-TC_All/Longitude,/All_Data/CrIMSS-EDR-GEO-TC_All/SatelliteAzimuthAngle:COMPA --layout=/All_Data/CrIMSS-EDR-GEO-TC_All/SatelliteRange,/All_Data/CrIMSS-EDR-GEO-TC_All/SatelliteZenithAngle,/All_Data/CrIMSS-EDR-GEO-TC_All/SolarAzimuthAngle,/All_Data/CrIMSS-EDR-GEO-TC_All/SolarZenithAngle,/All_Data/ATMS-REMAP-SDR_All/BrightnessTemperature,/All_Data/CrIMSS-CrIS-AVMP-LOS-IR-IP_All/moistProfile,/All_Data/CrIMSS-CrIS-AVMP-LOS-MW-IP_All/moistProfile,/All_Data/CrIMSS-CrIS-CLOUD-CLEARED-RAD-IP_All/radiance:CONTI --filter=NONE --minimum 125000 -n8 -tICALI,ICALM,ICCCR,SATMR $dfile

# Check that processing a file with missing region references will create them.
dfile=$dir/SVM01_npp_d20120423_t2359080_e0016117_b02534_c20140301163325890356_XXXX_XXX.h5
CHECK_OUTPUT svm01-repaired.txt h5dump_grep "-d/Data_Products/VIIRS-M1-SDR/VIIRS-M1-SDR_Gran_0 -d/Data_Products/VIIRS-M1-SDR/VIIRS-M1-SDR_Gran_1 -d/Data_Products/VIIRS-M1-SDR/VIIRS-M1-SDR_Gran_2 -d/Data_Products/VIIRS-M1-SDR/VIIRS-M1-SDR_Gran_3" "" 'DATASET' -n 12 -t SVM01 -g no $dfile

# This file has less granules than its Aggr dataset promises.
dfile="$dir/SVM01_npp_d20120423_t2359080_e0001576_b02534_c20140305142028894758_XXXX_XXX.h5"
CHECK -tSVM01 -g no $dfile
# This test previously failed, but nagg behavior has changed to aggregate the actual number
# of granules and print a warning to stderr.  Currently there is no test function to 
# handle a nagg operation that returns 0 with output to both stdout and stderr.
#CHECK_FAIL_OUTPUT toofewgranules.txt -tSVM01 -g no $dfile

#===================================#
# Negative tests: Expected to fail. #
# Exit code is non-zero.            #
#===================================#
echo =========================================
echo The following tests are expected to fail.
echo =========================================
# unknown option
CHECK_FAIL_OUTPUT usage-bad-option.txt -x
# Non-positive -n number of granules
CHECK_FAIL -n 0
# Missing input file argument
CHECK_FAIL -n 3

echo Test -d option - Errors expected
# missing option parameter
dfile=$dir/REDRO_npp_d20030125_t0702533_e0711257_b00014_c20111025170507190780_unkn_ada.h5
CHECK_FAIL -d -t REDRO $dfile
# output directory does not exist or not writable. Assume /tmpbad does not exist.
dfile=$dir/REDRO_npp_d20030125_t0702533_e0711257_b00014_c20111025170507190780_unkn_ada.h5
CHECK_FAIL -d /tmpbad -t REDRO $dfile
# Verify Nagg can detect if the outdir name is too long for the output file
# names buffer. Each output file name has two parts, Head and Tail.  The Tail
# size is fixed. The Head consists of the outdir, products and geolocation
# product.
# The Header part can hold (DPID_NUM_MAX+1) [31] of DPID_size[5] and "_". That
# is 31*6=186 chars.
# is 42*6=252.  252 - 18 - 1 = 233 
# Outdir length test 1: outdir length is 167, '/', two products and one Geo
# are 18.  Total is 186. Should pass.
dfile="$dir/SATMS_npp_d201204*.h5 $dir/TATMS_npp_d201204*.h5"
outdir=\
123456789a123456789b123456789c123456789d123456789e123456789f123456789g123456789h123456789i123456789j123456789k123456789l123456789l123456789m123456789n123456789o123456789p123456789q123456789r123456789s123456789t123456789u123456789v123456789 
test -e $outdir || mkdir $outdir
CHECK_OUTPUT satms_tatms2-d.txt -d $outdir -n 2 -t SATMS,TATMS $dfile
rm -rf $outdir
# Outdir length test 2: increase outdir length by 1 to 168.
# Total is 187. Should fail.
outdir="$outdir"8
test -e $outdir || mkdir $outdir
CHECK_FAIL -d $outdir -n 2 -t SATMS,TATMS $dfile
# Outdir length test 3: Use -g 0 to not generate Geolocation product. Total
# decreases to 181. Should pass.
CHECK_OUTPUT satms_tatms2-d-g0.txt -g 0 -d $outdir -n 2 -t SATMS,TATMS $dfile
rm -rf $outdir
# Outdir length test 4: increase outdir length by 6 to 174. With 2 products,
# Total is 187. Should fail.
outdir="$outdir"9p1234
test -e $outdir || mkdir $outdir
CHECK_FAIL -g 0 -d $outdir -n 2 -t SATMS,TATMS $dfile
rmdir $outdir
# Outdir length test 5: Use -S (Simple==Unpackaged form).  Only 1 product per
# file. Increase outdir by 5 to 179. Total is 186. Should pass.
outdir="$outdir"56789
test -e $outdir || mkdir $outdir
CHECK_OUTPUT satms_tatms2-d-S.txt -S -d $outdir -n 2 -t SATMS,TATMS $dfile
rm -rf $outdir
# Outdir length test 6: Increse outdir by 1 to 180. Total is 187. Should fail.
outdir="$outdir"q
test -e $outdir || mkdir $outdir
CHECK_FAIL -S -d $outdir -n 2 -t SATMS,TATMS $dfile
# Outdir length test 7: Use -g 0 will not help. Still one product per file,
# just no geoproduct file. Total is still 187. Should fail.
CHECK_FAIL -g 0 -S -d $outdir -n 2 -t SATMS,TATMS $dfile
rm -rf $outdir

echo Test -A bad values. All should fail.
dfile=$dir/REDRO_npp_d20030125_t0702533_e0711257_b00014_c20111025170507190780_unkn_ada.h5
# -A missting parameter
CHECK_FAIL -A -t REDRO $dfile
# -A non-positive integer parameter
CHECK_FAIL -A 0 -t REDRO $dfile
CHECK_FAIL -A -1 -t REDRO $dfile
CHECK_FAIL -A  319.97 -t REDRO $dfile
CHECK_FAIL -A 32s -t REDRO $dfile
# -A and -n are exclusive
CHECK_FAIL -A 150 -n 5 -t REDRO $dfile

echo Test -g strict - Errors expected
# ICSTT with non-matching geofilename using -g strict
dfile=$dir/ICSTT_npp_d20030125_t0705359_e0713337_b00014_c20111025184111991145_unkn_ada.h5
CHECK_FAIL -n 15 -t ICSTT -g strict $dfile

echo Test -g "<geoproduct>" error checking
dfile=$dir/GCRIO_npp_d20030125_t0702533_e0711257_b00014_c20111025170507184515_unkn_ada.h5
# -t with -g is now allowed.  Test will now fail for lack of REDRO input granules 
CHECK_FAIL -n 8 -t REDRO -g GCRIO $dfile
# One of -t or -g geoproduct should be used
CHECK_FAIL -n 8 $dfile
# Input files have the wrong geoproduct
CHECK_FAIL -n 8 -g GATMO $dfile
dfile=$dir/REDRO_npp_d20030125_t0702533_e0711257_b00014_c20111025170507190780_unkn_ada.h5
# Only one geoproduct is allowed
CHECK_FAIL -n 8 -t REDRO -g GATRO,GCRIO $dfile
# geoproduct is not compatible with data product
CHECK_FAIL -n 8 -t REDRO -g GMODO $dfile

echo Test --like -l with bad parameters. All should fail.
exfile=$dir/GATMO-SATMS-TATMS_npp_d20120403_t2359123_e0000159_b02250_c20120830155722719166_unkn_xxx.h5
dfile="$dir/SATMS_npp_d2012040*.h5 $dir/TATMS_npp_d2012040*.h5"
dfile2=$dir/REDRO_npp_d20030125_t0702533_e0711257_b00014_c20111025170507190780_unkn_ada.h5
# -l without example file.
CHECK_FAIL -l -n 2 $dfile
# -l with non-existing file
CHECK_FAIL -l nosuch -n 2 $dfile
# override with bad -t value. Example file has SATMS and TATMS products but
# -t asks for REDRO.
CHECK_FAIL -l $exfile -t REDRO $dfile
# Example file has SATMS and TATMS but datafiles have REDRO only.
CHECK_FAIL -l $exfile $dfile2

# Two global layout options should fail
CHECK_FAIL --layout CONTI --layout CHUNK=12x12x5 -n4 -t SATMS $dfile

dfile="$dir/GCRIO-ICALI-ICALM-ICCCR-SATMR_npp_d20130126_t0153459_e0155517_b06465_c20130301214009160670_unkn_xxx.h5"
CHECK_OUTPUT gcrio-mixed-applycompact-all.txt  h5dump_grep "-p -H" "" 'COMPRESSION\|COMPACT' --layout=COMPA -n8 -tICALI,ICALM,ICCCR,SATMR $dfile
# Global compact or contiguous layout with any global filter except NONE 
# should fail without trying to create a dataset.
CHECK_FAIL_OUTPUT gcrio-mixed-applycompactGZIP.txt --layout=COMPA --filter=GZIP=7 -n8 -tICALI,ICALM,ICCCR,SATMR $dfile

# Global compact or contiguous layout with any filter for a specific 
# dataset except NONE should fail without trying to create a dataset.
CHECK_FAIL_OUTPUT gcrio-mixed-applycontigSHUF.txt --layout=/All_Data/CrIMSS-CrIS-CLOUD-CLEARED-RAD-IP_All/radiance:CONTI --filter=/All_Data/CrIMSS-CrIS-CLOUD-CLEARED-RAD-IP_All/radiance:SHUF -n8 -tICALI,ICALM,ICCCR,SATMR $dfile

# Post results and exit.
echo
if test $nerrors -eq 0 ; then
    echo "All $TESTNAME tests passed."
    exit $EXIT_SUCCESS
else
    echo "$TESTNAME tests failed with $nerrors errors."
    exit $EXIT_FAILURE
fi
